(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var global = Package.meteor.global;
var meteorEnv = Package.meteor.meteorEnv;
var ECMAScript = Package.ecmascript.ECMAScript;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Random = Package.random.Random;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var Accounts = Package['accounts-base'].Accounts;
var HTTP = Package.http.HTTP;
var HTTPInternals = Package.http.HTTPInternals;
var Identicon = Package['sandstorm-identicons'].Identicon;
var meteorInstall = Package.modules.meteorInstall;
var meteorBabelHelpers = Package['babel-runtime'].meteorBabelHelpers;
var Promise = Package.promise.Promise;
var Symbol = Package['ecmascript-runtime-server'].Symbol;
var Map = Package['ecmascript-runtime-server'].Map;
var Set = Package['ecmascript-runtime-server'].Set;

/* Package-scope variables */
var Packages, DevPackages, UserActions, Grains, RoleAssignments, Contacts, Sessions, SignupKeys, ActivityStats, DeleteStats, FileTokens, SpkTokens, ApiTokens, ApiHosts, Notifications, ActivitySubscriptions, StatsTokens, Misc, Settings, Migrations, StaticAssets, AssetUploadTokens, Plans, AppIndex, KeybaseProfiles, FeatureKey, SetupSession, successfulReferralsCount, isAdmin, isAdminById, findAdminUserForToken, matchWildcardHost, makeWildcardHost, getWildcardOrigin, SandstormDb, appName;

var require = meteorInstall({"node_modules":{"meteor":{"sandstorm-db":{"db.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// packages/sandstorm-db/db.js                                                                                         //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _typeof2 = require("babel-runtime/helpers/typeof");                                                                //
                                                                                                                       //
var _typeof3 = _interopRequireDefault(_typeof2);                                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
// This file defines the database schema.                                                                              // 17
// Useful for debugging: Set the env variable LOG_MONGO_QUERIES to have the server write every                         // 19
// query it makes, so you can see if it's doing queries too often, etc.                                                // 20
if (Meteor.isServer && process.env.LOG_MONGO_QUERIES) {                                                                // 21
  var oldFind = Mongo.Collection.prototype.find;                                                                       // 22
                                                                                                                       //
  Mongo.Collection.prototype.find = function () {                                                                      // 23
    console.log(this._prefix, arguments);                                                                              // 24
    return oldFind.apply(this, arguments);                                                                             // 25
  };                                                                                                                   // 26
} // Helper so that we don't have to if (Meteor.isServer) before declaring indexes.                                    // 27
                                                                                                                       //
                                                                                                                       //
if (Meteor.isServer) {                                                                                                 // 30
  Mongo.Collection.prototype.ensureIndexOnServer = Mongo.Collection.prototype._ensureIndex;                            // 31
} else {                                                                                                               // 32
  Mongo.Collection.prototype.ensureIndexOnServer = function () {};                                                     // 33
} // TODO(soon): Systematically go through this file and add ensureIndexOnServer() as needed.                          // 34
                                                                                                                       //
                                                                                                                       //
var collectionOptions = {                                                                                              // 38
  defineMutationMethods: Meteor.isClient                                                                               // 38
}; // Set to `true` on the client so that method simulation works. Set to `false` on the server                        // 38
// so that we can be extra certain that all mutations must go through methods.                                         // 40
// Users = new Mongo.Collection("users");                                                                              // 42
// The users collection is special and can be accessed through `Meteor.users`.                                         // 43
// See https://docs.meteor.com/#/full/meteor_users.                                                                    // 44
//                                                                                                                     // 45
// There are two distinct types of entries in the users collection: credentials and accounts.                          // 46
//                                                                                                                     // 47
// An account is an owner of app actions, grains, contacts, notifications, profile info, and                           // 48
// payment info. Each account can have multiple credentials linked to it. To log in as an account                      // 49
// you must first authenticate as one of its linked credentials.                                                       // 50
//                                                                                                                     // 51
// A credential identifies a method for authenticating a user, such as an attached Google or Github                    // 52
// account. The reason these exist as separate entries in the collection is to make it easier to                       // 53
// reuse Meteor's authentication libraries while allowing multiple credentials per account. This is                    // 54
// a huge hack, but it works.                                                                                          // 55
//                                                                                                                     // 56
// Every user contains the following fields:                                                                           // 57
//   _id: Unique string ID. For accounts, this is random. For credentials, this is the globally                        // 58
//        stable ID of this credential (based on a SHA-256 hash of the credential description, hex                     // 59
//        encoded).                                                                                                    // 60
//   type: "credential" or "account"                                                                                   // 61
//   createdAt: Date when this entry was added to the collection.                                                      // 62
//   lastActive: Date of the user's most recent interaction with this Sandstorm server.                                // 63
//   services: Object containing login data used by Meteor authentication services.                                    // 64
//   expires: Date when this user should be deleted. Only present for demo users.                                      // 65
//   upgradedFromDemo: If present, the date when this user was upgraded from being a demo user.                        // 66
//                     TODO(cleanup): Unlike other dates in our database, this is stored as a number                   // 67
//                     rather than as a Date object. We should fix that.                                               // 68
//   appDemoId: If present and non-null, then the user is a demo user who arrived via an /appdemo/                     // 69
//              link. This field contains the app ID of the app that the user started out demoing.                     // 70
//              Unlike the `expires` field, this field is not cleared when the user upgrades from                      // 71
//              being a demo user.                                                                                     // 72
//   suspended: If this exists, this account/credential is supsended. Both accounts and credentials                    // 73
//              can be suspended. After some amount of time, the user will be completely deleted                       // 74
//              and removed from the DB.                                                                               // 75
//              It is an object with fields:                                                                           // 76
//                voluntary: Boolean. This is true if the user initiated it. They will have the                        // 77
//                  chance to still login and reverse the suspension/deletion.                                         // 78
//                admin: The userId of the admin who suspended the account.                                            // 79
//                timestamp: Date object. When the suspension occurred.                                                // 80
//                willDelete: Boolean. If true, this account will be deleted after some time.                          // 81
//                                                                                                                     // 82
// Credential users additionally contain the following fields:                                                         // 83
//   unverifiedEmail: If present, a string containing an email address specified by the user.                          // 84
//                    TODO(cleanup): Is this obsolete?                                                                 // 85
//                                                                                                                     // 86
// Account users additionally contain the following fields:                                                            // 87
//   loginCredentials: Array of objects, each containing a field `id` which is the _id of a                            // 88
//                    credential user which is authorized to authenticate this account.                                // 89
//   nonloginCredentials: Array of objects, of the same form as `loginCredentials`. We use                             // 90
//                       a separate array here so that we can use a Mongo index to enforce the                         // 91
//                       invariant that a credetial only be a login credetial for a single account.                    // 92
//   profile: Object containing the data that will be shared with users and grains that come into                      // 93
//            contact with this user. Includes the following fields:                                                   // 94
//       name: String containing the chosen display name of the user.                                                  // 95
//       handle: String containing the users's preferred handle.                                                       // 96
//       picture: _id into the StaticAssets table for the users's picture. If not present,                             // 97
//                an identicon will be used.                                                                           // 98
//       identicon: 32-character hex string used as the seed for the identicon generator, if no                        // 99
//                  profile picture is available.                                                                      // 100
//       pronoun: One of "male", "female", "neutral", or "robot".                                                      // 101
//   primaryEmail: String containing this account's primary email address. Must be a verified adress                   // 102
//                 of one of this account's linked credentials. Call SandstormDb.getUserEmails()                       // 103
//                 to do this checking automatically.                                                                  // 104
//   isAdmin: Boolean indicating whether this account is allowed to access the Sandstorm admin panel.                  // 105
//   signupKey: If this is an invited user, then this field contains their signup key.                                 // 106
//   signupNote: If the user was invited through a link, then this field contains the note that the                    // 107
//               inviter admin attached to the key.                                                                    // 108
//   signupEmail: If the user was invited by email, then this field contains the email address that                    // 109
//                the invite was sent to.                                                                              // 110
//   hasCompletedSignup: True if this account has confirmed its profile and agreed to this server's                    // 111
//                       terms of service.                                                                             // 112
//   plan: _id of an entry in the Plans table which determines the user's quota.                                       // 113
//   planBonus: {storage, compute, grains} bonus amounts to add to the user's plan. The payments                       // 114
//              module writes data here; we merely read it. Missing fields should be treated as                        // 115
//              zeroes. Does not yet include referral bonus, which is calculated separately.                           // 116
//              TODO(cleanup): Use for referral bonus too.                                                             // 117
//   storageUsage: Number of bytes this user is currently storing.                                                     // 118
//   payments: Object defined by payments module, if loaded.                                                           // 119
//   dailySentMailCount: Number of emails sent by this user today; used to limit spam.                                 // 120
//   accessRequests: Object containing the following fields; used to limit spam.                                       // 121
//       count: Number of "request access" emails during sent during the current interval.                             // 122
//       resetOn: Date when the count should be reset.                                                                 // 123
//   referredBy: ID of the Account that referred this user. Only present if this user themselves                       // 124
//       has not get completed a sharing action.                                                                       // 125
//   referredByComplete: ID of the Account that referred this user. `referredBy` becomes                               // 126
//       `referredByComplete` when this user performs their first sharing action. (For Alice to get                    // 127
//       referral credit, she must share with a new user Bob, and Bob must himself share something                     // 128
//       (possibly back to Alice, possibly to a third party).)                                                         // 129
//   referredCompleteDate: The Date at which the completed referral occurred.                                          // 130
//   referredAccountIds: List of Account IDs that this Account has referred. This is used for                          // 131
//                       reliably determining which users's names are safe to display.                                 // 132
//   experiments: Object where each field is an experiment that the user is in, and each value                         // 133
//           is the parameters for that experiment. Typically, the value simply names which                            // 134
//           experiment group which the user is in, where "control" is one group. If an experiment                     // 135
//           is not listed, then the user should not be considered at all for the purpose of that                      // 136
//           experiment. Each experiment may define a point in time where users not already in the                     // 137
//           experiment may be added to it and assigned to a group (for example, at user creation                      // 138
//           time). Current experiments:                                                                               // 139
//       firstTimeBillingPrompt: OBSOLETE                                                                              // 140
//       freeGrainLimit: Value is "control" or or a number indicating the grain limit that the                         // 141
//               user should receive when on the "free" plan, e.g. "Infinity".                                         // 142
                                                                                                                       //
Meteor.users.ensureIndexOnServer("services.google.email", {                                                            // 144
  sparse: 1                                                                                                            // 144
});                                                                                                                    // 144
Meteor.users.ensureIndexOnServer("services.github.emails.email", {                                                     // 145
  sparse: 1                                                                                                            // 145
});                                                                                                                    // 145
Meteor.users.ensureIndexOnServer("services.email.email", {                                                             // 146
  unique: 1,                                                                                                           // 146
  sparse: 1                                                                                                            // 146
});                                                                                                                    // 146
Meteor.users.ensureIndexOnServer("loginCredentials.id", {                                                              // 147
  unique: 1,                                                                                                           // 147
  sparse: 1                                                                                                            // 147
});                                                                                                                    // 147
Meteor.users.ensureIndexOnServer("nonloginCredentials.id", {                                                           // 148
  sparse: 1                                                                                                            // 148
});                                                                                                                    // 148
Meteor.users.ensureIndexOnServer("services.google.id", {                                                               // 149
  unique: 1,                                                                                                           // 149
  sparse: 1                                                                                                            // 149
});                                                                                                                    // 149
Meteor.users.ensureIndexOnServer("services.github.id", {                                                               // 150
  unique: 1,                                                                                                           // 150
  sparse: 1                                                                                                            // 150
});                                                                                                                    // 150
Meteor.users.ensureIndexOnServer("suspended.willDelete", {                                                             // 151
  sparse: 1                                                                                                            // 151
});                                                                                                                    // 151
Packages = new Mongo.Collection("packages", collectionOptions); // Packages which are installed or downloading.        // 153
//                                                                                                                     // 155
// Each contains:                                                                                                      // 156
//   _id:  128-bit prefix of SHA-256 hash of spk file, hex-encoded.                                                    // 157
//   status:  String.  One of "download", "verify", "unpack", "analyze", "ready", "failed", "delete"                   // 158
//   progress:  Float.  -1 = N/A, 0-1 = fractional progress (e.g. download percentage),                                // 159
//       >1 = download byte count.                                                                                     // 160
//   error:  If status is "failed", error message string.                                                              // 161
//   manifest:  If status is "ready", the package manifest.  See "Manifest" in package.capnp.                          // 162
//   appId:  If status is "ready", the application ID string.  Packages representing different                         // 163
//       versions of the same app have the same appId.  The spk tool defines the app ID format                         // 164
//       and can cryptographically verify that a package belongs to a particular app ID.                               // 165
//   shouldCleanup:  If true, a reference to this package was recently dropped, and the package                        // 166
//       collector should at some point check whether there are any other references and, if not,                      // 167
//       delete the package.                                                                                           // 168
//   url:  When status is "download", the URL from which the SPK can be obtained, if provided.                         // 169
//   isAutoUpdated: This package was downloaded as part of an auto-update. We shouldn't clean it up                    // 170
//     even if it has no users.                                                                                        // 171
//   authorPgpKeyFingerprint: Verified PGP key fingerprint (SHA-1, hex, all-caps) of the app                           // 172
//     packager.                                                                                                       // 173
                                                                                                                       //
DevPackages = new Mongo.Collection("devpackages", collectionOptions); // List of packages currently made available via the dev tools running on the local machine.
// This is normally empty; the only time it is non-empty is when a developer is using the spk tool                     // 177
// on the local machine to publish an under-development app to this server. That should only ever                      // 178
// happen on developers' desktop machines.                                                                             // 179
//                                                                                                                     // 180
// While a dev package is published, it automatically appears as installed by every user of the                        // 181
// server, and it overrides all packages with the same application ID. If any instances of those                       // 182
// packages are currently open, they are killed and reset on publish.                                                  // 183
//                                                                                                                     // 184
// When the dev tool disconnects, the package is automatically unpublished, and any open instances                     // 185
// are again killed and refreshed.                                                                                     // 186
//                                                                                                                     // 187
// Each contains:                                                                                                      // 188
//   _id:  The package ID string (as with Packages._id).                                                               // 189
//   appId: The app ID this package is intended to override (as with Packages.appId).                                  // 190
//   timestamp:  Time when the package was last updated. If this changes while the package is                          // 191
//     published, all running instances are reset. This is used e.g. to reset the app each time                        // 192
//     changes are made to the source code.                                                                            // 193
//   manifest:  The app's manifest, as with Packages.manifest.                                                         // 194
//   mountProc: True if the supervisor should mount /proc.                                                             // 195
                                                                                                                       //
UserActions = new Mongo.Collection("userActions", collectionOptions); // List of actions that each user has installed which create new grains.  Each app may install
// some number of actions (usually, one).                                                                              // 199
//                                                                                                                     // 200
// Each contains:                                                                                                      // 201
//   _id:  random                                                                                                      // 202
//   userId:  Account ID of the user who has installed this action.                                                    // 203
//   packageId:  Package used to run this action.                                                                      // 204
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                                // 205
//   appTitle:  Same as Packages.findOne(packageId).manifest.appTitle; denormalized so                                 // 206
//       that clients can access it without subscribing to the Packages collection.                                    // 207
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                            // 208
//       searchability.                                                                                                // 209
//   appMarketingVersion:  Human-readable presentation of the app version, e.g. "2.9.17"                               // 210
//   title: JSON-encoded LocalizedText title for this action, e.g.                                                     // 211
//       `{defaultText: "New Spreadsheet"}`.                                                                           // 212
//   nounPhrase: JSON-encoded LocalizedText describing what is created when this action is run.                        // 213
//   command:  Manifest.Command to run this action (see package.capnp).                                                // 214
                                                                                                                       //
Grains = new Mongo.Collection("grains", collectionOptions); // Grains belonging to users.                              // 216
//                                                                                                                     // 218
// Each contains:                                                                                                      // 219
//   _id:  random                                                                                                      // 220
//   packageId:  _id of the package of which this grain is an instance.                                                // 221
//   packageSalt: If present, a random string that will used in session ID generation. This field                      // 222
//       is usually updated when `packageId` is updated, triggering automatic refreshes for                            // 223
//       clients with active sessions.                                                                                 // 224
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                                // 225
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                            // 226
//       searchability.                                                                                                // 227
//   userId: The _id of the account that owns this grain.                                                              // 228
//   identityId: The identity ID by which the grain's owner is identified within this grain. Grains                    // 229
//       do not observe global user IDs, but rather grain-local user IDs, generated when each user                     // 230
//       first obtains access to the grain.                                                                            // 231
//   title:  Human-readable string title, as chosen by the user.                                                       // 232
//   lastUsed:  Date when the grain was last used by a user.                                                           // 233
//   private: If true, then knowledge of `_id` does not suffice to open this grain.                                    // 234
//   cachedViewInfo: The JSON-encoded result of `UiView.getViewInfo()`, cached from the most recent                    // 235
//                   time a session to this grain was opened.                                                          // 236
//   trashed: If present, the Date when this grain was moved to the trash bin. Thirty days after                       // 237
//            this date, the grain will be automatically deleted.                                                      // 238
//   suspended: If true, the owner of this grain has been suspended. They will soon be deleted,                        // 239
//              so treat this grain the same as "trashed". It is denormalized out of Users for ease                    // 240
//              of querying.                                                                                           // 241
//   ownerSeenAllActivity: True if the owner has viewed the grain since the last activity event                        // 242
//       occurred. See also ApiTokenOwner.user.seenAllActivity.                                                        // 243
//   size: On-disk size of the grain in bytes.                                                                         // 244
//                                                                                                                     // 245
// The following fields *might* also exist. These are temporary hacks used to implement e-mail and                     // 246
// web publishing functionality without powerbox support; they will be replaced once the powerbox                      // 247
// is implemented.                                                                                                     // 248
//   publicId:  An id used to publicly identify this grain. Used e.g. to route incoming e-mail and                     // 249
//       web publishing. This field is initialized when first requested by the app.                                    // 250
                                                                                                                       //
Grains.ensureIndexOnServer("userId");                                                                                  // 252
Grains.ensureIndexOnServer("cachedViewInfo.matchRequests.tags.id", {                                                   // 253
  sparse: 1                                                                                                            // 253
});                                                                                                                    // 253
RoleAssignments = new Mongo.Collection("roleAssignments", collectionOptions); // *OBSOLETE* Before `user` was a variant of ApiTokenOwner, this collection was used to store edges
// in the permissions sharing graph. This functionality has been subsumed by the ApiTokens                             // 257
// collection.                                                                                                         // 258
                                                                                                                       //
Contacts = new Mongo.Collection("contacts", collectionOptions); // Edges in the social graph.                          // 260
//                                                                                                                     // 262
// If Alice has Bob as a contact, then she is allowed to see Bob's profile information and Bob                         // 263
// will show up in her user-picker UI for actions like share-by-identity.                                              // 264
//                                                                                                                     // 265
// Contacts are not symmetric. Bob might be one of Alice's contacts even if Alice is not one of                        // 266
// Bob's.                                                                                                              // 267
//                                                                                                                     // 268
// Each contains:                                                                                                      // 269
//   _id: random                                                                                                       // 270
//   ownerId: The accountId of the user account who owns this contact.                                                 // 271
//   petname: Human-readable label chosen by and only visible to the owner. Uniquely identifies                        // 272
//            the contact to the owner.                                                                                // 273
//   created: Date when this contact was created.                                                                      // 274
//   accountId: The `_id` of the acount whose contact info this contains.                                              // 275
                                                                                                                       //
Sessions = new Mongo.Collection("sessions", collectionOptions); // UI sessions open to particular grains.  A new session is created each time a user opens a grain.
//                                                                                                                     // 279
// Each contains:                                                                                                      // 280
//   _id:  String generated as a SHA256 hash of the grain ID, the user ID, a salt generated by the                     // 281
//       client, and the grain's `packageSalt`.                                                                        // 282
//   grainId:  _id of the grain to which this session is connected.                                                    // 283
//   hostId: ID part of the hostname from which this grain is being served. I.e. this replaces the                     // 284
//       '*' in WILDCARD_HOST.                                                                                         // 285
//   tabId: Random value unique to the grain tab in which this session is displayed. Typically                         // 286
//       every session has a different `tabId`, but embedded sessions (including in the powerbox)                      // 287
//       have the same `tabId` as the outer session.                                                                   // 288
//   timestamp:  Time of last keep-alive message to this session.  Sessions time out after some                        // 289
//       period.                                                                                                       // 290
//   userId:  Account ID of the user who owns this session.                                                            // 291
//   identityId:  Identity ID by which the user is known to the grain.                                                 // 292
//   hashedToken: If the session is owned by an anonymous user, the _id of the entry in ApiTokens                      // 293
//       that was used to open it. Note that for old-style sharing (i.e. when !grain.private),                         // 294
//       anonymous users can get access without an API token and so neither userId nor hashedToken                     // 295
//       are present.                                                                                                  // 296
//   powerboxView: Information about a server-initiated powerbox interaction taking place in this                      // 297
//       session. When the client sees a `powerboxView` appear on the session, it opens the                            // 298
//       powerbox popup according to the contents. This field is an object containing one of:                          // 299
//     offer: A capability is being offered to the user by the app. This is an object containing:                      // 300
//       token: For a non-UiView capability, the API token that can be used to restore this                            // 301
//           capability.                                                                                               // 302
//       uiView: A UiView capability. This object contains one of:                                                     // 303
//         tokenId: The _id of an ApiToken belonging to the current user.                                              // 304
//         token: A full webkey token which can be opened by an anonymous user.                                        // 305
//     fulfill: A capability is being offered which fulfills the active powerbox request. This                         // 306
//         is an object with members:                                                                                  // 307
//       token: The SturdyRef of the fulfilling capability. This token can only be used in a call                      // 308
//           to claimRequest() by the requesting                                                                       // 309
//           grain.                                                                                                    // 310
//       descriptor: Packed-base64 PowerboxDescriptor for the capability.                                              // 311
//   powerboxRequest: If present, this session is a powerbox request session. Object containing:                       // 312
//     descriptors: Array of PowerboxDescriptors representing the request.                                             // 313
//     requestingSession: Session ID of the session initiating the request.                                            // 314
//   viewInfo: The UiView.ViewInfo corresponding to the underlying UiSession. This isn't populated                     // 315
//       until newSession is called on the UiView.                                                                     // 316
//   permissions: The permissions for the current user on this UiView. This isn't populated                            // 317
//       until newSession is called on the UiView.                                                                     // 318
//   hasLoaded: Marked as true by the proxy when the underlying UiSession has responded to its first                   // 319
//       request                                                                                                       // 320
                                                                                                                       //
SignupKeys = new Mongo.Collection("signupKeys", collectionOptions); // Invite keys which may be used by users to get access to Sandstorm.
//                                                                                                                     // 324
// Each contains:                                                                                                      // 325
//   _id:  random                                                                                                      // 326
//   used:  Boolean indicating whether this key has already been consumed.                                             // 327
//   note:  Text note assigned when creating key, to keep track of e.g. whom the key was for.                          // 328
//   email: If this key was sent as an email invite, the email address to which it was sent.                           // 329
                                                                                                                       //
ActivityStats = new Mongo.Collection("activityStats", collectionOptions); // Contains usage statistics taken on a regular interval. Each entry is a data point.
//                                                                                                                     // 333
// Each contains:                                                                                                      // 334
//   timestamp: Date when measurements were taken.                                                                     // 335
//   daily: Contains stats counts pertaining to the last day before the sample time.                                   // 336
//   weekly: Contains stats counts pertaining to the last seven days before the sample time.                           // 337
//   monthly: Contains stats counts pertaining to the last thirty days before the timestamp.                           // 338
//                                                                                                                     // 339
// Each of daily, weekly, and monthly contains:                                                                        // 340
//   activeUsers: The number of unique users who have used a grain on the server in the time                           // 341
//       interval. Only counts logged-in users.                                                                        // 342
//   demoUsers: Demo users.                                                                                            // 343
//   appDemoUsers: Users that came in through "app demo".                                                              // 344
//   activeGrains: The number of unique grains that have been used in the time interval.                               // 345
//   apps: An object indexed by app ID recording, for each app:                                                        // 346
//       owners: Number of unique owners of this app (counting only grains that still exist).                          // 347
//       sharedUsers: Number of users who have accessed other people's grains of this app (counting                    // 348
//         only grains that still exist).                                                                              // 349
//       grains: Number of active grains of this app (that still exist).                                               // 350
//       deleted: Number of non-demo grains of this app that were deleted.                                             // 351
//       demoed: Number of demo grains created and expired.                                                            // 352
//       appDemoUsers: Number of app demos initiated with this app.                                                    // 353
                                                                                                                       //
DeleteStats = new Mongo.Collection("deleteStats", collectionOptions); // Contains records of objects that were deleted, for stat-keeping purposes.
//                                                                                                                     // 357
// Each contains:                                                                                                      // 358
//   type: "grain" or "user" or "demoGrain" or "demoUser" or "appDemoUser"                                             // 359
//   lastActive: Date of the user's or grain's last activity.                                                          // 360
//   appId: For type = "grain", the app ID of the grain. For type = "appDemoUser", the app ID they                     // 361
//     arrived to demo. For others, undefined.                                                                         // 362
//   experiments: The experiments the user (or owner of the grain) was in. See user.experiments.                       // 363
                                                                                                                       //
FileTokens = new Mongo.Collection("fileTokens", collectionOptions); // Tokens corresponding to backup files that are currently stored on the server. A user receives
// a token when they create a backup file (either by uploading it, or by backing up one of their                       // 367
// grains) and may use the token to read the file (either to download it, or to restore a new                          // 368
// grain from it).                                                                                                     // 369
//                                                                                                                     // 370
// Each contains:                                                                                                      // 371
//   _id:       The unguessable token string.                                                                          // 372
//   name:      Suggested filename.                                                                                    // 373
//   timestamp: File creation time. Used to figure out when the token and file should be wiped.                        // 374
                                                                                                                       //
SpkTokens = new Mongo.Collection("spkTokens", collectionOptions); // A lot like FileTokens, but for SPK uploads.       // 376
//                                                                                                                     // 378
// Each contains:                                                                                                      // 379
//   _id:       The unguessable token string.                                                                          // 380
//   timestamp: Creation time. Used to figure out when the token should be wiped.                                      // 381
                                                                                                                       //
ApiTokens = new Mongo.Collection("apiTokens", collectionOptions); // Access tokens for APIs exported by apps.          // 383
//                                                                                                                     // 385
// Originally API tokens were only used by external users through the HTTP API endpoint. However,                      // 386
// now they are also used to implement SturdyRefs, not just held by external users, but also when                      // 387
// an app holds a SturdyRef to another app within the same server. See the various `save()`,                           // 388
// `restore()`, and `drop()` methods in `grain.capnp` (on `SandstormApi`, `AppPersistent`, and                         // 389
// `MainView`) -- the fields of type `Data` are API tokens.                                                            // 390
//                                                                                                                     // 391
// Each contains:                                                                                                      // 392
//   _id:       A SHA-256 hash of the token, base64-encoded.                                                           // 393
//   grainId:   The grain servicing this API. (Not present if the API isn't serviced by a grain.)                      // 394
//   accountId: For UiView capabilities, this is the account for which the view is attenuated.                         // 395
//              That is, the UiView's newSession() method will intersect the requested permissions                     // 396
//              with this account's permissions before forwarding on to the underlying app. Put                        // 397
//              another way, this identifies the user who *created* this ApiToken, i.e. the sharer,                    // 398
//              NOT the receiver (which is identified by `owner`). If `accountId` is not present,                      // 399
//              then no account attenuation is applied, i.e. this is a raw UiView as implemented by                    // 400
//              the app. The `roleAssignment` field, below, may still apply. For non-UiView                            // 401
//              capabilities, `accountId` is never present.                                                            // 402
//   roleAssignment: If this API token represents a UiView, this field contains a JSON-encoded                         // 403
//              Grain.ViewSharingLink.RoleAssignment representing the permissions it carries. These                    // 404
//              permissions will be intersected with those held by `accountId` when the view is                        // 405
//              opened.                                                                                                // 406
//   forSharing: If true, requests sent to the HTTP API endpoint with this token will be treated as                    // 407
//              anonymous rather than as directly associated with `accountId`. This has no effect                      // 408
//              on the permissions granted.                                                                            // 409
//   objectId:  If present, this token represents an arbitrary Cap'n Proto capability exported by                      // 410
//              the app or its supervisor (whereas without this it strictly represents UiView).                        // 411
//              sturdyRef is the JSON-encoded SupervisorObjectId (defined in `supervisor.capnp`).                      // 412
//              Note that if the SupervisorObjectId contains an AppObjectId, that field is                             // 413
//              treated as type AnyPointer, and so encoded as a raw Cap'n Proto message.                               // 414
//   frontendRef: If present, this token actually refers to an object implemented by the front-end,                    // 415
//              not a particular grain. (`grainId` and `accountId` are not set.) This is an object                     // 416
//              containing exactly one of the following fields:                                                        // 417
//       notificationHandle: A `Handle` for an ongoing notification, as returned by                                    // 418
//                           `NotificationTarget.addOngoing`. The value is an `_id` from the                           // 419
//                           `Notifications` collection.                                                               // 420
//       ipNetwork: An IpNetwork capability that is implemented by the frontend. Eventually, this                      // 421
//                  will be moved out of the frontend and into the backend, but we'll migrate the                      // 422
//                  database when that happens. This field contains the boolean true to signify that                   // 423
//                  it has been set.                                                                                   // 424
//       ipInterface: Ditto IpNetwork, except it's an IpInterface.                                                     // 425
//       emailVerifier: An EmailVerifier capability that is implemented by the frontend. The                           // 426
//                      value is an object containing the fields `id` and `services`. `id` is the                      // 427
//                      value returned by `EmailVerifier.getId()` and is used as part of a                             // 428
//                      powerbox query for matching verified emails. `services` is a                                   // 429
//                      list of names of authentication providers that are trusted to verify                           // 430
//                      addresses. If `services` is omitted or falsy, all configured authentication                    // 431
//                      providers are trusted. Note that a malicious user could specify invalid                        // 432
//                      names in the list; they should be ignored.                                                     // 433
//       verifiedEmail: An VerifiedEmail capability that is implemented by the frontend.                               // 434
//                      An object containing `verifierId`, `tabId`, and `address`.                                     // 435
//       identity: An Identity capability. The field is the account ID.                                                // 436
//       http: An ApiSession capability pointing to an external HTTP service. Object containing:                       // 437
//           url: Base URL of the external service.                                                                    // 438
//           auth: Authentication mechanism. Object containing one of:                                                 // 439
//               none: Value "null". Indicates no authorization.                                                       // 440
//               bearer: A bearer token to pass in the `Authorization: Bearer` header on all                           // 441
//                   requests. Encrypted with nonce 0.                                                                 // 442
//               basic: A `{username, password}` object. The password is encrypted with nonce 0.                       // 443
//                   Before encryption, the password is padded to 32 bytes by appending NUL bytes,                     // 444
//                   in order to mask the length of small passwords.                                                   // 445
//               refresh: An OAuth refresh token, which can be exchanged for an access token.                          // 446
//                   Encrypted with nonce 0.                                                                           // 447
//               TODO(security): How do we protect URLs that directly embed their secret? We don't                     // 448
//                   want to encrypt the full URL since this would make it hard to show a                              // 449
//                   meaningful audit UI, but maybe we could figure out a way to extract the key                       // 450
//                   part and encrypt it separately?                                                                   // 451
//   parentToken: If present, then this token represents exactly the capability represented by                         // 452
//              the ApiToken with _id = parentToken, except possibly (if it is a UiView) attenuated                    // 453
//              by `roleAssignment` (if present). To facilitate permissions computations, if the                       // 454
//              capability is a UiView, then `grainId` is set to the backing grain, and `accountId`                    // 455
//              is set to the account that shared the view. Neither `objectId` nor `frontendRef`                       // 456
//              is present when `parentToken` is present.                                                              // 457
//   parentTokenKey: The actual parent token -- whereas `parentToken` is only the parent token ID                      // 458
//              (hash). `parentTokenFull` is encrypted with nonce 0 (see below). This is needed                        // 459
//              in particular when the parent contains encrypted fields, since those would need to                     // 460
//              be decrypted using this key. If the parent contains no encrypted fields then                           // 461
//              `parentTokenKey` may be omitted from the child.                                                        // 462
//   petname:   Human-readable label for this access token, useful for identifying tokens for                          // 463
//              revocation. This should be displayed when visualizing incoming capabilities to                         // 464
//              the grain identified by `grainId`.                                                                     // 465
//   created:   Date when this token was created.                                                                      // 466
//   revoked:   If true, then this sturdyref has been revoked and can no longer be restored. It may                    // 467
//              become un-revoked in the future.                                                                       // 468
//   trashed:   If present, the Date when this token was moved to the trash bin. Thirty days after                     // 469
//              this date, the token will be automatically deleted.                                                    // 470
//   suspended: If true, the owner of this token has been suspended. They will soon be deleted,                        // 471
//              so treat this token the same as "trashed". It is denormalized out of Users for                         // 472
//              ease of querying.                                                                                      // 473
//   expires:   Optional expiration Date. If undefined, the token does not expire.                                     // 474
//   lastUsed:  Optional Date when this token was last used.                                                           // 475
//   owner:     A `ApiTokenOwner` (defined in `supervisor.capnp`, stored as a JSON object)                             // 476
//              as passed to the `save()` call that created this token. If not present, treat                          // 477
//              as `webkey` (the default for `ApiTokenOwner`).                                                         // 478
//   expiresIfUnused:                                                                                                  // 479
//              Optional Date after which the token, if it has not been used yet, expires.                             // 480
//              This field should be cleared on a token's first use.                                                   // 481
//   requirements: List of conditions which must hold for this token to be considered valid.                           // 482
//              Semantically, this list specifies the powers which were *used* to originally                           // 483
//              create the token. If any condition in the list becomes untrue, then the token must                     // 484
//              be considered revoked, and all live refs and sturdy refs obtained transitively                         // 485
//              through it must also become revoked. Each item is the JSON serialization of the                        // 486
//              `MembraneRequirement` structure defined in `supervisor.capnp`.                                         // 487
//   hasApiHost: If true, there is an entry in ApiHosts for this token, which will need to be                          // 488
//              cleaned up when the token is.                                                                          // 489
//                                                                                                                     // 490
// It is important to note that a token's owner and provider are independent from each other. To                       // 491
// illustrate, here is an approximate definition of ApiToken in pseudo Cap'n Proto schema language:                    // 492
//                                                                                                                     // 493
// struct ApiToken {                                                                                                   // 494
//   owner :ApiTokenOwner;                                                                                             // 495
//   provider :union {                                                                                                 // 496
//     grain :group {                                                                                                  // 497
//       grainId :Text;                                                                                                // 498
//       union {                                                                                                       // 499
//         uiView :group {                                                                                             // 500
//           accountId :Text;                                                                                          // 501
//           roleAssignment :RoleAssignment;                                                                           // 502
//           forSharing :Bool;                                                                                         // 503
//         }                                                                                                           // 504
//         objectId :SupervisorObjectId;                                                                               // 505
//       }                                                                                                             // 506
//     }                                                                                                               // 507
//     frontendRef :union {                                                                                            // 508
//       notificationHandle :Text;                                                                                     // 509
//       ipNetwork :Bool;                                                                                              // 510
//       ipInterface :Bool;                                                                                            // 511
//       emailVerifier :group {                                                                                        // 512
//         id :Text;                                                                                                   // 513
//         services :List(String);                                                                                     // 514
//       }                                                                                                             // 515
//       verifiedEmail :group {                                                                                        // 516
//         verifierId :Text;                                                                                           // 517
//         tabId :Text;                                                                                                // 518
//         address :Text;                                                                                              // 519
//       }                                                                                                             // 520
//       identity :Text;                                                                                               // 521
//       http :group {                                                                                                 // 522
//         url :Text;                                                                                                  // 523
//         auth :union {                                                                                               // 524
//           none :Void;                                                                                               // 525
//           bearer :Text;                                                                                             // 526
//           basic :group { username :Text; password :Text; }                                                          // 527
//           refresh :Text;                                                                                            // 528
//         }                                                                                                           // 529
//       }                                                                                                             // 530
//     }                                                                                                               // 531
//     child :group {                                                                                                  // 532
//       parentToken :Text;                                                                                            // 533
//       union {                                                                                                       // 534
//         uiView :group {                                                                                             // 535
//           grainId :Text;                                                                                            // 536
//           accountId :Text;                                                                                          // 537
//           roleAssignment :RoleAssignment = (allAccess = ());                                                        // 538
//         }                                                                                                           // 539
//         other :Void;                                                                                                // 540
//       }                                                                                                             // 541
//     }                                                                                                               // 542
//   }                                                                                                                 // 543
//   requirements: List(Supervisor.MembraneRequirement);                                                               // 544
//   ...                                                                                                               // 545
// }                                                                                                                   // 546
//                                                                                                                     // 547
// ENCRYPTION                                                                                                          // 548
//                                                                                                                     // 549
// We want to make sure that someone who obtains a copy of the database cannot use it to gain live                     // 550
// credentials.                                                                                                        // 551
//                                                                                                                     // 552
// The actual token corresponding to an ApiToken entry is not stored in the entry itself. Instead,                     // 553
// the ApiToken's `_id` is constructed as a SHA256 hash of the actual token. To use an ApiToken                        // 554
// in the live system, you must present the original token.                                                            // 555
//                                                                                                                     // 556
// Additionally, some ApiToken entries contain tokens to third-party services, e.g. OAuth tokens                       // 557
// or even passwords. Such tokens are encrypted, using the ApiToken entry's own full token (which,                     // 558
// again, is not stored in the database) as the encryption key.                                                        // 559
//                                                                                                                     // 560
// When such encryption is applied, the cipher used is ChaCha20. All API tokens are 256-bit base64                     // 561
// strings, hence can be used directly as the key. No MAC is applied, because this scheme is not                       // 562
// intended to protect against attackers who have write access to the database -- such an attacker                     // 563
// could almost certainly do more damage by modifying the non-encrypted fields anyway. (Put another                    // 564
// way, if we wanted to MAC something, we'd need to MAC the entire ApiToken structure, not just                        // 565
// the encrypted key. But we don't have a way to do that at present.)                                                  // 566
//                                                                                                                     // 567
// ChaCha20 requires a nonce. Luckily, all of the fields we wish to encrypt are immutable, so we                       // 568
// don't have to worry about tracking nonces over time -- we can just assign a static nonce to each                    // 569
// field. Moreover, many (currently, all) of these fields are mutually exclusive, so can even share                    // 570
// nonces. Currently, nonces map to fields as follows:                                                                 // 571
//                                                                                                                     // 572
// nonce 0:                                                                                                            // 573
//     parentTokenKey                                                                                                  // 574
//     frontendRef.http.auth.basic.password                                                                            // 575
//     frontendRef.http.auth.bearer                                                                                    // 576
//     frontendRef.http.auth.refresh                                                                                   // 577
                                                                                                                       //
ApiTokens.ensureIndexOnServer("grainId", {                                                                             // 579
  sparse: 1                                                                                                            // 579
});                                                                                                                    // 579
ApiTokens.ensureIndexOnServer("owner.user.accountId", {                                                                // 580
  sparse: 1                                                                                                            // 580
});                                                                                                                    // 580
ApiTokens.ensureIndexOnServer("frontendRef.emailVerifier.id", {                                                        // 581
  sparse: 1                                                                                                            // 581
});                                                                                                                    // 581
ApiHosts = new Mongo.Collection("apiHosts", collectionOptions); // Allows defining some limited static behavior for an API host when accessed unauthenticated. This
// mainly exists to allow backwards-compatibility with client applications that expect to be able                      // 585
// to probe an API host without authentication to determine capabilities such as DAV protocols                         // 586
// supported, before authenticating to perform real requests. An app can specify these properties                      // 587
// when creating an offerTemplate.                                                                                     // 588
//                                                                                                                     // 589
// Each contains:                                                                                                      // 590
//   _id:          apiHostIdHashForToken() of the corresponding API token.                                             // 591
//   hash2:        hash(hash(token)), aka hash(ApiToken._id). Used to allow ApiHosts to be cleaned                     // 592
//                 up when ApiTokens are deleted.                                                                      // 593
//   options:      Specifies how to respond to unauthenticated OPTIONS requests on this host.                          // 594
//                 This is an object containing fields:                                                                // 595
//     dav:        List of strings specifying DAV header `compliance-class`es, e.g. "1" or                             // 596
//                 "calendar-access". https://tools.ietf.org/html/rfc4918#section-10.1                                 // 597
//   resources:    Object mapping URL paths (including initial '/') to static HTTP responses to                        // 598
//                 give when those paths are accessed unauthenticated. Due to Mongo disliking '.'                      // 599
//                 and '$' in keys, these characters must be escaped as '\uFF0E' and '\uFF04'                          // 600
//                 (see SandstormDb.escapeMongoKey). Each value in this map is an object with                          // 601
//                 fields:                                                                                             // 602
//     type:       Content-Type.                                                                                       // 603
//     language:   Content-Language.                                                                                   // 604
//     encoding:   Content-Encoding.                                                                                   // 605
//     body:       Entity-body as a string or buffer.                                                                  // 606
                                                                                                                       //
Notifications = new Mongo.Collection("notifications", collectionOptions); // Notifications for a user.                 // 608
//                                                                                                                     // 610
// Each contains:                                                                                                      // 611
//   _id:          random                                                                                              // 612
//   grainId:      The grain originating this notification, if any.                                                    // 613
//   userId:       Account ID of the user receiving the notification.                                                  // 614
//   text:         The JSON-ified LocalizedText to display in the notification.                                        // 615
//   isUnread:     Boolean indicating if this notification is unread.                                                  // 616
//   timestamp:    Date when this notification was last updated                                                        // 617
//   eventType:    If this notification is due to an activity event, this is the numeric index                         // 618
//                 of the event type on the grain's ViewInfo.                                                          // 619
//   count:        The number of times this exact event has repeated. Identical events are                             // 620
//                 aggregated by incrementing the count.                                                               // 621
//   initiatingAccount: Account ID of the user who initiated this notification.                                        // 622
//   initiatorAnonymous: True if the initiator is an anonymous user. If neither this nor                               // 623
//                 initiatingAccount is present, the notification is not from a user.                                  // 624
//   path:         Path inside the grain to which the user should be directed if they click on                         // 625
//                 the notification.                                                                                   // 626
//   ongoing:      If present, this is an ongoing notification, and this field contains an                             // 627
//                 ApiToken referencing the `OngoingNotification` capability.                                          // 628
//   admin:        If present, this is a notification intended for an admin.                                           // 629
//     action:     If present, this is a (string) link that the notification should direct the                         // 630
//                 admin to.                                                                                           // 631
//     type:       The type of notification -- currently can only be "reportStats".                                    // 632
//   appUpdates:   If present, this is an app update notification. It is an object with the appIds                     // 633
//                 as keys.                                                                                            // 634
//     $appId:     The appId that has an outstanding update.                                                           // 635
//       packageId: The packageId that it will update to.                                                              // 636
//       name: The name of the app. (appTitle from package.manifest)                                                   // 637
//       version: The app's version number. (appVersion from package.manifest)                                         // 638
//       marketingVersion: String marketing version of this app. (appMarketingVersion from package.manifest)           // 639
//   referral:     If this boolean field is true, then treat this notification as a referral                           // 640
//                 notification. This causes text to be ignored, since we need custom logic.                           // 641
//   mailingListBonus: Like `referral`, but notify the user about the mailing list bonus. This is                      // 642
//                 a one-time notification only to Oasis users who existed when the bonus program                      // 643
//                 was implemented.                                                                                    // 644
//   identityChanges: If this boolean field is true, this notification should show a warning about                     // 645
//                 changes to the identity model described in:                                                         // 646
//                 https://sandstorm.io/news/2017-05-08-refactoring-identities                                         // 647
                                                                                                                       //
ActivitySubscriptions = new Mongo.Collection("activitySubscriptions", collectionOptions); // Activity events to which a user is subscribed.
//                                                                                                                     // 651
// Each contains:                                                                                                      // 652
//   _id:          random                                                                                              // 653
//   accountId:    Who is subscribed.                                                                                  // 654
//   grainId:      Grain to which subscription applies.                                                                // 655
//   threadPath:   If present, the subscription is on a specific thread. Otherwise, it is on the                       // 656
//                 whole grain.                                                                                        // 657
//   mute:         If true, this is an anti-subscription -- matching events should NOT notify.                         // 658
//                 This allows is useful to express:                                                                   // 659
//                 - A user wants to subscribe to a grain but mute a specific thread.                                  // 660
//                 - The owner of a grain does not want notifications (normally, they are                              // 661
//                   implicitly subscribed).                                                                           // 662
//                 - A user no longer wishes to be implicitly subscribed to threads in a grain on                      // 663
//                   which they comment, so they mute the grain.                                                       // 664
                                                                                                                       //
ActivitySubscriptions.ensureIndexOnServer("accountId");                                                                // 666
ActivitySubscriptions.ensureIndexOnServer({                                                                            // 667
  "grainId": 1,                                                                                                        // 667
  "threadPath": 1                                                                                                      // 667
});                                                                                                                    // 667
StatsTokens = new Mongo.Collection("statsTokens", collectionOptions); // Access tokens for the Stats collection        // 669
//                                                                                                                     // 671
// These tokens are used for accessing the ActivityStats collection remotely                                           // 672
// (ie. from a dashboard webapp)                                                                                       // 673
//                                                                                                                     // 674
// Each contains:                                                                                                      // 675
//   _id:       The token. At least 128 bits entropy (Random.id(22)).                                                  // 676
                                                                                                                       //
Misc = new Mongo.Collection("misc", collectionOptions); // Miscellaneous configuration and other settings              // 678
//                                                                                                                     // 680
// This table is currently only used for persisting BASE_URL from one session to the next,                             // 681
// but in general any miscellaneous settings should go in here                                                         // 682
//                                                                                                                     // 683
// Each contains:                                                                                                      // 684
//   _id:       The name of the setting. eg. "BASE_URL"                                                                // 685
//   value:     The value of the setting.                                                                              // 686
                                                                                                                       //
Settings = new Mongo.Collection("settings", collectionOptions); // Settings for this Sandstorm instance go here. They are configured through the adminSettings
// route. This collection differs from misc in that any admin user can update it through the admin                     // 690
// interface.                                                                                                          // 691
//                                                                                                                     // 692
// Each contains:                                                                                                      // 693
//   _id:       The name of the setting. eg. "smtpConfig"                                                              // 694
//   value:     The value of the setting.                                                                              // 695
//   automaticallyReset: Sometimes the server needs to automatically reset a setting. When it does                     // 696
//                       so, it will also write an object to this field indicating why the reset was                   // 697
//                       needed. That object can have the following variants:                                          // 698
//       baseUrlChangedFrom: The reset was due to BASE_URL changing. This field contains a string                      // 699
//                           with the old BASE_URL.                                                                    // 700
//   preinstalledApps: A list of objects:                                                                              // 701
//     appId: The Packages.appId of the app to install                                                                 // 702
//     status: packageId                                                                                               // 703
//     packageId: The Packages._id of the app to install                                                               // 704
//                                                                                                                     // 705
//   potentially other fields that are unique to the setting                                                           // 706
                                                                                                                       //
Migrations = new Mongo.Collection("migrations", collectionOptions); // This table tracks which migrations we have applied to this instance.
// It contains a single entry:                                                                                         // 710
//   _id:       "migrations_applied"                                                                                   // 711
//   value:     The number of migrations this instance has successfully completed.                                     // 712
                                                                                                                       //
StaticAssets = new Mongo.Collection("staticAssets", collectionOptions); // Collection of static assets served up from the Sandstorm server's "static" host. We only
// support relatively small assets: under 1MB each.                                                                    // 716
//                                                                                                                     // 717
// Each contains:                                                                                                      // 718
//   _id:       Random ID; will be used in the URL.                                                                    // 719
//   hash:      A base64-encoded SHA-256 hash of the data, used to de-dupe.                                            // 720
//   mimeType:  MIME type of the asset, suitable for Content-Type header.                                              // 721
//   encoding:  Either "gzip" or not present, suitable for Content-Encoding header.                                    // 722
//   content:   The asset content (byte buffer).                                                                       // 723
//   refcount:  Number of places where this asset's ID appears in the database. Since Mongo doesn't                    // 724
//       have transactions, this needs to bias towards over-counting; a backup GC could be used                        // 725
//       to catch leaked assets, although it's probably not a big deal in practice.                                    // 726
                                                                                                                       //
AssetUploadTokens = new Mongo.Collection("assetUploadTokens", collectionOptions); // Collection of tokens representing a single-use permission to upload an asset, such as a new
// profile picture.                                                                                                    // 730
//                                                                                                                     // 731
// Each contains:                                                                                                      // 732
//   _id:       Random ID.                                                                                             // 733
//   purpose:   Contains one of the following, indicating how the asset is to be used:                                 // 734
//       profilePicture: Indicates that the upload is a new profile picture. Contains fields:                          // 735
//           userId: Account ID of user whose picture shall be replaced.                                               // 736
//   expires:   Time when this token will go away if unused.                                                           // 737
                                                                                                                       //
Plans = new Mongo.Collection("plans", collectionOptions); // Subscription plans, which determine quota.                // 739
//                                                                                                                     // 741
// Each contains:                                                                                                      // 742
//   _id: Plan ID, usually a short string like "free", "standard", "large", "mega", ...                                // 743
//   storage: Number of bytes this user is allowed to store.                                                           // 744
//   compute: Number of kilobyte-RAM-seconds this user is allowed to consume.                                          // 745
//   computeLabel: Label to display to the user describing this plan's compute units.                                  // 746
//   grains: Total number of grains this user can create (often `Infinity`).                                           // 747
//   price: Price per month in US cents.                                                                               // 748
//   hidden: If true, a user cannot switch to this plan, but some users may be on it and are                           // 749
//       allowed to switch away.                                                                                       // 750
//   title: Title from display purposes. If missing, default to capitalizing _id.                                      // 751
                                                                                                                       //
AppIndex = new Mongo.Collection("appIndex", collectionOptions); // A mirror of the data from the App Market index      // 753
//                                                                                                                     // 755
// Each contains:                                                                                                      // 756
//   _id: the appId of the app                                                                                         // 757
//  The rest of the fields are defined in src/sandstorm/app-index/app-index.capnp:AppIndexForMarket                    // 758
                                                                                                                       //
KeybaseProfiles = new Mongo.Collection("keybaseProfiles", collectionOptions); // Cache of Keybase profile information. The profile for a user is re-fetched every time a package
// by that user is installed, as well as if the keybase profile is requested and not already                           // 762
// present for some reason.                                                                                            // 763
//                                                                                                                     // 764
// Each contains:                                                                                                      // 765
//   _id: PGP key fingerprint (SHA-1, hex, all-caps)                                                                   // 766
//   displayName: Display name from Keybase. (NOT VERIFIED AT ALL.)                                                    // 767
//   handle: Keybase handle.                                                                                           // 768
//   proofs: The "proofs_summary.all" array from the Keybase lookup. See the non-existent Keybase                      // 769
//     docs for details. We also add a boolean "status" field to each proof indicating whether                         // 770
//     we have directly verified the proof ourselves. Its values may be "unverified" (Keybase                          // 771
//     returned this but we haven't checked it directly), "verified" (we verified the proof and it                     // 772
//     is valid), "invalid" (we checked the proof and it was definitely bogus), or "checking" (the                     // 773
//     server is currently actively checking this proof). Note that if a check fails due to network                    // 774
//     errors, the status goes back to "unverified".                                                                   // 775
//                                                                                                                     // 776
//     WARNING: Currently verification is NOT IMPLEMENTED, so all proofs will be "unverified"                          // 777
//       for now and we just trust Keybase.                                                                            // 778
                                                                                                                       //
FeatureKey = new Mongo.Collection("featureKey", collectionOptions); // OBSOLETE: This was used to implement the Sandstorm for Work paywall, which has been removed.
//   Collection object still defined because it could have old data in it, for servers that used                       // 782
//   to have a feature key.                                                                                            // 783
                                                                                                                       //
SetupSession = new Mongo.Collection("setupSession", collectionOptions); // Responsible for storing information about setup sessions.  Contains a single document with three
// keys:                                                                                                               // 787
//                                                                                                                     // 788
//   _id: "current-session"                                                                                            // 789
//   creationDate: Date object indicating when this session was created.                                               // 790
//   hashedSessionId: the sha256 of the secret session id that was returned to the client                              // 791
                                                                                                                       //
var DesktopNotifications = new Mongo.Collection("desktopNotifications", collectionOptions); // Responsible for very short-lived queueing of desktop notification information.
// Entries are removed when they are ~30 seconds old.  This collection is a bit                                        // 795
// odd in that it is intended primarily for edge-triggered communications, but                                         // 796
// Meteor's collections aren't really designed to support that organization.                                           // 797
// Fields for each :                                                                                                   // 798
//                                                                                                                     // 799
//   _id: String.  Used as the tag to coordinate notification merging between browser tabs.                            // 800
//   creationDate: Date object. indicating when this notification was posted.                                          // 801
//   userId: String. Account id to which this notification was published.                                              // 802
//   notificationId: String.  ID of the matching event in the Notifications table to dismiss if this                   // 803
//                            notification is activated.                                                               // 804
//   appActivity: Object with fields:                                                                                  // 805
//     user: Optional Object. Not present if this notification wasn't generated by a user. If                          // 806
//           present, it will have one of the following shapes:                                                        // 807
//       { anonymous: true } if this notification was generated by an anonymous user.  Otherwise:                      // 808
//       {                                                                                                             // 809
//         accountId: String   The user's account ID.                                                                  // 810
//         name: String        The user's display name.                                                                // 811
//         avatarUrl: String   The URL for the user's profile picture.                                                 // 812
//       },                                                                                                            // 813
//     grainId: String,      Which grain this action took place on                                                     // 814
//     path: String,         The path of the notification.                                                             // 815
//     body: Util.LocalizedText,  The main body of the activity event.                                                 // 816
//     actionText: Util.LocalizedText, What action the user took, e.g.                                                 // 817
//                                     { defaultText: "added a comment" }                                              // 818
                                                                                                                       //
var StandaloneDomains = new Mongo.Collection("standaloneDomains", collectionOptions); // A standalone domain that points to a single share link. These domains act a little different
// than a normal shared Sandstorm grain. They completely drop any Sandstorm topbar/sidebar, and at                     // 822
// first glance look completely like a non-Sandstorm hosted webserver. The apps instead act in                         // 823
// concert with Sandstorm through the postMessage API, which allows it to do things like prompt for                    // 824
// login.                                                                                                              // 825
// Fields for each :                                                                                                   // 826
//                                                                                                                     // 827
//   _id: String. The domain name to use.                                                                              // 828
//   token: String. _id of a sharing token (it must be a webkey).                                                      // 829
                                                                                                                       //
if (Meteor.isServer) {                                                                                                 // 831
  Meteor.publish("credentials", function () {                                                                          // 832
    // Data needed for isSignedUp() and isAdmin() to work.                                                             // 833
    if (this.userId) {                                                                                                 // 835
      var db = this.connection.sandstormDb;                                                                            // 836
      return [Meteor.users.find({                                                                                      // 837
        _id: this.userId                                                                                               // 838
      }, {                                                                                                             // 838
        fields: {                                                                                                      // 839
          type: 1,                                                                                                     // 839
          signupKey: 1,                                                                                                // 839
          isAdmin: 1,                                                                                                  // 839
          expires: 1,                                                                                                  // 839
          storageUsage: 1,                                                                                             // 839
          plan: 1,                                                                                                     // 840
          planBonus: 1,                                                                                                // 840
          hasCompletedSignup: 1,                                                                                       // 840
          experiments: 1,                                                                                              // 840
          referredAccountIds: 1,                                                                                       // 841
          cachedStorageQuota: 1,                                                                                       // 841
          suspended: 1                                                                                                 // 841
        }                                                                                                              // 839
      }), db.collections.plans.find()];                                                                                // 839
    } else {                                                                                                           // 844
      return [];                                                                                                       // 845
    }                                                                                                                  // 846
  });                                                                                                                  // 847
}                                                                                                                      // 848
                                                                                                                       //
var countReferrals = function (user) {                                                                                 // 850
  var referredAccountIds = user.referredAccountIds;                                                                    // 851
  return referredAccountIds && referredAccountIds.length || 0;                                                         // 852
};                                                                                                                     // 853
                                                                                                                       //
var calculateReferralBonus = function (user) {                                                                         // 855
  // This function returns an object of the form:                                                                      // 856
  //                                                                                                                   // 857
  // - {grains: 0, storage: 0}                                                                                         // 858
  //                                                                                                                   // 859
  // which are extra resources this account gets as part of participating in the referral                              // 860
  // program. (Storage is measured in bytes, as usual for plans.)                                                      // 861
  // TODO(cleanup): Consider moving referral bonus logic into Oasis payments module (since it's                        // 863
  //   payments-specific) and aggregating into `planBonus`.                                                            // 864
  // Authorization note: Only call this if accountId is the current user!                                              // 866
  var isPaid = user.plan && user.plan !== "free";                                                                      // 867
  successfulReferralsCount = countReferrals(user);                                                                     // 869
                                                                                                                       //
  if (isPaid) {                                                                                                        // 870
    var maxPaidStorageBonus = 30 * 1e9;                                                                                // 871
    return {                                                                                                           // 872
      grains: 0,                                                                                                       // 872
      storage: Math.min(successfulReferralsCount * 2 * 1e9, maxPaidStorageBonus)                                       // 873
    };                                                                                                                 // 872
  } else {                                                                                                             // 876
    var maxFreeStorageBonus = 2 * 1e9;                                                                                 // 877
    var bonus = {                                                                                                      // 878
      storage: Math.min(successfulReferralsCount * 50 * 1e6, maxFreeStorageBonus)                                      // 879
    };                                                                                                                 // 878
                                                                                                                       //
    if (successfulReferralsCount > 0) {                                                                                // 883
      bonus.grains = Infinity;                                                                                         // 884
    } else {                                                                                                           // 885
      bonus.grains = 0;                                                                                                // 886
    }                                                                                                                  // 887
                                                                                                                       //
    return bonus;                                                                                                      // 889
  }                                                                                                                    // 890
};                                                                                                                     // 891
                                                                                                                       //
isAdmin = function () {                                                                                                // 893
  // Returns true if the user is the administrator.                                                                    // 894
  var user = Meteor.user();                                                                                            // 896
                                                                                                                       //
  if (user && user.isAdmin) {                                                                                          // 897
    return true;                                                                                                       // 898
  } else {                                                                                                             // 899
    return false;                                                                                                      // 900
  }                                                                                                                    // 901
};                                                                                                                     // 902
                                                                                                                       //
isAdminById = function (id) {                                                                                          // 904
  // Returns true if the user's id is the administrator.                                                               // 905
  var user = Meteor.users.findOne({                                                                                    // 907
    _id: id                                                                                                            // 907
  }, {                                                                                                                 // 907
    fields: {                                                                                                          // 907
      isAdmin: 1                                                                                                       // 907
    }                                                                                                                  // 907
  });                                                                                                                  // 907
                                                                                                                       //
  if (user && user.isAdmin) {                                                                                          // 908
    return true;                                                                                                       // 909
  } else {                                                                                                             // 910
    return false;                                                                                                      // 911
  }                                                                                                                    // 912
};                                                                                                                     // 913
                                                                                                                       //
findAdminUserForToken = function (token) {                                                                             // 915
  if (!token.requirements) {                                                                                           // 916
    return;                                                                                                            // 917
  }                                                                                                                    // 918
                                                                                                                       //
  var requirements = token.requirements.filter(function (requirement) {                                                // 920
    return "userIsAdmin" in requirement;                                                                               // 921
  });                                                                                                                  // 922
                                                                                                                       //
  if (requirements.length > 1) {                                                                                       // 924
    return;                                                                                                            // 925
  }                                                                                                                    // 926
                                                                                                                       //
  if (requirements.length === 0) {                                                                                     // 928
    return;                                                                                                            // 929
  }                                                                                                                    // 930
                                                                                                                       //
  return requirements[0].userIsAdmin;                                                                                  // 932
};                                                                                                                     // 933
                                                                                                                       //
var wildcardHost = Meteor.settings.public.wildcardHost.toLowerCase().split("*");                                       // 935
                                                                                                                       //
if (wildcardHost.length != 2) {                                                                                        // 937
  throw new Error("Wildcard host must contain exactly one asterisk.");                                                 // 938
}                                                                                                                      // 939
                                                                                                                       //
matchWildcardHost = function (host) {                                                                                  // 941
  // See if the hostname is a member of our wildcard. If so, extract the ID.                                           // 942
  // We remove everything after the first ":" character so that our                                                    // 944
  // comparison logic ignores port numbers.                                                                            // 945
  var prefix = wildcardHost[0];                                                                                        // 946
  var suffix = wildcardHost[1].split(":")[0];                                                                          // 947
  var hostSansPort = host.split(":")[0];                                                                               // 948
                                                                                                                       //
  if (hostSansPort.lastIndexOf(prefix, 0) >= 0 && hostSansPort.indexOf(suffix, -suffix.length) >= 0 && hostSansPort.length >= prefix.length + suffix.length) {
    var id = hostSansPort.slice(prefix.length, -suffix.length);                                                        // 953
                                                                                                                       //
    if (id.match(/^[-a-z0-9]*$/)) {                                                                                    // 954
      return id;                                                                                                       // 955
    }                                                                                                                  // 956
  }                                                                                                                    // 957
                                                                                                                       //
  return null;                                                                                                         // 959
};                                                                                                                     // 960
                                                                                                                       //
makeWildcardHost = function (id) {                                                                                     // 962
  return wildcardHost[0] + id + wildcardHost[1];                                                                       // 963
};                                                                                                                     // 964
                                                                                                                       //
var isApiHostId = function (hostId) {                                                                                  // 966
  if (hostId) {                                                                                                        // 967
    var split = hostId.split("-");                                                                                     // 968
    if (split[0] === "api") return split[1] || "*";                                                                    // 969
  }                                                                                                                    // 970
                                                                                                                       //
  return false;                                                                                                        // 972
};                                                                                                                     // 973
                                                                                                                       //
var isTokenSpecificHostId = function (hostId) {                                                                        // 975
  return hostId.lastIndexOf("api-", 0) === 0;                                                                          // 976
};                                                                                                                     // 977
                                                                                                                       //
var apiHostIdHashForToken = void 0;                                                                                    // 979
                                                                                                                       //
if (Meteor.isServer) {                                                                                                 // 980
  var Crypto = Npm.require("crypto");                                                                                  // 981
                                                                                                                       //
  apiHostIdHashForToken = function (token) {                                                                           // 982
    // Given an API token, compute the host ID that must be used when requesting this token.                           // 983
    // We add a leading 'x' to the hash so that knowing the hostname alone is not sufficient to                        // 985
    // find the corresponding API token in the ApiTokens table (whose _id values are also hashes                       // 986
    // of tokens). This doesn't technically add any security, but helps prove that we don't have                       // 987
    // any bugs which would allow someone who knows only the hostname to access the app API.                           // 988
    return Crypto.createHash("sha256").update("x" + token).digest("hex").slice(0, 32);                                 // 989
  };                                                                                                                   // 990
} else {                                                                                                               // 991
  apiHostIdHashForToken = function (token) {                                                                           // 992
    // Given an API token, compute the host ID that must be used when requesting this token.                           // 993
    // We add a leading 'x' to the hash so that knowing the hostname alone is not sufficient to                        // 995
    // find the corresponding API token in the ApiTokens table (whose _id values are also hashes                       // 996
    // of tokens). This doesn't technically add any security, but helps prove that we don't have                       // 997
    // any bugs which would allow someone who knows only the hostname to access the app API.                           // 998
    return SHA256("x" + token).slice(0, 32);                                                                           // 999
  };                                                                                                                   // 1000
}                                                                                                                      // 1001
                                                                                                                       //
var apiHostIdForToken = function (token) {                                                                             // 1003
  return "api-" + apiHostIdHashForToken(token);                                                                        // 1004
};                                                                                                                     // 1005
                                                                                                                       //
var makeApiHost = function (token) {                                                                                   // 1007
  return makeWildcardHost(apiHostIdForToken(token));                                                                   // 1008
};                                                                                                                     // 1009
                                                                                                                       //
if (Meteor.isServer) {                                                                                                 // 1011
  var Url = Npm.require("url");                                                                                        // 1012
                                                                                                                       //
  getWildcardOrigin = function () {                                                                                    // 1013
    // The wildcard URL can be something like "foo-*-bar.example.com", but sometimes when we're                        // 1014
    // trying to specify a pattern matching hostnames (say, a Content-Security-Policy directive),                      // 1015
    // an astrisk is only allowed as the first character and must be followed by a period. So we need                  // 1016
    // "*.example.com" instead -- which matches more than we actually want, but is the best we can                     // 1017
    // really do. We also add the protocol to the front (again, that's what CSP wants).                                // 1018
    // TODO(cleanup): `protocol` is computed in other files, like proxy.js. Put it somewhere common.                   // 1020
    var protocol = Url.parse(process.env.ROOT_URL).protocol;                                                           // 1021
    var dotPos = wildcardHost[1].indexOf(".");                                                                         // 1023
                                                                                                                       //
    if (dotPos < 0) {                                                                                                  // 1024
      return protocol + "//*";                                                                                         // 1025
    } else {                                                                                                           // 1026
      return protocol + "//*" + wildcardHost[1].slice(dotPos);                                                         // 1027
    }                                                                                                                  // 1028
  };                                                                                                                   // 1029
}                                                                                                                      // 1030
                                                                                                                       //
SandstormDb = function (quotaManager) {                                                                                // 1032
  // quotaManager is an object with the following method:                                                              // 1033
  //   updateUserQuota: It is provided two arguments                                                                   // 1034
  //     db: This SandstormDb object                                                                                   // 1035
  //     user: A collections.users account object                                                                      // 1036
  //   and returns a quota object:                                                                                     // 1037
  //     storage: A number (can be Infinity)                                                                           // 1038
  //     compute: A number (can be Infinity)                                                                           // 1039
  //     grains: A number (can be Infinity)                                                                            // 1040
  this.quotaManager = quotaManager;                                                                                    // 1042
  this.collections = {                                                                                                 // 1043
    // Direct access to underlying collections. DEPRECATED, but better than accessing the top-level                    // 1044
    // collection globals directly.                                                                                    // 1045
    //                                                                                                                 // 1046
    // TODO(cleanup): Over time, we will provide methods covering each supported query and remove                      // 1047
    //   direct access to the collections.                                                                             // 1048
    users: Meteor.users,                                                                                               // 1049
    packages: Packages,                                                                                                // 1051
    devPackages: DevPackages,                                                                                          // 1052
    userActions: UserActions,                                                                                          // 1053
    grains: Grains,                                                                                                    // 1054
    roleAssignments: RoleAssignments,                                                                                  // 1055
    // Deprecated, only used by the migration that eliminated it.                                                      // 1055
    contacts: Contacts,                                                                                                // 1056
    sessions: Sessions,                                                                                                // 1057
    signupKeys: SignupKeys,                                                                                            // 1058
    activityStats: ActivityStats,                                                                                      // 1059
    deleteStats: DeleteStats,                                                                                          // 1060
    fileTokens: FileTokens,                                                                                            // 1061
    spkTokens: SpkTokens,                                                                                              // 1062
    apiTokens: ApiTokens,                                                                                              // 1063
    apiHosts: ApiHosts,                                                                                                // 1064
    notifications: Notifications,                                                                                      // 1065
    activitySubscriptions: ActivitySubscriptions,                                                                      // 1066
    statsTokens: StatsTokens,                                                                                          // 1067
    misc: Misc,                                                                                                        // 1068
    settings: Settings,                                                                                                // 1069
    migrations: Migrations,                                                                                            // 1070
    staticAssets: StaticAssets,                                                                                        // 1071
    assetUploadTokens: AssetUploadTokens,                                                                              // 1072
    plans: Plans,                                                                                                      // 1073
    appIndex: AppIndex,                                                                                                // 1074
    keybaseProfiles: KeybaseProfiles,                                                                                  // 1075
    setupSession: SetupSession,                                                                                        // 1076
    desktopNotifications: DesktopNotifications,                                                                        // 1077
    standaloneDomains: StandaloneDomains                                                                               // 1078
  };                                                                                                                   // 1043
}; // TODO(cleanup): These methods should not be defined freestanding and should use collection                        // 1080
//   objects created in SandstormDb's constructor rather than globals.                                                 // 1083
                                                                                                                       //
                                                                                                                       //
_.extend(SandstormDb.prototype, {                                                                                      // 1085
  isAdmin: isAdmin,                                                                                                    // 1086
  isAdminById: isAdminById,                                                                                            // 1087
  findAdminUserForToken: findAdminUserForToken,                                                                        // 1088
  matchWildcardHost: matchWildcardHost,                                                                                // 1089
  makeWildcardHost: makeWildcardHost,                                                                                  // 1090
  isApiHostId: isApiHostId,                                                                                            // 1091
  isTokenSpecificHostId: isTokenSpecificHostId,                                                                        // 1092
  apiHostIdHashForToken: apiHostIdHashForToken,                                                                        // 1093
  apiHostIdForToken: apiHostIdForToken,                                                                                // 1094
  makeApiHost: makeApiHost,                                                                                            // 1095
  allowDevAccounts: function () {                                                                                      // 1096
    var setting = this.collections.settings.findOne({                                                                  // 1097
      _id: "devAccounts"                                                                                               // 1097
    });                                                                                                                // 1097
                                                                                                                       //
    if (setting) {                                                                                                     // 1098
      return setting.value;                                                                                            // 1099
    } else {                                                                                                           // 1100
      return Meteor.settings && Meteor.settings.public && Meteor.settings.public.allowDevAccounts;                     // 1101
    }                                                                                                                  // 1103
  },                                                                                                                   // 1104
  roleAssignmentPattern: {                                                                                             // 1106
    none: Match.Optional(null),                                                                                        // 1107
    allAccess: Match.Optional(null),                                                                                   // 1108
    roleId: Match.Optional(Match.Integer),                                                                             // 1109
    addPermissions: Match.Optional([Boolean]),                                                                         // 1110
    removePermissions: Match.Optional([Boolean])                                                                       // 1111
  },                                                                                                                   // 1106
  isDemoUser: function () {                                                                                            // 1114
    // Returns true if this is a demo user.                                                                            // 1115
    var user = Meteor.user();                                                                                          // 1117
                                                                                                                       //
    if (user && user.expires) {                                                                                        // 1118
      return true;                                                                                                     // 1119
    } else {                                                                                                           // 1120
      return false;                                                                                                    // 1121
    }                                                                                                                  // 1122
  },                                                                                                                   // 1123
  isSignedUp: function () {                                                                                            // 1125
    var user = Meteor.user();                                                                                          // 1126
    return this.isAccountSignedUp(user);                                                                               // 1127
  },                                                                                                                   // 1128
  isAccountSignedUp: function (user) {                                                                                 // 1130
    // Returns true if the user has presented an invite key.                                                           // 1131
    if (!user) return false; // not signed in                                                                          // 1133
                                                                                                                       //
    if (user.type != "account") return false; // not an account                                                        // 1135
                                                                                                                       //
    if (user.expires) return false; // demo user.                                                                      // 1137
                                                                                                                       //
    if (Meteor.settings.public.allowUninvited) return true; // all accounts qualify                                    // 1139
                                                                                                                       //
    if (user.signupKey) return true; // user is invited                                                                // 1141
                                                                                                                       //
    if (this.isUserInOrganization(user)) return true;                                                                  // 1143
    return false;                                                                                                      // 1145
  },                                                                                                                   // 1146
  isSignedUpOrDemo: function () {                                                                                      // 1148
    var user = Meteor.user();                                                                                          // 1149
    return this.isAccountSignedUpOrDemo(user);                                                                         // 1150
  },                                                                                                                   // 1151
  isAccountSignedUpOrDemo: function (user) {                                                                           // 1153
    if (!user) return false; // not signed in                                                                          // 1154
                                                                                                                       //
    if (user.type != "account") return false; // not an account                                                        // 1156
                                                                                                                       //
    if (user.expires) return true; // demo user.                                                                       // 1158
                                                                                                                       //
    if (Meteor.settings.public.allowUninvited) return true; // all accounts qualify                                    // 1160
                                                                                                                       //
    if (user.signupKey) return true; // user is invited                                                                // 1162
                                                                                                                       //
    if (this.isUserInOrganization(user)) return true;                                                                  // 1164
    return false;                                                                                                      // 1166
  },                                                                                                                   // 1167
  isCredentialInOrganization: function (credential) {                                                                  // 1169
    if (!credential || !credential.services) {                                                                         // 1170
      return false;                                                                                                    // 1171
    }                                                                                                                  // 1172
                                                                                                                       //
    var orgMembership = this.getOrganizationMembership();                                                              // 1174
    var googleEnabled = orgMembership && orgMembership.google && orgMembership.google.enabled;                         // 1175
    var googleDomain = orgMembership && orgMembership.google && orgMembership.google.domain;                           // 1176
    var emailEnabled = orgMembership && orgMembership.emailToken && orgMembership.emailToken.enabled;                  // 1177
    var emailDomain = orgMembership && orgMembership.emailToken && orgMembership.emailToken.domain;                    // 1178
    var ldapEnabled = orgMembership && orgMembership.ldap && orgMembership.ldap.enabled;                               // 1179
    var samlEnabled = orgMembership && orgMembership.saml && orgMembership.saml.enabled;                               // 1180
                                                                                                                       //
    if (emailEnabled && emailDomain && credential.services.email) {                                                    // 1181
      var domainSuffixes = emailDomain.split(/\s*,\s*/);                                                               // 1182
                                                                                                                       //
      for (var i = 0; i < domainSuffixes.length; i++) {                                                                // 1183
        var suffix = domainSuffixes[i];                                                                                // 1184
        var domain = credential.services.email.email.toLowerCase().split("@").pop();                                   // 1185
                                                                                                                       //
        if (suffix.startsWith("*.")) {                                                                                 // 1186
          if (domain.endsWith(suffix.substr(1))) {                                                                     // 1187
            return true;                                                                                               // 1188
          }                                                                                                            // 1189
        } else if (domain === suffix) {                                                                                // 1190
          return true;                                                                                                 // 1191
        }                                                                                                              // 1192
      }                                                                                                                // 1193
    } else if (ldapEnabled && credential.services.ldap) {                                                              // 1194
      return true;                                                                                                     // 1195
    } else if (samlEnabled && credential.services.saml) {                                                              // 1196
      return true;                                                                                                     // 1197
    } else if (googleEnabled && googleDomain && credential.services.google && credential.services.google.hd) {         // 1198
      if (credential.services.google.hd.toLowerCase() === googleDomain) {                                              // 1199
        return true;                                                                                                   // 1200
      }                                                                                                                // 1201
    }                                                                                                                  // 1202
                                                                                                                       //
    return false;                                                                                                      // 1204
  },                                                                                                                   // 1205
  isUserInOrganization: function (user) {                                                                              // 1207
    for (var i = 0; i < user.loginCredentials.length; i++) {                                                           // 1208
      var credential = Meteor.users.findOne({                                                                          // 1209
        _id: user.loginCredentials[i].id                                                                               // 1209
      });                                                                                                              // 1209
                                                                                                                       //
      if (this.isCredentialInOrganization(credential)) {                                                               // 1210
        return true;                                                                                                   // 1211
      }                                                                                                                // 1212
    }                                                                                                                  // 1213
                                                                                                                       //
    return false;                                                                                                      // 1215
  }                                                                                                                    // 1216
});                                                                                                                    // 1085
                                                                                                                       //
if (Meteor.isServer) {                                                                                                 // 1219
  SandstormDb.prototype.getWildcardOrigin = getWildcardOrigin;                                                         // 1220
                                                                                                                       //
  var _Crypto = Npm.require("crypto");                                                                                 // 1222
                                                                                                                       //
  SandstormDb.prototype.removeApiTokens = function (query) {                                                           // 1223
    var _this2 = this;                                                                                                 // 1223
                                                                                                                       //
    // Remove all API tokens matching the query, making sure to clean up ApiHosts as well.                             // 1224
    this.collections.apiTokens.find(query).forEach(function (token) {                                                  // 1226
      // Clean up ApiHosts for webkey tokens.                                                                          // 1227
      if (token.hasApiHost) {                                                                                          // 1228
        var hash2 = _Crypto.createHash("sha256").update(token._id).digest("base64");                                   // 1229
                                                                                                                       //
        _this2.collections.apiHosts.remove({                                                                           // 1230
          hash2: hash2                                                                                                 // 1230
        });                                                                                                            // 1230
      } // TODO(soon): Drop remote OAuth tokens for frontendRef.http. Unfortunately the way to do                      // 1231
      //   this is different for every service. :( Also we may need to clarify with the "bearer"                       // 1234
      //   type whether or not the token is "owned" by us...                                                           // 1235
                                                                                                                       //
    });                                                                                                                // 1236
    this.collections.apiTokens.remove(query);                                                                          // 1238
  };                                                                                                                   // 1239
} // TODO(someday): clean this up.  Logic for building static asset urls on client and server                          // 1240
// appears all over the codebase.                                                                                      // 1243
                                                                                                                       //
                                                                                                                       //
var httpProtocol = void 0;                                                                                             // 1244
                                                                                                                       //
if (Meteor.isServer) {                                                                                                 // 1245
  var _Url = Npm.require("url");                                                                                       // 1246
                                                                                                                       //
  httpProtocol = _Url.parse(process.env.ROOT_URL).protocol;                                                            // 1247
} else {                                                                                                               // 1248
  httpProtocol = window.location.protocol;                                                                             // 1249
} // =======================================================================================                           // 1250
// Below this point are newly-written or refactored functions.                                                         // 1253
                                                                                                                       //
                                                                                                                       //
_.extend(SandstormDb.prototype, {                                                                                      // 1255
  getUser: function (userId) {                                                                                         // 1256
    check(userId, Match.OneOf(String, undefined, null));                                                               // 1257
                                                                                                                       //
    if (userId) {                                                                                                      // 1258
      return Meteor.users.findOne(userId);                                                                             // 1259
    }                                                                                                                  // 1260
  },                                                                                                                   // 1261
  getCredential: function (credentialId) {                                                                             // 1263
    check(credentialId, String);                                                                                       // 1264
    return Meteor.users.findOne({                                                                                      // 1265
      _id: credentialId                                                                                                // 1265
    });                                                                                                                // 1265
  },                                                                                                                   // 1266
  userHasCredential: function (userId, credentialId) {                                                                 // 1268
    check(userId, String);                                                                                             // 1269
    check(credentialId, String);                                                                                       // 1270
    if (userId === credentialId) return true;                                                                          // 1272
    var user = Meteor.users.findOne(userId);                                                                           // 1274
    return SandstormDb.getUserCredentialIds(user).indexOf(credentialId) != -1;                                         // 1275
  },                                                                                                                   // 1276
  userGrains: function (userId, options) {                                                                             // 1278
    check(userId, Match.OneOf(String, undefined, null));                                                               // 1279
    check(options, Match.OneOf(undefined, null, {                                                                      // 1280
      includeTrashOnly: Match.Optional(Boolean),                                                                       // 1281
      includeTrash: Match.Optional(Boolean)                                                                            // 1281
    }));                                                                                                               // 1281
    var query = {                                                                                                      // 1283
      userId: userId                                                                                                   // 1283
    };                                                                                                                 // 1283
                                                                                                                       //
    if (options && options.includeTrashOnly) {                                                                         // 1284
      query.trashed = {                                                                                                // 1285
        $exists: true                                                                                                  // 1285
      };                                                                                                               // 1285
    } else if (options && options.includeTrash) {// Keep query as-is.                                                  // 1286
    } else {                                                                                                           // 1288
      query.trashed = {                                                                                                // 1289
        $exists: false                                                                                                 // 1289
      };                                                                                                               // 1289
    }                                                                                                                  // 1290
                                                                                                                       //
    return this.collections.grains.find(query);                                                                        // 1292
  },                                                                                                                   // 1293
  currentUserGrains: function (options) {                                                                              // 1295
    return this.userGrains(Meteor.userId(), options);                                                                  // 1296
  },                                                                                                                   // 1297
  getGrain: function (grainId) {                                                                                       // 1299
    check(grainId, String);                                                                                            // 1300
    return this.collections.grains.findOne(grainId);                                                                   // 1301
  },                                                                                                                   // 1302
  userApiTokens: function (userId, trashed) {                                                                          // 1304
    check(userId, Match.OneOf(String, undefined, null));                                                               // 1305
    check(trashed, Match.OneOf(Boolean, undefined, null));                                                             // 1306
    return this.collections.apiTokens.find({                                                                           // 1307
      "owner.user.accountId": userId || "invalid user don't match anything please!",                                   // 1308
      trashed: {                                                                                                       // 1309
        $exists: !!trashed                                                                                             // 1309
      }                                                                                                                // 1309
    });                                                                                                                // 1307
  },                                                                                                                   // 1311
  currentUserApiTokens: function (trashed) {                                                                           // 1313
    return this.userApiTokens(Meteor.userId(), trashed);                                                               // 1314
  },                                                                                                                   // 1315
  userActions: function (user) {                                                                                       // 1317
    return this.collections.userActions.find({                                                                         // 1318
      userId: user                                                                                                     // 1318
    });                                                                                                                // 1318
  },                                                                                                                   // 1319
  currentUserActions: function () {                                                                                    // 1321
    return this.userActions(Meteor.userId());                                                                          // 1322
  },                                                                                                                   // 1323
  iconSrcForPackage: function (pkg, usage) {                                                                           // 1325
    return Identicon.iconSrcForPackage(pkg, usage, httpProtocol + "//" + this.makeWildcardHost("static"));             // 1326
  },                                                                                                                   // 1327
  getDenormalizedGrainInfo: function (grainId) {                                                                       // 1329
    var grain = this.getGrain(grainId);                                                                                // 1330
    var pkg = this.collections.packages.findOne(grain.packageId);                                                      // 1331
                                                                                                                       //
    if (!pkg) {                                                                                                        // 1333
      pkg = this.collections.devPackages.findOne(grain.packageId);                                                     // 1334
    }                                                                                                                  // 1335
                                                                                                                       //
    var appTitle = pkg && pkg.manifest && pkg.manifest.appTitle || {                                                   // 1337
      defaultText: ""                                                                                                  // 1337
    };                                                                                                                 // 1337
    var grainInfo = {                                                                                                  // 1338
      appTitle: appTitle                                                                                               // 1338
    };                                                                                                                 // 1338
                                                                                                                       //
    if (pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.icons) {                                 // 1340
      var icons = pkg.manifest.metadata.icons;                                                                         // 1341
      var icon = icons.grain || icons.appGrid;                                                                         // 1342
                                                                                                                       //
      if (icon) {                                                                                                      // 1343
        grainInfo.icon = icon;                                                                                         // 1344
      }                                                                                                                // 1345
    } // Only provide an app ID if we have no icon asset to provide and need to offer an identicon.                    // 1346
                                                                                                                       //
                                                                                                                       //
    if (!grainInfo.icon && pkg) {                                                                                      // 1349
      grainInfo.appId = pkg.appId;                                                                                     // 1350
    }                                                                                                                  // 1351
                                                                                                                       //
    return grainInfo;                                                                                                  // 1353
  },                                                                                                                   // 1354
  getPlan: function (id, user) {                                                                                       // 1356
    check(id, String); // `user`, if provided, is the user observing the plan. This matters only for checking if the   // 1357
    // user is in an experiment.                                                                                       // 1360
                                                                                                                       //
    var plan = this.collections.plans.findOne(id);                                                                     // 1362
                                                                                                                       //
    if (!plan) {                                                                                                       // 1363
      throw new Error("no such plan: " + id);                                                                          // 1364
    }                                                                                                                  // 1365
                                                                                                                       //
    if (plan._id === "free") {                                                                                         // 1367
      user = user || Meteor.user();                                                                                    // 1368
                                                                                                                       //
      if (user && user.experiments && typeof user.experiments.freeGrainLimit === "number") {                           // 1369
        plan.grains = user.experiments.freeGrainLimit;                                                                 // 1371
      }                                                                                                                // 1372
    }                                                                                                                  // 1373
                                                                                                                       //
    return plan;                                                                                                       // 1375
  },                                                                                                                   // 1376
  listPlans: function (user) {                                                                                         // 1378
    user = user || Meteor.user();                                                                                      // 1379
                                                                                                                       //
    if (user && user.experiments && typeof user.experiments.freeGrainLimit === "number") {                             // 1380
      return this.collections.plans.find({}, {                                                                         // 1382
        sort: {                                                                                                        // 1382
          price: 1                                                                                                     // 1382
        }                                                                                                              // 1382
      }).map(function (plan) {                                                                                         // 1382
        if (plan._id === "free") {                                                                                     // 1384
          plan.grains = user.experiments.freeGrainLimit;                                                               // 1385
        }                                                                                                              // 1386
                                                                                                                       //
        return plan;                                                                                                   // 1388
      });                                                                                                              // 1389
    } else {                                                                                                           // 1390
      return this.collections.plans.find({}, {                                                                         // 1391
        sort: {                                                                                                        // 1391
          price: 1                                                                                                     // 1391
        }                                                                                                              // 1391
      }).fetch();                                                                                                      // 1391
    }                                                                                                                  // 1392
  },                                                                                                                   // 1393
  getMyPlan: function () {                                                                                             // 1395
    var user = Meteor.user();                                                                                          // 1396
    return user && this.collections.plans.findOne(user.plan || "free");                                                // 1397
  },                                                                                                                   // 1398
  getMyReferralBonus: function (user) {                                                                                // 1400
    // This function is called from the server and from the client, similar to getMyPlan().                            // 1401
    //                                                                                                                 // 1402
    // The parameter may be omitted in which case the current user is assumed.                                         // 1403
    return calculateReferralBonus(user || Meteor.user());                                                              // 1405
  },                                                                                                                   // 1406
  getMyUsage: function (user) {                                                                                        // 1408
    user = user || Meteor.user();                                                                                      // 1409
                                                                                                                       //
    if (user && (Meteor.isServer || user.pseudoUsage)) {                                                               // 1410
      if (Meteor.isClient) {                                                                                           // 1411
        // Filled by pseudo-subscription to "getMyUsage". WARNING: The subscription is currently                       // 1412
        // not reactive.                                                                                               // 1413
        return user.pseudoUsage;                                                                                       // 1414
      } else {                                                                                                         // 1415
        return {                                                                                                       // 1416
          grains: this.collections.grains.find({                                                                       // 1417
            userId: user._id                                                                                           // 1417
          }).count(),                                                                                                  // 1417
          storage: user.storageUsage || 0,                                                                             // 1418
          compute: 0 // not tracked yet                                                                                // 1419
                                                                                                                       //
        };                                                                                                             // 1416
      }                                                                                                                // 1421
    } else {                                                                                                           // 1422
      return {                                                                                                         // 1423
        grains: 0,                                                                                                     // 1423
        storage: 0,                                                                                                    // 1423
        compute: 0                                                                                                     // 1423
      };                                                                                                               // 1423
    }                                                                                                                  // 1424
  },                                                                                                                   // 1425
  isUninvitedFreeUser: function () {                                                                                   // 1427
    if (!Meteor.settings.public.allowUninvited) return false;                                                          // 1428
    var user = Meteor.user();                                                                                          // 1430
    return user && !user.expires && (!user.plan || user.plan === "free");                                              // 1431
  },                                                                                                                   // 1432
  getSetting: function (name) {                                                                                        // 1434
    var setting = this.collections.settings.findOne(name);                                                             // 1435
    return setting && setting.value;                                                                                   // 1436
  },                                                                                                                   // 1437
  getSettingWithFallback: function (name, fallbackValue) {                                                             // 1439
    var value = this.getSetting(name);                                                                                 // 1440
                                                                                                                       //
    if (value === undefined) {                                                                                         // 1441
      return fallbackValue;                                                                                            // 1442
    }                                                                                                                  // 1443
                                                                                                                       //
    return value;                                                                                                      // 1445
  },                                                                                                                   // 1446
  addUserActions: function (userId, packageId, simulation) {                                                           // 1448
    check(userId, String);                                                                                             // 1449
    check(packageId, String);                                                                                          // 1450
    var pack = this.collections.packages.findOne({                                                                     // 1452
      _id: packageId                                                                                                   // 1452
    });                                                                                                                // 1452
                                                                                                                       //
    if (pack) {                                                                                                        // 1453
      // Remove old versions.                                                                                          // 1454
      var numRemoved = this.collections.userActions.remove({                                                           // 1455
        userId: userId,                                                                                                // 1455
        appId: pack.appId                                                                                              // 1455
      }); // Install new.                                                                                              // 1455
                                                                                                                       //
      var actions = pack.manifest.actions;                                                                             // 1458
                                                                                                                       //
      for (var i in meteorBabelHelpers.sanitizeForInObject(actions)) {                                                 // 1459
        var action = actions[i];                                                                                       // 1460
                                                                                                                       //
        if ("none" in action.input) {                                                                                  // 1461
          var userAction = {                                                                                           // 1462
            userId: userId,                                                                                            // 1463
            packageId: pack._id,                                                                                       // 1464
            appId: pack.appId,                                                                                         // 1465
            appTitle: pack.manifest.appTitle,                                                                          // 1466
            appMarketingVersion: pack.manifest.appMarketingVersion,                                                    // 1467
            appVersion: pack.manifest.appVersion,                                                                      // 1468
            title: action.title,                                                                                       // 1469
            nounPhrase: action.nounPhrase,                                                                             // 1470
            command: action.command                                                                                    // 1471
          };                                                                                                           // 1462
          this.collections.userActions.insert(userAction);                                                             // 1473
        } else {// TODO(someday):  Implement actions with capability inputs.                                           // 1474
        }                                                                                                              // 1476
      }                                                                                                                // 1477
                                                                                                                       //
      if (numRemoved > 0 && !simulation) {                                                                             // 1479
        this.deleteUnusedPackages(pack.appId);                                                                         // 1480
      }                                                                                                                // 1481
    }                                                                                                                  // 1482
  },                                                                                                                   // 1483
  sendAdminNotification: function (type, action) {                                                                     // 1485
    Meteor.users.find({                                                                                                // 1486
      isAdmin: true                                                                                                    // 1486
    }, {                                                                                                               // 1486
      fields: {                                                                                                        // 1486
        _id: 1                                                                                                         // 1486
      }                                                                                                                // 1486
    }).forEach(function (user) {                                                                                       // 1486
      Notifications.insert({                                                                                           // 1487
        admin: {                                                                                                       // 1488
          action: action,                                                                                              // 1488
          type: type                                                                                                   // 1488
        },                                                                                                             // 1488
        userId: user._id,                                                                                              // 1489
        timestamp: new Date(),                                                                                         // 1490
        isUnread: true                                                                                                 // 1491
      });                                                                                                              // 1487
    });                                                                                                                // 1493
  },                                                                                                                   // 1494
  getKeybaseProfile: function (keyFingerprint) {                                                                       // 1496
    return this.collections.keybaseProfiles.findOne(keyFingerprint) || {};                                             // 1497
  },                                                                                                                   // 1498
  getServerTitle: function () {                                                                                        // 1500
    var setting = this.collections.settings.findOne({                                                                  // 1501
      _id: "serverTitle"                                                                                               // 1501
    });                                                                                                                // 1501
    return setting ? setting.value : ""; // empty if subscription is not ready.                                        // 1502
  },                                                                                                                   // 1503
  getSmtpConfig: function () {                                                                                         // 1505
    var setting = this.collections.settings.findOne({                                                                  // 1506
      _id: "smtpConfig"                                                                                                // 1506
    });                                                                                                                // 1506
    return setting ? setting.value : undefined; // undefined if subscription is not ready.                             // 1507
  },                                                                                                                   // 1508
  getReturnAddress: function () {                                                                                      // 1510
    var config = this.getSmtpConfig();                                                                                 // 1511
    return config && config.returnAddress || ""; // empty if subscription is not ready.                                // 1512
  },                                                                                                                   // 1513
  getReturnAddressWithDisplayName: function (userId) {                                                                 // 1515
    check(userId, String);                                                                                             // 1516
    var user = Meteor.users.findOne(userId);                                                                           // 1517
    var displayName = user.profile.name + " (via " + this.getServerTitle() + ")"; // First remove any instances of characters that cause trouble for SimpleSmtp. Ideally,
    // we could escape such characters with a backslash, but that does not seem to help here.                          // 1521
    // TODO(cleanup): Unclear whether this sanitization is still necessary now that we return a                        // 1522
    //   structured object and have moved to nodemailer. I'm not touching it for now.                                  // 1523
                                                                                                                       //
    var sanitized = displayName.replace(/"|<|>|\\|\r/g, "");                                                           // 1524
    return {                                                                                                           // 1526
      name: sanitized,                                                                                                 // 1526
      address: this.getReturnAddress()                                                                                 // 1526
    };                                                                                                                 // 1526
  },                                                                                                                   // 1527
  getPrimaryEmail: function (accountId) {                                                                              // 1529
    check(accountId, String);                                                                                          // 1530
    var result = null;                                                                                                 // 1532
    SandstormDb.getUserEmails(Meteor.users.findOne(accountId)).forEach(function (email) {                              // 1533
      if (email.primary) result = email.email;                                                                         // 1534
    });                                                                                                                // 1535
    return result;                                                                                                     // 1537
  },                                                                                                                   // 1538
  incrementDailySentMailCount: function (accountId) {                                                                  // 1540
    check(accountId, String);                                                                                          // 1541
    var DAILY_LIMIT = 50;                                                                                              // 1543
    var result = Meteor.users.findAndModify({                                                                          // 1544
      query: {                                                                                                         // 1545
        _id: accountId                                                                                                 // 1545
      },                                                                                                               // 1545
      update: {                                                                                                        // 1546
        $inc: {                                                                                                        // 1547
          dailySentMailCount: 1                                                                                        // 1548
        }                                                                                                              // 1547
      },                                                                                                               // 1546
      fields: {                                                                                                        // 1551
        dailySentMailCount: 1                                                                                          // 1551
      }                                                                                                                // 1551
    });                                                                                                                // 1544
                                                                                                                       //
    if (!result.ok) {                                                                                                  // 1554
      throw new Error("Couldn't update daily sent mail count.");                                                       // 1555
    }                                                                                                                  // 1556
                                                                                                                       //
    var user = result.value;                                                                                           // 1558
                                                                                                                       //
    if (user.dailySentMailCount >= DAILY_LIMIT) {                                                                      // 1559
      throw new Error("Sorry, you've reached your e-mail sending limit for today. Currently, Sandstorm " + "limits each user to " + DAILY_LIMIT + " e-mails per day for spam control reasons. " + "Please feel free to contact us if this is a problem.");
    }                                                                                                                  // 1564
  },                                                                                                                   // 1565
  getLdapUrl: function () {                                                                                            // 1567
    var setting = this.collections.settings.findOne({                                                                  // 1568
      _id: "ldapUrl"                                                                                                   // 1568
    });                                                                                                                // 1568
    return setting ? setting.value : ""; // empty if subscription is not ready.                                        // 1569
  },                                                                                                                   // 1570
  getLdapBase: function () {                                                                                           // 1572
    var setting = this.collections.settings.findOne({                                                                  // 1573
      _id: "ldapBase"                                                                                                  // 1573
    });                                                                                                                // 1573
    return setting ? setting.value : ""; // empty if subscription is not ready.                                        // 1574
  },                                                                                                                   // 1575
  getLdapDnPattern: function () {                                                                                      // 1577
    var setting = this.collections.settings.findOne({                                                                  // 1578
      _id: "ldapDnPattern"                                                                                             // 1578
    });                                                                                                                // 1578
    return setting ? setting.value : ""; // empty if subscription is not ready.                                        // 1579
  },                                                                                                                   // 1580
  getLdapSearchUsername: function () {                                                                                 // 1582
    var setting = this.collections.settings.findOne({                                                                  // 1583
      _id: "ldapSearchUsername"                                                                                        // 1583
    });                                                                                                                // 1583
    return setting ? setting.value : ""; // empty if subscription is not ready.                                        // 1584
  },                                                                                                                   // 1585
  getLdapNameField: function () {                                                                                      // 1587
    var setting = this.collections.settings.findOne({                                                                  // 1588
      _id: "ldapNameField"                                                                                             // 1588
    });                                                                                                                // 1588
    return setting ? setting.value : ""; // empty if subscription is not ready.                                        // 1589
  },                                                                                                                   // 1590
  getLdapEmailField: function () {                                                                                     // 1592
    var setting = this.collections.settings.findOne({                                                                  // 1593
      _id: "ldapEmailField"                                                                                            // 1593
    });                                                                                                                // 1593
    return setting ? setting.value : "mail"; // default to "mail". This setting was added later, and so could potentially be unset.
  },                                                                                                                   // 1596
  getLdapExplicitDnSelected: function () {                                                                             // 1598
    var setting = this.collections.settings.findOne({                                                                  // 1599
      _id: "ldapExplicitDnSelected"                                                                                    // 1599
    });                                                                                                                // 1599
    return setting && setting.value;                                                                                   // 1600
  },                                                                                                                   // 1601
  getLdapFilter: function () {                                                                                         // 1603
    var setting = this.collections.settings.findOne({                                                                  // 1604
      _id: "ldapFilter"                                                                                                // 1604
    });                                                                                                                // 1604
    return setting ? setting.value : ""; // empty if subscription is not ready.                                        // 1605
  },                                                                                                                   // 1606
  getLdapSearchBindDn: function () {                                                                                   // 1608
    var setting = this.collections.settings.findOne({                                                                  // 1609
      _id: "ldapSearchBindDn"                                                                                          // 1609
    });                                                                                                                // 1609
    return setting ? setting.value : ""; // empty if subscription is not ready.                                        // 1610
  },                                                                                                                   // 1611
  getLdapSearchBindPassword: function () {                                                                             // 1613
    var setting = this.collections.settings.findOne({                                                                  // 1614
      _id: "ldapSearchBindPassword"                                                                                    // 1614
    });                                                                                                                // 1614
    return setting ? setting.value : ""; // empty if subscription is not ready.                                        // 1615
  },                                                                                                                   // 1616
  getLdapCaCert: function () {                                                                                         // 1618
    var setting = this.collections.settings.findOne({                                                                  // 1619
      _id: "ldapCaCert"                                                                                                // 1619
    });                                                                                                                // 1619
    return setting ? setting.value : ""; // empty if subscription is not ready.                                        // 1620
  },                                                                                                                   // 1621
  getOrganizationMembership: function () {                                                                             // 1623
    var setting = this.collections.settings.findOne({                                                                  // 1624
      _id: "organizationMembership"                                                                                    // 1624
    });                                                                                                                // 1624
    return setting && setting.value;                                                                                   // 1625
  },                                                                                                                   // 1626
  getOrganizationEmailEnabled: function () {                                                                           // 1628
    var membership = this.getOrganizationMembership();                                                                 // 1629
    return membership && membership.emailToken && membership.emailToken.enabled;                                       // 1630
  },                                                                                                                   // 1631
  getOrganizationEmailDomain: function () {                                                                            // 1633
    var membership = this.getOrganizationMembership();                                                                 // 1634
    return membership && membership.emailToken && membership.emailToken.domain;                                        // 1635
  },                                                                                                                   // 1636
  getOrganizationGoogleEnabled: function () {                                                                          // 1638
    var membership = this.getOrganizationMembership();                                                                 // 1639
    return membership && membership.google && membership.google.enabled;                                               // 1640
  },                                                                                                                   // 1641
  getOrganizationGoogleDomain: function () {                                                                           // 1643
    var membership = this.getOrganizationMembership();                                                                 // 1644
    return membership && membership.google && membership.google.domain;                                                // 1645
  },                                                                                                                   // 1646
  getOrganizationLdapEnabled: function () {                                                                            // 1648
    var membership = this.getOrganizationMembership();                                                                 // 1649
    return membership && membership.ldap && membership.ldap.enabled;                                                   // 1650
  },                                                                                                                   // 1651
  getOrganizationSamlEnabled: function () {                                                                            // 1653
    var membership = this.getOrganizationMembership();                                                                 // 1654
    return membership && membership.saml && membership.saml.enabled;                                                   // 1655
  },                                                                                                                   // 1656
  getOrganizationDisallowGuests: function () {                                                                         // 1658
    return this.getOrganizationDisallowGuestsRaw();                                                                    // 1659
  },                                                                                                                   // 1660
  getOrganizationDisallowGuestsRaw: function () {                                                                      // 1662
    var setting = this.collections.settings.findOne({                                                                  // 1663
      _id: "organizationSettings"                                                                                      // 1663
    });                                                                                                                // 1663
    return setting && setting.value && setting.value.disallowGuests;                                                   // 1664
  },                                                                                                                   // 1665
  getOrganizationShareContacts: function () {                                                                          // 1667
    return this.getOrganizationShareContactsRaw();                                                                     // 1668
  },                                                                                                                   // 1669
  getOrganizationShareContactsRaw: function () {                                                                       // 1671
    var setting = this.collections.settings.findOne({                                                                  // 1672
      _id: "organizationSettings"                                                                                      // 1672
    });                                                                                                                // 1672
                                                                                                                       //
    if (!setting || !setting.value || setting.value.shareContacts === undefined) {                                     // 1673
      // default to true if undefined                                                                                  // 1674
      return true;                                                                                                     // 1675
    } else {                                                                                                           // 1676
      return setting.value.shareContacts;                                                                              // 1677
    }                                                                                                                  // 1678
  },                                                                                                                   // 1679
  getSamlEntryPoint: function () {                                                                                     // 1681
    var setting = this.collections.settings.findOne({                                                                  // 1682
      _id: "samlEntryPoint"                                                                                            // 1682
    });                                                                                                                // 1682
    return setting ? setting.value : ""; // empty if subscription is not ready.                                        // 1683
  },                                                                                                                   // 1684
  getSamlLogout: function () {                                                                                         // 1686
    var setting = this.collections.settings.findOne({                                                                  // 1687
      _id: "samlLogout"                                                                                                // 1687
    });                                                                                                                // 1687
    return setting ? setting.value : ""; // empty if subscription is not ready.                                        // 1688
  },                                                                                                                   // 1689
  getSamlPublicCert: function () {                                                                                     // 1691
    var setting = this.collections.settings.findOne({                                                                  // 1692
      _id: "samlPublicCert"                                                                                            // 1692
    });                                                                                                                // 1692
    return setting ? setting.value : ""; // empty if subscription is not ready.                                        // 1693
  },                                                                                                                   // 1694
  getSamlEntityId: function () {                                                                                       // 1696
    var setting = this.collections.settings.findOne({                                                                  // 1697
      _id: "samlEntityId"                                                                                              // 1697
    });                                                                                                                // 1697
    return setting ? setting.value : ""; // empty if subscription is not ready.                                        // 1698
  },                                                                                                                   // 1699
  userHasSamlLoginCredential: function () {                                                                            // 1701
    var user = Meteor.user();                                                                                          // 1702
                                                                                                                       //
    if (!user.loginCredentials) {                                                                                      // 1703
      return false;                                                                                                    // 1704
    }                                                                                                                  // 1705
                                                                                                                       //
    var hasSaml = false;                                                                                               // 1707
    user.loginCredentials.forEach(function (credential) {                                                              // 1708
      if (Meteor.users.findOne({                                                                                       // 1709
        _id: credential.id                                                                                             // 1709
      }).services.saml) {                                                                                              // 1709
        hasSaml = true;                                                                                                // 1710
      }                                                                                                                // 1711
    });                                                                                                                // 1712
    return hasSaml;                                                                                                    // 1714
  },                                                                                                                   // 1715
  getActivitySubscriptions: function (grainId, threadPath) {                                                           // 1717
    return this.collections.activitySubscriptions.find({                                                               // 1718
      grainId: grainId,                                                                                                // 1719
      threadPath: threadPath || {                                                                                      // 1720
        $exists: false                                                                                                 // 1720
      }                                                                                                                // 1720
    }, {                                                                                                               // 1718
      fields: {                                                                                                        // 1722
        accountId: 1,                                                                                                  // 1722
        mute: 1,                                                                                                       // 1722
        _id: 0                                                                                                         // 1722
      }                                                                                                                // 1722
    }).fetch();                                                                                                        // 1721
  },                                                                                                                   // 1724
  subscribeToActivity: function (accountId, grainId, threadPath) {                                                     // 1726
    // Subscribe the given user to activity events with the given grainId and (optional)                               // 1727
    // threadPath -- unless the user has previously muted this grainId/threadPath, in which                            // 1728
    // case do nothing.                                                                                                // 1729
    var record = {                                                                                                     // 1731
      accountId: accountId,                                                                                            // 1731
      grainId: grainId                                                                                                 // 1731
    };                                                                                                                 // 1731
                                                                                                                       //
    if (threadPath) {                                                                                                  // 1732
      record.threadPath = threadPath;                                                                                  // 1733
    } // The $set here is redundant since an upsert automatically initializes a new record to contain                  // 1734
    // the fields from the query, but if we try to do { $set: {} } Mongo throws an exception, and                      // 1737
    // if we try to just pass {}, Mongo interprets it as "replace the record with an empty record".                    // 1738
    // What a wonderful query language.                                                                                // 1739
                                                                                                                       //
                                                                                                                       //
    this.collections.activitySubscriptions.upsert(record, {                                                            // 1740
      $set: record                                                                                                     // 1740
    });                                                                                                                // 1740
  },                                                                                                                   // 1741
  muteActivity: function (accountId, grainId, threadPath) {                                                            // 1743
    // Mute notifications for the given user originating from the given grainId and                                    // 1744
    // (optional) threadPath.                                                                                          // 1745
    var record = {                                                                                                     // 1747
      accountId: accountId,                                                                                            // 1747
      grainId: grainId                                                                                                 // 1747
    };                                                                                                                 // 1747
                                                                                                                       //
    if (threadPath) {                                                                                                  // 1748
      record.threadPath = threadPath;                                                                                  // 1749
    }                                                                                                                  // 1750
                                                                                                                       //
    this.collections.activitySubscriptions.upsert(record, {                                                            // 1752
      $set: {                                                                                                          // 1752
        mute: true                                                                                                     // 1752
      }                                                                                                                // 1752
    });                                                                                                                // 1752
  },                                                                                                                   // 1753
  updateAppIndex: function () {                                                                                        // 1755
    var _this3 = this;                                                                                                 // 1755
                                                                                                                       //
    var appUpdatesEnabledSetting = this.collections.settings.findOne({                                                 // 1756
      _id: "appUpdatesEnabled"                                                                                         // 1756
    });                                                                                                                // 1756
    var appUpdatesEnabled = appUpdatesEnabledSetting && appUpdatesEnabledSetting.value;                                // 1757
                                                                                                                       //
    if (!appUpdatesEnabled) {                                                                                          // 1758
      // It's much simpler to check appUpdatesEnabled here rather than reactively deactivate the                       // 1759
      // timer that triggers this call.                                                                                // 1760
      return;                                                                                                          // 1761
    }                                                                                                                  // 1762
                                                                                                                       //
    var appIndexUrl = this.collections.settings.findOne({                                                              // 1764
      _id: "appIndexUrl"                                                                                               // 1764
    }).value;                                                                                                          // 1764
    var appIndex = this.collections.appIndex;                                                                          // 1765
    var data = HTTP.get(appIndexUrl + "/apps/index.json").data;                                                        // 1766
    var preinstalledAppIds = this.getAllPreinstalledAppIds(); // We make sure to get all preinstalled appIds, even ones that are currently
    // downloading/failed.                                                                                             // 1769
                                                                                                                       //
    data.apps.forEach(function (app) {                                                                                 // 1770
      app._id = app.appId;                                                                                             // 1771
      var oldApp = appIndex.findOne({                                                                                  // 1773
        _id: app.appId                                                                                                 // 1773
      });                                                                                                              // 1773
      app.hasSentNotifications = false;                                                                                // 1774
      appIndex.upsert({                                                                                                // 1775
        _id: app._id                                                                                                   // 1775
      }, app);                                                                                                         // 1775
                                                                                                                       //
      var isAppPreinstalled = _.contains(preinstalledAppIds, app.appId);                                               // 1776
                                                                                                                       //
      if ((!oldApp || app.versionNumber > oldApp.versionNumber) && (_this3.collections.userActions.findOne({           // 1777
        appId: app.appId                                                                                               // 1778
      }) || isAppPreinstalled)) {                                                                                      // 1778
        var pack = _this3.collections.packages.findOne({                                                               // 1780
          _id: app.packageId                                                                                           // 1780
        });                                                                                                            // 1780
                                                                                                                       //
        var url = appIndexUrl + "/packages/" + app.packageId;                                                          // 1781
                                                                                                                       //
        if (pack) {                                                                                                    // 1782
          if (pack.status === "ready") {                                                                               // 1783
            if (pack.appId && pack.appId !== app.appId) {                                                              // 1784
              console.error("app index returned app ID and package ID that don't match:", JSON.stringify(app));        // 1785
            } else {                                                                                                   // 1787
              _this3.sendAppUpdateNotifications(app.appId, app.packageId, app.name, app.versionNumber, app.version);   // 1788
                                                                                                                       //
              if (isAppPreinstalled) {                                                                                 // 1790
                _this3.setPreinstallAppAsReady(app.appId, app.packageId);                                              // 1791
              }                                                                                                        // 1792
            }                                                                                                          // 1793
          } else {                                                                                                     // 1794
            var result = _this3.collections.packages.findAndModify({                                                   // 1795
              query: {                                                                                                 // 1796
                _id: app.packageId                                                                                     // 1796
              },                                                                                                       // 1796
              update: {                                                                                                // 1797
                $set: {                                                                                                // 1797
                  isAutoUpdated: true                                                                                  // 1797
                }                                                                                                      // 1797
              }                                                                                                        // 1797
            });                                                                                                        // 1795
                                                                                                                       //
            if (!result.ok) {                                                                                          // 1800
              return;                                                                                                  // 1801
            }                                                                                                          // 1802
                                                                                                                       //
            var newPack = result.value;                                                                                // 1804
                                                                                                                       //
            if (newPack.status === "ready") {                                                                          // 1805
              // The package was marked as ready before we applied isAutoUpdated=true. We should send                  // 1806
              // notifications ourselves to be sure there's no timing issue (sending more than one is                  // 1807
              // fine, since it will de-dupe).                                                                         // 1808
              if (pack.appId && pack.appId !== app.appId) {                                                            // 1809
                console.error("app index returned app ID and package ID that don't match:", JSON.stringify(app));      // 1810
              } else {                                                                                                 // 1812
                _this3.sendAppUpdateNotifications(app.appId, app.packageId, app.name, app.versionNumber, app.version);
                                                                                                                       //
                if (isAppPreinstalled) {                                                                               // 1815
                  _this3.setPreinstallAppAsReady(app.appId, app.packageId);                                            // 1816
                }                                                                                                      // 1817
              }                                                                                                        // 1818
            } else if (newPack.status === "failed") {                                                                  // 1819
              // If the package has failed, retry it                                                                   // 1820
              _this3.startInstall(app.packageId, url, true, true);                                                     // 1821
            }                                                                                                          // 1822
          }                                                                                                            // 1823
        } else {                                                                                                       // 1824
          _this3.startInstall(app.packageId, url, false, true);                                                        // 1825
        }                                                                                                              // 1826
      }                                                                                                                // 1827
    });                                                                                                                // 1828
  },                                                                                                                   // 1829
  isPackagePreinstalled: function (packageId) {                                                                        // 1831
    return this.collections.settings.find({                                                                            // 1832
      _id: "preinstalledApps",                                                                                         // 1832
      "value.packageId": packageId                                                                                     // 1832
    }).count() === 1;                                                                                                  // 1832
  },                                                                                                                   // 1833
  getAppIdForPreinstalledPackage: function (packageId) {                                                               // 1835
    var setting = this.collections.settings.findOne({                                                                  // 1836
      _id: "preinstalledApps",                                                                                         // 1836
      "value.packageId": packageId                                                                                     // 1836
    }, {                                                                                                               // 1836
      fields: {                                                                                                        // 1837
        "value.$": 1                                                                                                   // 1837
      }                                                                                                                // 1837
    }); // value.$ causes mongo to transform the result and only return the first matching element in                  // 1837
    // the array                                                                                                       // 1839
                                                                                                                       //
    return setting && setting.value && setting.value[0] && setting.value[0].appId;                                     // 1840
  },                                                                                                                   // 1841
  getPackageIdForPreinstalledApp: function (appId) {                                                                   // 1843
    var setting = this.collections.settings.findOne({                                                                  // 1844
      _id: "preinstalledApps",                                                                                         // 1844
      "value.appId": appId                                                                                             // 1844
    }, {                                                                                                               // 1844
      fields: {                                                                                                        // 1845
        "value.$": 1                                                                                                   // 1845
      }                                                                                                                // 1845
    }); // value.$ causes mongo to transform the result and only return the first matching element in                  // 1845
    // the array                                                                                                       // 1847
                                                                                                                       //
    return setting && setting.value && setting.value[0] && setting.value[0].packageId;                                 // 1848
  },                                                                                                                   // 1849
  getReadyPreinstalledAppIds: function () {                                                                            // 1851
    var setting = this.collections.settings.findOne({                                                                  // 1852
      _id: "preinstalledApps"                                                                                          // 1852
    });                                                                                                                // 1852
    var ret = setting && setting.value || [];                                                                          // 1853
    return _.chain(ret).filter(function (app) {                                                                        // 1854
      return app.status === "ready";                                                                                   // 1855
    }).map(function (app) {                                                                                            // 1855
      return app.appId;                                                                                                // 1856
    }).value();                                                                                                        // 1856
  },                                                                                                                   // 1858
  getAllPreinstalledAppIds: function () {                                                                              // 1860
    var setting = this.collections.settings.findOne({                                                                  // 1861
      _id: "preinstalledApps"                                                                                          // 1861
    });                                                                                                                // 1861
    var ret = setting && setting.value || [];                                                                          // 1862
    return _.map(ret, function (app) {                                                                                 // 1863
      return app.appId;                                                                                                // 1863
    });                                                                                                                // 1863
  },                                                                                                                   // 1864
  preinstallAppsForUser: function (userId) {                                                                           // 1866
    var _this4 = this;                                                                                                 // 1866
                                                                                                                       //
    var appIds = this.getReadyPreinstalledAppIds();                                                                    // 1867
    appIds.forEach(function (appId) {                                                                                  // 1868
      try {                                                                                                            // 1869
        _this4.addUserActions(userId, _this4.getPackageIdForPreinstalledApp(appId));                                   // 1870
      } catch (e) {                                                                                                    // 1871
        console.error("failed to install app for user:", e);                                                           // 1872
      }                                                                                                                // 1873
    });                                                                                                                // 1874
  },                                                                                                                   // 1875
  setPreinstallAppAsDownloading: function (appId, packageId) {                                                         // 1877
    this.collections.settings.update({                                                                                 // 1878
      _id: "preinstalledApps",                                                                                         // 1879
      "value.appId": appId,                                                                                            // 1879
      "value.packageId": packageId                                                                                     // 1879
    }, {                                                                                                               // 1879
      $set: {                                                                                                          // 1880
        "value.$.status": "downloading"                                                                                // 1880
      }                                                                                                                // 1880
    });                                                                                                                // 1880
  },                                                                                                                   // 1881
  setPreinstallAppAsReady: function (appId, packageId) {                                                               // 1883
    // This function both sets the appId as ready and updates the packageId for the given appId                        // 1884
    // Setting the packageId is especially useful in installer.js, as it always ensures the                            // 1885
    // latest installed package will be set as ready.                                                                  // 1886
    this.collections.settings.update({                                                                                 // 1887
      _id: "preinstalledApps",                                                                                         // 1888
      "value.appId": appId                                                                                             // 1888
    }, {                                                                                                               // 1888
      $set: {                                                                                                          // 1889
        "value.$.status": "ready",                                                                                     // 1889
        "value.$.packageId": packageId                                                                                 // 1889
      }                                                                                                                // 1889
    });                                                                                                                // 1889
  },                                                                                                                   // 1890
  ensureAppPreinstall: function (appId, packageId) {                                                                   // 1892
    check(appId, String);                                                                                              // 1893
    var appIndexUrl = this.collections.settings.findOne({                                                              // 1894
      _id: "appIndexUrl"                                                                                               // 1894
    }).value;                                                                                                          // 1894
    var pack = this.collections.packages.findOne({                                                                     // 1895
      _id: packageId                                                                                                   // 1895
    });                                                                                                                // 1895
    var url = appIndexUrl + "/packages/" + packageId;                                                                  // 1896
                                                                                                                       //
    if (pack && pack.status === "ready") {                                                                             // 1897
      this.setPreinstallAppAsReady(appId, packageId);                                                                  // 1898
    } else if (pack && pack.status === "failed") {                                                                     // 1899
      this.setPreinstallAppAsDownloading(appId, packageId);                                                            // 1900
      this.startInstall(packageId, url, true, false);                                                                  // 1901
    } else {                                                                                                           // 1902
      this.setPreinstallAppAsDownloading(appId, packageId);                                                            // 1903
      this.startInstall(packageId, url, false, false);                                                                 // 1904
    }                                                                                                                  // 1905
  },                                                                                                                   // 1906
  setPreinstalledApps: function (appAndPackageIds) {                                                                   // 1908
    var _this5 = this;                                                                                                 // 1908
                                                                                                                       //
    // appAndPackageIds: A List[Object] where each element has fields:                                                 // 1909
    //     appId: The Packages.appId of the app to install                                                             // 1910
    //     packageId: The Packages._id of the app to install                                                           // 1911
    check(appAndPackageIds, [{                                                                                         // 1912
      appId: String,                                                                                                   // 1912
      packageId: String                                                                                                // 1912
    }]); // Start by clearing out the setting. We'll push appIds one by one to it                                      // 1912
                                                                                                                       //
    this.collections.settings.upsert({                                                                                 // 1915
      _id: "preinstalledApps"                                                                                          // 1915
    }, {                                                                                                               // 1915
      $set: {                                                                                                          // 1915
        value: appAndPackageIds.map(function (data) {                                                                  // 1916
          return {                                                                                                     // 1917
            appId: data.appId,                                                                                         // 1918
            status: "notReady",                                                                                        // 1919
            packageId: data.packageId                                                                                  // 1920
          };                                                                                                           // 1917
        })                                                                                                             // 1922
      }                                                                                                                // 1915
    });                                                                                                                // 1915
    appAndPackageIds.forEach(function (data) {                                                                         // 1924
      _this5.ensureAppPreinstall(data.appId, data.packageId);                                                          // 1925
    });                                                                                                                // 1926
  },                                                                                                                   // 1927
  getProductivitySuiteAppIds: function () {                                                                            // 1929
    return ["8aspz4sfjnp8u89000mh2v1xrdyx97ytn8hq71mdzv4p4d8n0n3h", // Davros                                          // 1930
    "h37dm17aa89yrd8zuqpdn36p6zntumtv08fjpu8a8zrte7q1cn60", // Etherpad                                                // 1932
    "vfnwptfn02ty21w715snyyczw0nqxkv3jvawcah10c6z7hj1hnu0", // Rocket.Chat                                             // 1933
    "m86q05rdvj14yvn78ghaxynqz7u2svw6rnttptxx49g1785cdv1h"];                                                           // 1934
  },                                                                                                                   // 1936
  getSystemSuiteAppIds: function () {                                                                                  // 1938
    return ["s3u2xgmqwznz2n3apf30sm3gw1d85y029enw5pymx734cnk5n78h"];                                                   // 1939
  },                                                                                                                   // 1942
  isPreinstalledAppsReady: function () {                                                                               // 1944
    var setting = this.collections.settings.findOne({                                                                  // 1945
      _id: "preinstalledApps"                                                                                          // 1945
    });                                                                                                                // 1945
                                                                                                                       //
    if (!setting || !setting.value) {                                                                                  // 1946
      return true;                                                                                                     // 1947
    }                                                                                                                  // 1948
                                                                                                                       //
    var packageIds = _.pluck(setting.value, "packageId");                                                              // 1950
                                                                                                                       //
    var readyApps = this.collections.packages.find({                                                                   // 1951
      _id: {                                                                                                           // 1952
        $in: packageIds                                                                                                // 1953
      },                                                                                                               // 1952
      status: "ready"                                                                                                  // 1955
    });                                                                                                                // 1951
    return readyApps.count() === packageIds.length;                                                                    // 1957
  },                                                                                                                   // 1958
  getBillingPromptUrl: function () {                                                                                   // 1960
    var setting = this.collections.settings.findOne({                                                                  // 1961
      _id: "billingPromptUrl"                                                                                          // 1961
    });                                                                                                                // 1961
    return setting && setting.value;                                                                                   // 1962
  },                                                                                                                   // 1963
  isReferralEnabled: function () {                                                                                     // 1965
    // This function is a bit weird, in that we've transitioned from                                                   // 1966
    // Meteor.settings.public.quotaEnabled to DB settings. For now,                                                    // 1967
    // Meteor.settings.public.quotaEnabled implies bothisReferralEnabled and isQuotaEnabled are true.                  // 1968
    return Meteor.settings.public.quotaEnabled;                                                                        // 1969
  },                                                                                                                   // 1970
  isHideAboutEnabled: function () {                                                                                    // 1972
    var setting = this.collections.settings.findOne({                                                                  // 1973
      _id: "whiteLabelHideAbout"                                                                                       // 1973
    });                                                                                                                // 1973
    return setting && setting.value;                                                                                   // 1974
  },                                                                                                                   // 1975
  isQuotaEnabled: function () {                                                                                        // 1977
    if (Meteor.settings.public.quotaEnabled) return true;                                                              // 1978
    var setting = this.collections.settings.findOne({                                                                  // 1980
      _id: "quotaEnabled"                                                                                              // 1980
    });                                                                                                                // 1980
    return setting && setting.value;                                                                                   // 1981
  },                                                                                                                   // 1982
  isQuotaLdapEnabled: function () {                                                                                    // 1984
    var setting = this.collections.settings.findOne({                                                                  // 1985
      _id: "quotaLdapEnabled"                                                                                          // 1985
    });                                                                                                                // 1985
    return setting && setting.value;                                                                                   // 1986
  },                                                                                                                   // 1987
  updateUserQuota: function (user) {                                                                                   // 1989
    if (this.quotaManager) {                                                                                           // 1990
      return this.quotaManager.updateUserQuota(this, user);                                                            // 1991
    }                                                                                                                  // 1992
  },                                                                                                                   // 1993
  getUserQuota: function (user) {                                                                                      // 1995
    if (this.isQuotaLdapEnabled()) {                                                                                   // 1996
      return this.quotaManager.updateUserQuota(this, user);                                                            // 1997
    } else {                                                                                                           // 1998
      var plan = this.getPlan(user.plan || "free", user);                                                              // 1999
      var referralBonus = calculateReferralBonus(user);                                                                // 2000
      var bonus = user.planBonus || {};                                                                                // 2001
      var userQuota = {                                                                                                // 2002
        storage: plan.storage + referralBonus.storage + (bonus.storage || 0),                                          // 2003
        grains: plan.grains + referralBonus.grains + (bonus.grains || 0),                                              // 2004
        compute: plan.compute + (bonus.compute || 0)                                                                   // 2005
      };                                                                                                               // 2002
      return userQuota;                                                                                                // 2007
    }                                                                                                                  // 2008
  },                                                                                                                   // 2009
  isUserOverQuota: function (user) {                                                                                   // 2011
    // Return false if user has quota space remaining, true if it is full. When this returns true,                     // 2012
    // we will not allow the user to create new grains, though they may be able to open existing ones                  // 2013
    // which may still increase their storage usage.                                                                   // 2014
    //                                                                                                                 // 2015
    // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                                // 2016
    if (!this.isQuotaEnabled() || user.isAdmin) return false;                                                          // 2018
    var plan = this.getUserQuota(user);                                                                                // 2020
                                                                                                                       //
    if (plan.grains < Infinity) {                                                                                      // 2021
      var count = this.collections.grains.find({                                                                       // 2022
        userId: user._id,                                                                                              // 2022
        trashed: {                                                                                                     // 2022
          $exists: false                                                                                               // 2022
        }                                                                                                              // 2022
      }, {                                                                                                             // 2022
        fields: {},                                                                                                    // 2023
        limit: plan.grains                                                                                             // 2023
      }).count();                                                                                                      // 2023
      if (count >= plan.grains) return "outOfGrains";                                                                  // 2024
    }                                                                                                                  // 2025
                                                                                                                       //
    return plan && user.storageUsage && user.storageUsage >= plan.storage && "outOfStorage";                           // 2027
  },                                                                                                                   // 2028
  isUserExcessivelyOverQuota: function (user) {                                                                        // 2030
    // Return true if user is so far over quota that we should prevent their existing grains from                      // 2031
    // running at all.                                                                                                 // 2032
    //                                                                                                                 // 2033
    // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                                // 2034
    if (!this.isQuotaEnabled() || user.isAdmin) return false;                                                          // 2036
    var quota = this.getUserQuota(user); // quota.grains = Infinity means unlimited grains. IEEE754 defines Infinity == Infinity.
                                                                                                                       //
    if (quota.grains < Infinity) {                                                                                     // 2041
      var count = this.collections.grains.find({                                                                       // 2042
        userId: user._id,                                                                                              // 2042
        trashed: {                                                                                                     // 2042
          $exists: false                                                                                               // 2042
        }                                                                                                              // 2042
      }, {                                                                                                             // 2042
        fields: {},                                                                                                    // 2043
        limit: quota.grains * 2                                                                                        // 2043
      }).count();                                                                                                      // 2043
      if (count >= quota.grains * 2) return "outOfGrains";                                                             // 2044
    }                                                                                                                  // 2045
                                                                                                                       //
    return quota && user.storageUsage && user.storageUsage >= quota.storage * 1.2 && "outOfStorage";                   // 2047
  },                                                                                                                   // 2048
  suspendCredential: function (credentialId, suspension) {                                                             // 2050
    check(credentialId, String);                                                                                       // 2051
    check(suspension, {                                                                                                // 2052
      timestamp: Date,                                                                                                 // 2053
      admin: Match.Optional(String),                                                                                   // 2054
      voluntary: Match.Optional(Boolean)                                                                               // 2055
    });                                                                                                                // 2052
    this.collections.users.update({                                                                                    // 2058
      _id: credentialId                                                                                                // 2058
    }, {                                                                                                               // 2058
      $set: {                                                                                                          // 2058
        suspended: suspension                                                                                          // 2058
      }                                                                                                                // 2058
    });                                                                                                                // 2058
  },                                                                                                                   // 2059
  unsuspendCredential: function (credentialId) {                                                                       // 2061
    check(credentialId, String);                                                                                       // 2062
    this.collections.users.update({                                                                                    // 2064
      _id: credentialId                                                                                                // 2064
    }, {                                                                                                               // 2064
      $unset: {                                                                                                        // 2064
        suspended: 1                                                                                                   // 2064
      }                                                                                                                // 2064
    });                                                                                                                // 2064
  },                                                                                                                   // 2065
  suspendAccount: function (userId, byAdminUserId, willDelete) {                                                       // 2067
    var _this6 = this;                                                                                                 // 2067
                                                                                                                       //
    check(userId, String);                                                                                             // 2068
    check(byAdminUserId, Match.OneOf(String, null, undefined));                                                        // 2069
    check(willDelete, Boolean);                                                                                        // 2070
    var user = this.collections.users.findOne({                                                                        // 2072
      _id: userId                                                                                                      // 2072
    });                                                                                                                // 2072
    var suspension = {                                                                                                 // 2073
      timestamp: new Date(),                                                                                           // 2074
      willDelete: willDelete || false                                                                                  // 2075
    };                                                                                                                 // 2073
                                                                                                                       //
    if (byAdminUserId) {                                                                                               // 2077
      suspension.admin = byAdminUserId;                                                                                // 2078
    } else {                                                                                                           // 2079
      suspension.voluntary = true;                                                                                     // 2080
    }                                                                                                                  // 2081
                                                                                                                       //
    this.collections.users.update({                                                                                    // 2083
      _id: userId                                                                                                      // 2083
    }, {                                                                                                               // 2083
      $set: {                                                                                                          // 2083
        suspended: suspension                                                                                          // 2083
      }                                                                                                                // 2083
    });                                                                                                                // 2083
    this.collections.grains.update({                                                                                   // 2084
      userId: userId                                                                                                   // 2084
    }, {                                                                                                               // 2084
      $set: {                                                                                                          // 2084
        suspended: true                                                                                                // 2084
      }                                                                                                                // 2084
    }, {                                                                                                               // 2084
      multi: true                                                                                                      // 2084
    });                                                                                                                // 2084
    this.collections.apiTokens.update({                                                                                // 2086
      "owner.user.accountId": userId                                                                                   // 2086
    }, {                                                                                                               // 2086
      $set: {                                                                                                          // 2087
        suspended: true                                                                                                // 2087
      }                                                                                                                // 2087
    }, {                                                                                                               // 2087
      multi: true                                                                                                      // 2087
    });                                                                                                                // 2087
    delete suspension.willDelete; // Only mark the parent account for deletion. This makes the query simpler later.    // 2089
                                                                                                                       //
    user.loginCredentials.forEach(function (credential) {                                                              // 2092
      _this6.suspendCredential(credential.id, suspension);                                                             // 2093
    });                                                                                                                // 2094
    user.nonloginCredentials.forEach(function (credential) {                                                           // 2095
      if (_this6.collections.users.find({                                                                              // 2096
        $or: [{                                                                                                        // 2096
          "loginCredentials.id": credential.id                                                                         // 2097
        }, {                                                                                                           // 2097
          "nonloginCredentials.id": credential.id                                                                      // 2098
        }]                                                                                                             // 2098
      }).count() === 1) {                                                                                              // 2096
        // Only suspend non-login credential that are unique to this account.                                          // 2100
        _this6.suspendCredential(credential.id, suspension);                                                           // 2101
      }                                                                                                                // 2102
    }); // Force logout this user                                                                                      // 2103
                                                                                                                       //
    this.collections.users.update({                                                                                    // 2106
      _id: userId                                                                                                      // 2106
    }, {                                                                                                               // 2106
      $unset: {                                                                                                        // 2107
        "services.resume.loginTokens": 1                                                                               // 2107
      }                                                                                                                // 2107
    });                                                                                                                // 2107
                                                                                                                       //
    if (user && user.loginCredentials) {                                                                               // 2108
      user.loginCredentials.forEach(function (credential) {                                                            // 2109
        Meteor.users.update({                                                                                          // 2110
          _id: credential.id                                                                                           // 2110
        }, {                                                                                                           // 2110
          $unset: {                                                                                                    // 2111
            "services.resume.loginTokens": 1                                                                           // 2111
          }                                                                                                            // 2111
        });                                                                                                            // 2111
      });                                                                                                              // 2112
    }                                                                                                                  // 2113
  },                                                                                                                   // 2114
  unsuspendAccount: function (userId) {                                                                                // 2116
    var _this7 = this;                                                                                                 // 2116
                                                                                                                       //
    check(userId, String);                                                                                             // 2117
    var user = this.collections.users.findOne({                                                                        // 2119
      _id: userId                                                                                                      // 2119
    });                                                                                                                // 2119
    this.collections.users.update({                                                                                    // 2120
      _id: userId                                                                                                      // 2120
    }, {                                                                                                               // 2120
      $unset: {                                                                                                        // 2120
        suspended: 1                                                                                                   // 2120
      }                                                                                                                // 2120
    });                                                                                                                // 2120
    this.collections.grains.update({                                                                                   // 2121
      userId: userId                                                                                                   // 2121
    }, {                                                                                                               // 2121
      $unset: {                                                                                                        // 2121
        suspended: 1                                                                                                   // 2121
      }                                                                                                                // 2121
    }, {                                                                                                               // 2121
      multi: true                                                                                                      // 2121
    });                                                                                                                // 2121
    this.collections.apiTokens.update({                                                                                // 2123
      "owner.user.accountId": userId                                                                                   // 2123
    }, {                                                                                                               // 2123
      $unset: {                                                                                                        // 2124
        suspended: true                                                                                                // 2124
      }                                                                                                                // 2124
    }, {                                                                                                               // 2124
      multi: true                                                                                                      // 2124
    });                                                                                                                // 2124
    user.loginCredentials.forEach(function (credential) {                                                              // 2126
      _this7.unsuspendCredential(credential.id);                                                                       // 2127
    });                                                                                                                // 2128
    user.nonloginCredentials.forEach(function (credential) {                                                           // 2130
      _this7.unsuspendCredential(credential.id);                                                                       // 2131
    });                                                                                                                // 2132
  },                                                                                                                   // 2133
  deletePendingAccounts: function (deletionCoolingOffTime, backend, cb) {                                              // 2135
    var _this8 = this;                                                                                                 // 2135
                                                                                                                       //
    check(deletionCoolingOffTime, Number);                                                                             // 2136
    var queryDate = new Date(Date.now() - deletionCoolingOffTime);                                                     // 2138
    this.collections.users.find({                                                                                      // 2139
      "suspended.willDelete": true,                                                                                    // 2140
      "suspended.timestamp": {                                                                                         // 2141
        $lt: queryDate                                                                                                 // 2141
      }                                                                                                                // 2141
    }).forEach(function (user) {                                                                                       // 2139
      if (cb) cb(_this8, user);                                                                                        // 2143
                                                                                                                       //
      _this8.deleteAccount(user._id, backend);                                                                         // 2144
    });                                                                                                                // 2145
  },                                                                                                                   // 2146
  hostIsStandalone: function (hostname) {                                                                              // 2148
    check(hostname, String);                                                                                           // 2149
    return !!this.collections.standaloneDomains.findOne({                                                              // 2151
      _id: hostname                                                                                                    // 2151
    });                                                                                                                // 2151
  }                                                                                                                    // 2152
});                                                                                                                    // 1255
                                                                                                                       //
SandstormDb.escapeMongoKey = function (key) {                                                                          // 2155
  // This incredibly poor mechanism for escaping Mongo keys is recommended by the Mongo docs here:                     // 2156
  //   https://docs.mongodb.org/manual/faq/developers/#dollar-sign-operator-escaping                                   // 2157
  // and seems to be a de facto standard, for example:                                                                 // 2158
  //   https://www.npmjs.com/package/mongo-key-escape                                                                  // 2159
  return key.replace(".", "\uFF0E").replace("$", "\uFF04");                                                            // 2160
};                                                                                                                     // 2161
                                                                                                                       //
var appNameFromPackage = function (packageObj) {                                                                       // 2163
  // This function takes a Package object from Mongo and returns an                                                    // 2164
  // app title.                                                                                                        // 2165
  var manifest = packageObj.manifest;                                                                                  // 2166
  if (!manifest) return packageObj.appId || packageObj._id || "unknown";                                               // 2167
  var action = manifest.actions[0];                                                                                    // 2168
  appName = manifest.appTitle && manifest.appTitle.defaultText || appNameFromActionName(action.title.defaultText);     // 2169
  return appName;                                                                                                      // 2171
};                                                                                                                     // 2172
                                                                                                                       //
var appNameFromActionName = function (name) {                                                                          // 2174
  // Hack: Historically we only had action titles, like "New Etherpad Document", not app                               // 2175
  //   titles. But for this UI we want app titles. As a transitionary measure, try to                                  // 2176
  //   derive the app title from the action title.                                                                     // 2177
  // TODO(cleanup): Get rid of this once apps have real titles.                                                        // 2178
  if (!name) {                                                                                                         // 2179
    return "(unnamed)";                                                                                                // 2180
  }                                                                                                                    // 2181
                                                                                                                       //
  if (name.lastIndexOf("New ", 0) === 0) {                                                                             // 2183
    name = name.slice(4);                                                                                              // 2184
  }                                                                                                                    // 2185
                                                                                                                       //
  if (name.lastIndexOf("Hacker CMS", 0) === 0) {                                                                       // 2187
    name = "Hacker CMS";                                                                                               // 2188
  } else {                                                                                                             // 2189
    var space = name.indexOf(" ");                                                                                     // 2190
                                                                                                                       //
    if (space > 0) {                                                                                                   // 2191
      name = name.slice(0, space);                                                                                     // 2192
    }                                                                                                                  // 2193
  }                                                                                                                    // 2194
                                                                                                                       //
  return name;                                                                                                         // 2196
};                                                                                                                     // 2197
                                                                                                                       //
var appShortDescriptionFromPackage = function (pkg) {                                                                  // 2199
  return pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.shortDescription && pkg.manifest.metadata.shortDescription.defaultText;
};                                                                                                                     // 2203
                                                                                                                       //
var nounPhraseForActionAndAppTitle = function (action, appTitle) {                                                     // 2205
  // A hack to deal with legacy apps not including fields in their manifests.                                          // 2206
  // I look forward to the day I can remove most of this code.                                                         // 2207
  // Attempt to figure out the appropriate noun that this action will create.                                          // 2208
  // Use an explicit noun phrase is one is available.  Apps should add these in the future.                            // 2209
  if (action.nounPhrase) return action.nounPhrase.defaultText; // Otherwise, try to guess one from the structure of the action title field
                                                                                                                       //
  if (action.title && action.title.defaultText) {                                                                      // 2212
    var text = action.title.defaultText; // Strip a leading "New "                                                     // 2213
                                                                                                                       //
    if (text.lastIndexOf("New ", 0) === 0) {                                                                           // 2215
      var candidate = text.slice(4); // Strip a leading appname too, if provided                                       // 2216
                                                                                                                       //
      if (candidate.lastIndexOf(appTitle, 0) === 0) {                                                                  // 2218
        var newCandidate = candidate.slice(appTitle.length); // Unless that leaves you with no noun, in which case, use "grain"
                                                                                                                       //
        if (newCandidate.length > 0) {                                                                                 // 2221
          return newCandidate.toLowerCase();                                                                           // 2222
        } else {                                                                                                       // 2223
          return "grain";                                                                                              // 2224
        }                                                                                                              // 2225
      }                                                                                                                // 2226
                                                                                                                       //
      return candidate.toLowerCase();                                                                                  // 2228
    } // Some other verb phrase was given.  Just use it verbatim, and hope the app author updates                      // 2229
    // the package soon.                                                                                               // 2231
                                                                                                                       //
                                                                                                                       //
    return text;                                                                                                       // 2232
  } else {                                                                                                             // 2233
    return "grain";                                                                                                    // 2234
  }                                                                                                                    // 2235
}; // Static methods on SandstormDb that don't need an instance.                                                       // 2236
// Largely things that deal with backwards-compatibility.                                                              // 2239
                                                                                                                       //
                                                                                                                       //
_.extend(SandstormDb, {                                                                                                // 2240
  appNameFromActionName: appNameFromActionName,                                                                        // 2241
  appNameFromPackage: appNameFromPackage,                                                                              // 2242
  appShortDescriptionFromPackage: appShortDescriptionFromPackage,                                                      // 2243
  nounPhraseForActionAndAppTitle: nounPhraseForActionAndAppTitle                                                       // 2244
});                                                                                                                    // 2240
                                                                                                                       //
if (Meteor.isServer) {                                                                                                 // 2247
  var _Crypto2 = Npm.require("crypto");                                                                                // 2248
                                                                                                                       //
  var ContentType = Npm.require("content-type");                                                                       // 2249
                                                                                                                       //
  var Zlib = Npm.require("zlib");                                                                                      // 2250
                                                                                                                       //
  var _Url2 = Npm.require("url");                                                                                      // 2251
                                                                                                                       //
  var replicaNumber = Meteor.settings.replicaNumber || 0;                                                              // 2253
                                                                                                                       //
  var computeStagger = function (n) {                                                                                  // 2255
    // Compute a fraction in the range [0, 1) such that, for any natural number k, the values                          // 2256
    // of computeStagger(n) for all n in [1, 2^k) are uniformly distributed between 0 and 1.                           // 2257
    // The sequence looks like:                                                                                        // 2258
    //   0, 1/2, 1/4, 3/4, 1/8, 3/8, 5/8, 7/8, 1/16, ...                                                               // 2259
    //                                                                                                                 // 2260
    // We use this to determine how we'll stagger periodic events performed by this replica.                           // 2261
    // Notice that this allows us to compute a stagger which is independent of the number of                           // 2262
    // front-end replicas present; we can add more replicas to the end without affecting how the                       // 2263
    // earlier ones schedule their events.                                                                             // 2264
    var denom = 1;                                                                                                     // 2265
                                                                                                                       //
    while (denom <= n) {                                                                                               // 2266
      denom <<= 1;                                                                                                     // 2266
    }                                                                                                                  // 2266
                                                                                                                       //
    var num = n * 2 - denom + 1;                                                                                       // 2267
    return num / denom;                                                                                                // 2268
  };                                                                                                                   // 2269
                                                                                                                       //
  var stagger = computeStagger(replicaNumber);                                                                         // 2271
                                                                                                                       //
  SandstormDb.periodicCleanup = function (intervalMs, callback) {                                                      // 2273
    // Register a database cleanup function than should run periodically, roughly once every                           // 2274
    // interval of the given length.                                                                                   // 2275
    //                                                                                                                 // 2276
    // In a blackrock deployment with multiple front-ends, the frequency of the cleanup will be                        // 2277
    // scaled appropriately on the assumption that more data is being generated demanding more                         // 2278
    // frequent cleanups.                                                                                              // 2279
    check(intervalMs, Number);                                                                                         // 2281
    check(callback, Function);                                                                                         // 2282
                                                                                                                       //
    if (intervalMs < 120000) {                                                                                         // 2284
      throw new Error("less than 2-minute cleanup interval seems too fast; " + "are you using the right units?");      // 2285
    } // Schedule first cleanup to happen at the next intervalMs interval from the epoch, so that                      // 2287
    // the schedule is independent of the exact startup time.                                                          // 2290
                                                                                                                       //
                                                                                                                       //
    var first = intervalMs - Date.now() % intervalMs; // Stagger cleanups across replicas so that we don't have all replicas trying to clean the
    // same data at the same time.                                                                                     // 2294
                                                                                                                       //
    first += Math.floor(intervalMs * computeStagger(replicaNumber)); // If the stagger put us more than an interval away from now, back up.
                                                                                                                       //
    if (first > intervalMs) first -= intervalMs;                                                                       // 2298
    Meteor.setTimeout(function () {                                                                                    // 2300
      callback();                                                                                                      // 2301
      Meteor.setInterval(callback, intervalMs);                                                                        // 2302
    }, first);                                                                                                         // 2303
  }; // TODO(cleanup): Node 0.12 has a `gzipSync` but 0.10 (which Meteor still uses) does not.                         // 2304
                                                                                                                       //
                                                                                                                       //
  var gzipSync = Meteor.wrapAsync(Zlib.gzip, Zlib);                                                                    // 2307
                                                                                                                       //
  var BufferSmallerThan = function (limit) {                                                                           // 2309
    return Match.Where(function (buf) {                                                                                // 2310
      check(buf, Buffer);                                                                                              // 2311
      return buf.length < limit;                                                                                       // 2312
    });                                                                                                                // 2313
  };                                                                                                                   // 2314
                                                                                                                       //
  var DatabaseId = Match.Where(function (s) {                                                                          // 2316
    check(s, String);                                                                                                  // 2317
    return !!s.match(/^[a-zA-Z0-9_]+$/);                                                                               // 2318
  });                                                                                                                  // 2319
                                                                                                                       //
  SandstormDb.prototype.addStaticAsset = function (metadata, content) {                                                // 2321
    // Add a new static asset to the database. If `content` is a string rather than a buffer, it                       // 2322
    // will be automatically gzipped before storage; do not specify metadata.encoding in this case.                    // 2323
    if (typeof content === "string" && !metadata.encoding) {                                                           // 2325
      content = gzipSync(new Buffer(content, "utf8"));                                                                 // 2326
      metadata.encoding = "gzip";                                                                                      // 2327
    }                                                                                                                  // 2328
                                                                                                                       //
    check(metadata, {                                                                                                  // 2330
      mimeType: String,                                                                                                // 2331
      encoding: Match.Optional("gzip")                                                                                 // 2332
    });                                                                                                                // 2330
    check(content, BufferSmallerThan(1 << 20)); // Validate content type.                                              // 2334
                                                                                                                       //
    metadata.mimeType = ContentType.format(ContentType.parse(metadata.mimeType));                                      // 2337
                                                                                                                       //
    var hasher = _Crypto2.createHash("sha256");                                                                        // 2339
                                                                                                                       //
    hasher.update(metadata.mimeType + "\n" + metadata.encoding + "\n", "utf8");                                        // 2340
    hasher.update(content);                                                                                            // 2341
    var hash = hasher.digest("base64");                                                                                // 2342
    var result = this.collections.staticAssets.findAndModify({                                                         // 2344
      query: {                                                                                                         // 2345
        hash: hash,                                                                                                    // 2345
        refcount: {                                                                                                    // 2345
          $gte: 1                                                                                                      // 2345
        }                                                                                                              // 2345
      },                                                                                                               // 2345
      update: {                                                                                                        // 2346
        $inc: {                                                                                                        // 2346
          refcount: 1                                                                                                  // 2346
        }                                                                                                              // 2346
      },                                                                                                               // 2346
      fields: {                                                                                                        // 2347
        _id: 1,                                                                                                        // 2347
        refcount: 1                                                                                                    // 2347
      }                                                                                                                // 2347
    });                                                                                                                // 2344
                                                                                                                       //
    if (!result.ok) {                                                                                                  // 2350
      throw new Error("Couldn't increment refcount of asset with hash " + hash);                                       // 2351
    }                                                                                                                  // 2352
                                                                                                                       //
    var existing = result.value;                                                                                       // 2354
                                                                                                                       //
    if (existing) {                                                                                                    // 2355
      return existing._id;                                                                                             // 2356
    }                                                                                                                  // 2357
                                                                                                                       //
    return this.collections.staticAssets.insert(_.extend({                                                             // 2359
      hash: hash,                                                                                                      // 2360
      content: content,                                                                                                // 2361
      refcount: 1                                                                                                      // 2362
    }, metadata));                                                                                                     // 2359
  };                                                                                                                   // 2364
                                                                                                                       //
  SandstormDb.prototype.refStaticAsset = function (id) {                                                               // 2366
    // Increment the refcount on an existing static asset. Returns the asset on success.                               // 2367
    // If the asset does not exist, returns a falsey value.                                                            // 2368
    //                                                                                                                 // 2369
    // You must call this BEFORE adding the new reference to the DB, in case of failure between                        // 2370
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big                       // 2371
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                                         // 2372
    check(id, String);                                                                                                 // 2374
    var result = this.collections.staticAssets.findAndModify({                                                         // 2376
      query: {                                                                                                         // 2377
        hash: hash                                                                                                     // 2377
      },                                                                                                               // 2377
      update: {                                                                                                        // 2378
        $inc: {                                                                                                        // 2378
          refcount: 1                                                                                                  // 2378
        }                                                                                                              // 2378
      },                                                                                                               // 2378
      fields: {                                                                                                        // 2379
        _id: 1,                                                                                                        // 2379
        content: 1,                                                                                                    // 2379
        mimeType: 1                                                                                                    // 2379
      }                                                                                                                // 2379
    });                                                                                                                // 2376
                                                                                                                       //
    if (!result.ok) {                                                                                                  // 2382
      throw new Error("Couldn't increment refcount of asset with hash " + hash);                                       // 2383
    }                                                                                                                  // 2384
                                                                                                                       //
    var existing = result.value;                                                                                       // 2386
    return existing;                                                                                                   // 2387
  };                                                                                                                   // 2388
                                                                                                                       //
  SandstormDb.prototype.unrefStaticAsset = function (id) {                                                             // 2390
    // Decrement refcount on a static asset and delete if it has reached zero.                                         // 2391
    //                                                                                                                 // 2392
    // You must call this AFTER removing the reference from the DB, in case of failure between                         // 2393
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big                       // 2394
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                                         // 2395
    check(id, String);                                                                                                 // 2397
    var result = this.collections.staticAssets.findAndModify({                                                         // 2399
      query: {                                                                                                         // 2400
        _id: id                                                                                                        // 2400
      },                                                                                                               // 2400
      update: {                                                                                                        // 2401
        $inc: {                                                                                                        // 2401
          refcount: -1                                                                                                 // 2401
        }                                                                                                              // 2401
      },                                                                                                               // 2401
      fields: {                                                                                                        // 2402
        _id: 1,                                                                                                        // 2402
        refcount: 1                                                                                                    // 2402
      },                                                                                                               // 2402
      "new": true                                                                                                      // 2403
    });                                                                                                                // 2399
                                                                                                                       //
    if (!result.ok) {                                                                                                  // 2406
      throw new Error("Couldn't unref static asset " + id);                                                            // 2407
    }                                                                                                                  // 2408
                                                                                                                       //
    var existing = result.value;                                                                                       // 2410
                                                                                                                       //
    if (!existing) {                                                                                                   // 2411
      console.error(new Error("unrefStaticAsset() called on asset that doesn't exist").stack);                         // 2412
    } else if (existing.refcount <= 0) {                                                                               // 2413
      this.collections.staticAssets.remove({                                                                           // 2414
        _id: existing._id                                                                                              // 2414
      });                                                                                                              // 2414
    }                                                                                                                  // 2415
  };                                                                                                                   // 2416
                                                                                                                       //
  SandstormDb.prototype.getStaticAsset = function (id) {                                                               // 2418
    // Get a static asset's mimeType, encoding, and raw content.                                                       // 2419
    check(id, String);                                                                                                 // 2421
    var asset = this.collections.staticAssets.findOne(id, {                                                            // 2423
      fields: {                                                                                                        // 2423
        _id: 0,                                                                                                        // 2423
        mimeType: 1,                                                                                                   // 2423
        encoding: 1,                                                                                                   // 2423
        content: 1                                                                                                     // 2423
      }                                                                                                                // 2423
    });                                                                                                                // 2423
                                                                                                                       //
    if (asset) {                                                                                                       // 2424
      // TODO(perf): Mongo converts buffers to something else. Figure out a way to avoid a copy                        // 2425
      //   here.                                                                                                       // 2426
      asset.content = new Buffer(asset.content);                                                                       // 2427
    }                                                                                                                  // 2428
                                                                                                                       //
    return asset;                                                                                                      // 2430
  };                                                                                                                   // 2431
                                                                                                                       //
  SandstormDb.prototype.newAssetUpload = function (purpose) {                                                          // 2433
    check(purpose, Match.OneOf({                                                                                       // 2434
      profilePicture: {                                                                                                // 2435
        userId: DatabaseId                                                                                             // 2435
      }                                                                                                                // 2435
    }, {                                                                                                               // 2435
      loginLogo: {}                                                                                                    // 2436
    }));                                                                                                               // 2436
    return this.collections.assetUploadTokens.insert({                                                                 // 2439
      purpose: purpose,                                                                                                // 2440
      expires: new Date(Date.now() + 300000) // in 5 minutes                                                           // 2441
                                                                                                                       //
    });                                                                                                                // 2439
  };                                                                                                                   // 2443
                                                                                                                       //
  SandstormDb.prototype.fulfillAssetUpload = function (id) {                                                           // 2445
    // Indicates that the given asset upload has completed. It will be removed and its purpose                         // 2446
    // returned. If no matching upload exists, returns undefined.                                                      // 2447
    check(id, String);                                                                                                 // 2449
    var result = this.collections.assetUploadTokens.findAndModify({                                                    // 2451
      query: {                                                                                                         // 2452
        _id: id                                                                                                        // 2452
      },                                                                                                               // 2452
      remove: true                                                                                                     // 2453
    });                                                                                                                // 2451
                                                                                                                       //
    if (!result.ok) {                                                                                                  // 2456
      throw new Error("Failed to remove asset upload token");                                                          // 2457
    }                                                                                                                  // 2458
                                                                                                                       //
    var upload = result.value;                                                                                         // 2460
                                                                                                                       //
    if (upload.expires.valueOf() < Date.now()) {                                                                       // 2462
      return undefined; // already expired                                                                             // 2463
    } else {                                                                                                           // 2464
      return upload.purpose;                                                                                           // 2465
    }                                                                                                                  // 2466
  };                                                                                                                   // 2467
                                                                                                                       //
  SandstormDb.prototype.cleanupExpiredAssetUploads = function () {                                                     // 2469
    this.collections.assetUploadTokens.remove({                                                                        // 2470
      expires: {                                                                                                       // 2470
        $lt: Date.now()                                                                                                // 2470
      }                                                                                                                // 2470
    });                                                                                                                // 2470
  }; // TODO(cleanup): lift this out of the package so it can share with the ones in async-helpers.js                  // 2471
                                                                                                                       //
                                                                                                                       //
  var Future = Npm.require("fibers/future");                                                                           // 2474
                                                                                                                       //
  var promiseToFuture = function (promise) {                                                                           // 2475
    var result = new Future();                                                                                         // 2476
    promise.then(result.return.bind(result), result.throw.bind(result));                                               // 2477
    return result;                                                                                                     // 2478
  };                                                                                                                   // 2479
                                                                                                                       //
  var waitPromise = function (promise) {                                                                               // 2481
    return promiseToFuture(promise).wait();                                                                            // 2482
  };                                                                                                                   // 2483
                                                                                                                       //
  SandstormDb.prototype.deleteGrains = function (query, backend, type) {                                               // 2485
    var _this9 = this;                                                                                                 // 2485
                                                                                                                       //
    // Returns the number of grains deleted.                                                                           // 2486
    check(type, Match.OneOf("grain", "demoGrain"));                                                                    // 2488
    var numDeleted = 0;                                                                                                // 2490
    this.collections.grains.find(query).forEach(function (grain) {                                                     // 2491
      var user = Meteor.users.findOne(grain.userId);                                                                   // 2492
      waitPromise(backend.deleteGrain(grain._id, grain.userId));                                                       // 2494
      numDeleted += _this9.collections.grains.remove({                                                                 // 2495
        _id: grain._id                                                                                                 // 2495
      });                                                                                                              // 2495
                                                                                                                       //
      _this9.removeApiTokens({                                                                                         // 2496
        grainId: grain._id,                                                                                            // 2497
        $or: [{                                                                                                        // 2498
          owner: {                                                                                                     // 2499
            $exists: false                                                                                             // 2499
          }                                                                                                            // 2499
        }, {                                                                                                           // 2499
          owner: {                                                                                                     // 2500
            webkey: null                                                                                               // 2500
          }                                                                                                            // 2500
        }]                                                                                                             // 2500
      });                                                                                                              // 2496
                                                                                                                       //
      _this9.removeApiTokens({                                                                                         // 2504
        "owner.grain.grainId": grain._id                                                                               // 2504
      });                                                                                                              // 2504
                                                                                                                       //
      _this9.collections.activitySubscriptions.remove({                                                                // 2506
        grainId: grain._id                                                                                             // 2506
      });                                                                                                              // 2506
                                                                                                                       //
      if (grain.lastUsed) {                                                                                            // 2508
        var record = {                                                                                                 // 2509
          type: "grain",                                                                                               // 2510
          // Demo grains can never get here!                                                                           // 2510
          lastActive: grain.lastUsed,                                                                                  // 2511
          appId: grain.appId                                                                                           // 2512
        };                                                                                                             // 2509
                                                                                                                       //
        if (user && user.experiments) {                                                                                // 2514
          record.experiments = user.experiments;                                                                       // 2515
        }                                                                                                              // 2516
                                                                                                                       //
        _this9.collections.deleteStats.insert(record);                                                                 // 2518
      }                                                                                                                // 2519
                                                                                                                       //
      _this9.deleteUnusedPackages(grain.appId);                                                                        // 2521
                                                                                                                       //
      if (grain.size) {                                                                                                // 2523
        Meteor.users.update(grain.userId, {                                                                            // 2524
          $inc: {                                                                                                      // 2524
            storageUsage: -grain.size                                                                                  // 2524
          }                                                                                                            // 2524
        });                                                                                                            // 2524
      }                                                                                                                // 2525
    });                                                                                                                // 2526
    return numDeleted;                                                                                                 // 2527
  };                                                                                                                   // 2528
                                                                                                                       //
  SandstormDb.prototype.userGrainTitle = function (grainId, accountId, obsolete) {                                     // 2530
    check(grainId, String);                                                                                            // 2531
    check(accountId, Match.OneOf(String, undefined, null));                                                            // 2532
    check(obsolete, undefined);                                                                                        // 2533
    var grain = this.getGrain(grainId);                                                                                // 2535
                                                                                                                       //
    if (!grain) {                                                                                                      // 2536
      throw new Error("called userGrainTitle() for a grain that doesn't exist");                                       // 2537
    }                                                                                                                  // 2538
                                                                                                                       //
    var title = grain.title;                                                                                           // 2540
                                                                                                                       //
    if (grain.userId !== accountId) {                                                                                  // 2541
      var sharerToken = this.collections.apiTokens.findOne({                                                           // 2542
        grainId: grainId,                                                                                              // 2543
        "owner.user.accountId": accountId                                                                              // 2544
      }, {                                                                                                             // 2542
        sort: {                                                                                                        // 2546
          lastUsed: -1                                                                                                 // 2547
        }                                                                                                              // 2546
      });                                                                                                              // 2545
                                                                                                                       //
      if (sharerToken) {                                                                                               // 2550
        title = sharerToken.owner.user.title;                                                                          // 2551
      } else {                                                                                                         // 2552
        title = "shared grain";                                                                                        // 2553
      }                                                                                                                // 2554
    }                                                                                                                  // 2555
                                                                                                                       //
    return title;                                                                                                      // 2557
  };                                                                                                                   // 2558
                                                                                                                       //
  var packageCache = {}; // Package info is immutable. Let's cache to save on mongo queries.                           // 2560
                                                                                                                       //
  SandstormDb.prototype.getPackage = function (packageId) {                                                            // 2563
    // Get the given package record. Since package info is immutable, cache the data in the server                     // 2564
    // to reduce mongo query overhead, since it turns out we have to fetch specific packages a                         // 2565
    // lot.                                                                                                            // 2566
    if (packageId in packageCache) {                                                                                   // 2568
      return packageCache[packageId];                                                                                  // 2569
    }                                                                                                                  // 2570
                                                                                                                       //
    var pkg = this.collections.packages.findOne(packageId);                                                            // 2572
                                                                                                                       //
    if (pkg && pkg.status === "ready") {                                                                               // 2573
      packageCache[packageId] = pkg;                                                                                   // 2574
    }                                                                                                                  // 2575
                                                                                                                       //
    return pkg;                                                                                                        // 2577
  };                                                                                                                   // 2578
                                                                                                                       //
  SandstormDb.prototype.deleteUnusedPackages = function (appId) {                                                      // 2580
    var _this10 = this;                                                                                                // 2580
                                                                                                                       //
    check(appId, String);                                                                                              // 2581
    this.collections.packages.find({                                                                                   // 2582
      appId: appId                                                                                                     // 2582
    }).forEach(function (pkg) {                                                                                        // 2582
      // Mark package for possible deletion;                                                                           // 2583
      _this10.collections.packages.update({                                                                            // 2584
        _id: pkg._id,                                                                                                  // 2584
        status: "ready"                                                                                                // 2584
      }, {                                                                                                             // 2584
        $set: {                                                                                                        // 2584
          shouldCleanup: true                                                                                          // 2584
        }                                                                                                              // 2584
      });                                                                                                              // 2584
    });                                                                                                                // 2585
  };                                                                                                                   // 2586
                                                                                                                       //
  SandstormDb.prototype.sendAppUpdateNotifications = function (appId, packageId, name, versionNumber, marketingVersion) {
    var _this11 = this;                                                                                                // 2589
                                                                                                                       //
    var actions = this.collections.userActions.find({                                                                  // 2590
      appId: appId,                                                                                                    // 2590
      appVersion: {                                                                                                    // 2590
        $lt: versionNumber                                                                                             // 2590
      }                                                                                                                // 2590
    }, {                                                                                                               // 2590
      fields: {                                                                                                        // 2591
        userId: 1                                                                                                      // 2591
      }                                                                                                                // 2591
    });                                                                                                                // 2591
    actions.forEach(function (action) {                                                                                // 2592
      var userId = action.userId;                                                                                      // 2593
      var updater = {                                                                                                  // 2594
        timestamp: new Date(),                                                                                         // 2595
        isUnread: true                                                                                                 // 2596
      };                                                                                                               // 2594
                                                                                                                       //
      var inserter = _.extend({                                                                                        // 2598
        userId: userId,                                                                                                // 2598
        appUpdates: {}                                                                                                 // 2598
      }, updater); // Set only the appId that we care about. Use mongo's dot notation to specify only a single         // 2598
      // field inside of an object to update                                                                           // 2601
                                                                                                                       //
                                                                                                                       //
      inserter.appUpdates[appId] = updater["appUpdates." + appId] = {                                                  // 2602
        marketingVersion: marketingVersion,                                                                            // 2603
        packageId: packageId,                                                                                          // 2604
        name: name,                                                                                                    // 2605
        version: versionNumber                                                                                         // 2606
      }; // We unfortunately cannot upsert because upserts can only have field equality conditions in                  // 2602
      // the query. If we try to upsert, Mongo complaints that "$exists" isn't valid to store.                         // 2610
                                                                                                                       //
      if (_this11.collections.notifications.update({                                                                   // 2611
        userId: userId,                                                                                                // 2612
        appUpdates: {                                                                                                  // 2612
          $exists: true                                                                                                // 2612
        }                                                                                                              // 2612
      }, {                                                                                                             // 2612
        $set: updater                                                                                                  // 2613
      }) == 0) {                                                                                                       // 2613
        // Update failed; try an insert instead.                                                                       // 2614
        _this11.collections.notifications.insert(inserter);                                                            // 2615
      }                                                                                                                // 2616
    });                                                                                                                // 2617
    this.collections.appIndex.update({                                                                                 // 2619
      _id: appId                                                                                                       // 2619
    }, {                                                                                                               // 2619
      $set: {                                                                                                          // 2619
        hasSentNotifications: true                                                                                     // 2619
      }                                                                                                                // 2619
    }); // In the case where we replaced a previous notification and that was the only reference to the                // 2619
    // package, we need to clean it up                                                                                 // 2622
                                                                                                                       //
    this.deleteUnusedPackages(appId);                                                                                  // 2623
  };                                                                                                                   // 2624
                                                                                                                       //
  SandstormDb.prototype.sendReferralProgramNotification = function (userId) {                                          // 2626
    this.collections.notifications.upsert({                                                                            // 2627
      userId: userId,                                                                                                  // 2628
      referral: true                                                                                                   // 2629
    }, {                                                                                                               // 2627
      userId: userId,                                                                                                  // 2631
      referral: true,                                                                                                  // 2632
      timestamp: new Date(),                                                                                           // 2633
      isUnread: true                                                                                                   // 2634
    });                                                                                                                // 2630
  };                                                                                                                   // 2636
                                                                                                                       //
  SandstormDb.prototype.upgradeGrains = function (appId, version, packageId, backend) {                                // 2638
    check(appId, String);                                                                                              // 2639
    check(version, Match.Integer);                                                                                     // 2640
    check(packageId, String);                                                                                          // 2641
    var selector = {                                                                                                   // 2643
      userId: Meteor.userId(),                                                                                         // 2644
      appId: appId,                                                                                                    // 2645
      appVersion: {                                                                                                    // 2646
        $lte: version                                                                                                  // 2646
      },                                                                                                               // 2646
      packageId: {                                                                                                     // 2647
        $ne: packageId                                                                                                 // 2647
      }                                                                                                                // 2647
    };                                                                                                                 // 2643
    this.collections.grains.find(selector).forEach(function (grain) {                                                  // 2650
      backend.shutdownGrain(grain._id, grain.userId);                                                                  // 2651
    });                                                                                                                // 2652
    this.collections.grains.update(selector, {                                                                         // 2654
      $set: {                                                                                                          // 2655
        appVersion: version,                                                                                           // 2655
        packageId: packageId,                                                                                          // 2655
        packageSalt: Random.secret()                                                                                   // 2655
      }                                                                                                                // 2655
    }, {                                                                                                               // 2654
      multi: true                                                                                                      // 2656
    });                                                                                                                // 2656
  };                                                                                                                   // 2657
                                                                                                                       //
  SandstormDb.prototype.startInstall = function (packageId, url, retryFailed, isAutoUpdated) {                         // 2659
    // Mark package for possible installation.                                                                         // 2660
    var fields = {                                                                                                     // 2662
      status: "download",                                                                                              // 2663
      progress: 0,                                                                                                     // 2664
      url: url,                                                                                                        // 2665
      isAutoUpdated: !!isAutoUpdated                                                                                   // 2666
    };                                                                                                                 // 2662
                                                                                                                       //
    if (retryFailed) {                                                                                                 // 2669
      this.collections.packages.update({                                                                               // 2670
        _id: packageId,                                                                                                // 2670
        status: "failed"                                                                                               // 2670
      }, {                                                                                                             // 2670
        $set: fields                                                                                                   // 2670
      });                                                                                                              // 2670
    } else {                                                                                                           // 2671
      try {                                                                                                            // 2672
        fields._id = packageId;                                                                                        // 2673
        this.collections.packages.insert(fields);                                                                      // 2674
      } catch (err) {                                                                                                  // 2675
        console.error("Simultaneous startInstall()s?", err.stack);                                                     // 2676
      }                                                                                                                // 2677
    }                                                                                                                  // 2678
  };                                                                                                                   // 2679
                                                                                                                       //
  var ValidKeyFingerprint = Match.Where(function (keyFingerprint) {                                                    // 2681
    check(keyFingerprint, String);                                                                                     // 2682
    return !!keyFingerprint.match(/^[0-9A-F]{40}$/);                                                                   // 2683
  });                                                                                                                  // 2684
                                                                                                                       //
  SandstormDb.prototype.updateKeybaseProfileAsync = function (keyFingerprint) {                                        // 2686
    var _this12 = this;                                                                                                // 2686
                                                                                                                       //
    // Asynchronously fetch the given Keybase profile and populate the KeybaseProfiles collection.                     // 2687
    check(keyFingerprint, ValidKeyFingerprint);                                                                        // 2689
    console.log("fetching keybase", keyFingerprint);                                                                   // 2691
    HTTP.get("https://keybase.io/_/api/1.0/user/lookup.json?key_fingerprint=" + keyFingerprint + "&fields=basics,profile,proofs_summary", {
      timeout: 5000                                                                                                    // 2696
    }, function (err, keybaseResponse) {                                                                               // 2695
      if (err) {                                                                                                       // 2698
        console.log("keybase lookup error:", err.stack);                                                               // 2699
        return;                                                                                                        // 2700
      }                                                                                                                // 2701
                                                                                                                       //
      if (!keybaseResponse.data) {                                                                                     // 2703
        console.log("keybase didn't return JSON? Headers:", keybaseResponse.headers);                                  // 2704
        return;                                                                                                        // 2705
      }                                                                                                                // 2706
                                                                                                                       //
      var profile = (keybaseResponse.data.them || [])[0];                                                              // 2708
                                                                                                                       //
      if (profile) {                                                                                                   // 2710
        // jscs:disable requireCamelCaseOrUpperCaseIdentifiers                                                         // 2711
        var record = {                                                                                                 // 2712
          displayName: (profile.profile || {}).full_name,                                                              // 2713
          handle: (profile.basics || {}).username,                                                                     // 2714
          proofs: (profile.proofs_summary || {}).all || []                                                             // 2715
        }; // jscs:enable requireCamelCaseOrUpperCaseIdentifiers                                                       // 2712
                                                                                                                       //
        record.proofs.forEach(function (proof) {                                                                       // 2719
          // Remove potentially Mongo-incompatible stuff. (Currently Keybase returns nothing that                      // 2720
          // this would filter.)                                                                                       // 2721
          for (var field in meteorBabelHelpers.sanitizeForInObject(proof)) {                                           // 2722
            // Don't allow field names containing '.' or '$'. Also don't allow sub-objects mainly                      // 2723
            // because I'm too lazy to check the field names recursively (and Keybase doesn't                          // 2724
            // return any objects anyway).                                                                             // 2725
            if (field.match(/[.$]/) || (0, _typeof3.default)(proof[field]) === "object") {                             // 2726
              delete proof[field];                                                                                     // 2727
            }                                                                                                          // 2728
          } // Indicate not verified.                                                                                  // 2729
          // TODO(security): Asynchronously verify proofs. Presumably we can borrow code from the                      // 2732
          //   Keybase node-based CLI.                                                                                 // 2733
                                                                                                                       //
                                                                                                                       //
          proof.status = "unverified";                                                                                 // 2734
        });                                                                                                            // 2735
                                                                                                                       //
        _this12.collections.keybaseProfiles.update(keyFingerprint, {                                                   // 2737
          $set: record                                                                                                 // 2737
        }, {                                                                                                           // 2737
          upsert: true                                                                                                 // 2737
        });                                                                                                            // 2737
      } else {                                                                                                         // 2738
        // Keybase reports no match, so remove what we know of this user. We don't want to remove                      // 2739
        // the item entirely from the cache as this will cause us to repeatedly re-fetch the data                      // 2740
        // from Keybase.                                                                                               // 2741
        //                                                                                                             // 2742
        // TODO(someday): We could perhaps keep the proofs if we can still verify them directly,                       // 2743
        //   but at present we don't have the ability to verify proofs.                                                // 2744
        _this12.collections.keybaseProfiles.update(keyFingerprint, {                                                   // 2745
          $unset: {                                                                                                    // 2746
            displayName: "",                                                                                           // 2746
            handle: "",                                                                                                // 2746
            proofs: ""                                                                                                 // 2746
          }                                                                                                            // 2746
        }, {                                                                                                           // 2746
          upsert: true                                                                                                 // 2746
        });                                                                                                            // 2746
      }                                                                                                                // 2747
    });                                                                                                                // 2748
  };                                                                                                                   // 2749
                                                                                                                       //
  SandstormDb.prototype.deleteUnusedAccount = function (backend, credentialId) {                                       // 2751
    // If there is an *unused* account that has `credentialId` as a login credential, deletes it.                      // 2752
    check(credentialId, String);                                                                                       // 2754
    var account = this.collections.users.findOne({                                                                     // 2755
      "loginCredentials.id": credentialId                                                                              // 2755
    });                                                                                                                // 2755
                                                                                                                       //
    if (account && account.loginCredentials.length == 1 && account.nonloginCredentials.length == 0 && !this.collections.grains.findOne({
      userId: account._id                                                                                              // 2759
    }) && !this.collections.apiTokens.findOne({                                                                        // 2759
      accountId: account._id                                                                                           // 2760
    }) && (!account.plan || account.plan === "free") && !(account.payments && account.payments.id) && !this.collections.contacts.findOne({
      ownerId: account._id                                                                                             // 2763
    })) {                                                                                                              // 2763
      this.collections.users.remove({                                                                                  // 2764
        _id: account._id                                                                                               // 2764
      });                                                                                                              // 2764
      backend.deleteUser(account._id);                                                                                 // 2765
    }                                                                                                                  // 2766
  };                                                                                                                   // 2767
                                                                                                                       //
  Meteor.publish("keybaseProfile", function (keyFingerprint) {                                                         // 2769
    check(keyFingerprint, ValidKeyFingerprint);                                                                        // 2770
    var db = this.connection.sandstormDb;                                                                              // 2771
    var cursor = db.collections.keybaseProfiles.find(keyFingerprint);                                                  // 2773
                                                                                                                       //
    if (cursor.count() === 0) {                                                                                        // 2774
      // Fire off async update.                                                                                        // 2775
      db.updateKeybaseProfileAsync(keyFingerprint);                                                                    // 2776
    }                                                                                                                  // 2777
                                                                                                                       //
    return cursor;                                                                                                     // 2779
  });                                                                                                                  // 2780
  Meteor.publish("appIndex", function (appId) {                                                                        // 2782
    check(appId, String);                                                                                              // 2783
    var db = this.connection.sandstormDb;                                                                              // 2784
    var cursor = db.collections.appIndex.find({                                                                        // 2785
      _id: appId                                                                                                       // 2785
    });                                                                                                                // 2785
    return cursor;                                                                                                     // 2786
  });                                                                                                                  // 2787
  Meteor.publish("userPackages", function () {                                                                         // 2789
    var _this13 = this;                                                                                                // 2789
                                                                                                                       //
    // Users should be able to see packages that are either:                                                           // 2790
    // 1. referenced by one of their userActions                                                                       // 2791
    // 2. referenced by one of their grains                                                                            // 2792
    var db = this.connection.sandstormDb; // Note that package information, once it is in the database, is static. There's no need to
    // reactively subscribe to changes to a package since they don't change. It's also unecessary                      // 2796
    // to reactively remove a package from the client side when it is removed on the server, or                        // 2797
    // when the client stops using it, because the worst case is the client has a small amount                         // 2798
    // of extra info on a no-longer-used package held in memory until they refresh Sandstorm.                          // 2799
    // So, we implement this as a cache: the first time each package ID shows up among the user's                      // 2800
    // stuff, we push the package info to the client, and then we never update it.                                     // 2801
    //                                                                                                                 // 2802
    // Alternatively, we could subscribe to each individual package query, but this would waste                        // 2803
    // lots of server-side resources watching for events that will never happen or don't matter.                       // 2804
                                                                                                                       //
    var hasPackage = {};                                                                                               // 2805
                                                                                                                       //
    var refPackage = function (packageId) {                                                                            // 2806
      // Ignore dev apps.                                                                                              // 2807
      if (packageId.lastIndexOf("dev-", 0) === 0) return;                                                              // 2808
                                                                                                                       //
      if (!hasPackage[packageId]) {                                                                                    // 2810
        hasPackage[packageId] = true;                                                                                  // 2811
        var pkg = db.getPackage(packageId);                                                                            // 2812
                                                                                                                       //
        if (pkg) {                                                                                                     // 2813
          _this13.added("packages", packageId, pkg);                                                                   // 2814
        } else {                                                                                                       // 2815
          console.error("shouldn't happen: missing package referenced by user's stuff:", packageId);                   // 2816
        }                                                                                                              // 2818
      }                                                                                                                // 2819
    }; // package source 1: packages referred to by actions                                                            // 2820
                                                                                                                       //
                                                                                                                       //
    var actions = db.userActions(this.userId);                                                                         // 2823
    var actionsHandle = actions.observe({                                                                              // 2824
      added: function (newAction) {                                                                                    // 2825
        refPackage(newAction.packageId);                                                                               // 2826
      },                                                                                                               // 2827
      changed: function (newAction, oldAction) {                                                                       // 2829
        refPackage(newAction.packageId);                                                                               // 2830
      }                                                                                                                // 2831
    }); // package source 2: packages referred to by grains directly                                                   // 2824
                                                                                                                       //
    var grains = db.userGrains(this.userId, {                                                                          // 2835
      includeTrash: true                                                                                               // 2835
    });                                                                                                                // 2835
    var grainsHandle = grains.observe({                                                                                // 2836
      added: function (newGrain) {                                                                                     // 2837
        // Watch out: DevApp grains can lack a packageId.                                                              // 2838
        if (newGrain.packageId) {                                                                                      // 2839
          refPackage(newGrain.packageId);                                                                              // 2840
        }                                                                                                              // 2841
      },                                                                                                               // 2842
      changed: function (newGrain, oldGrain) {                                                                         // 2844
        // Watch out: DevApp grains can lack a packageId.                                                              // 2845
        if (newGrain.packageId) {                                                                                      // 2846
          refPackage(newGrain.packageId);                                                                              // 2847
        }                                                                                                              // 2848
      }                                                                                                                // 2849
    });                                                                                                                // 2836
    this.onStop(function () {                                                                                          // 2852
      actionsHandle.stop();                                                                                            // 2853
      grainsHandle.stop();                                                                                             // 2854
    });                                                                                                                // 2855
    this.ready();                                                                                                      // 2857
  });                                                                                                                  // 2858
                                                                                                                       //
  SandstormDb.generateIdentityId = function () {                                                                       // 2860
    return _Crypto2.randomBytes(32).toString("hex");                                                                   // 2861
  };                                                                                                                   // 2862
                                                                                                                       //
  SandstormDb.prototype.getOrGenerateIdentityId = function (accountId, grain) {                                        // 2864
    // Determine the identity ID by which the given user is known within the given grain. May                          // 2865
    // generate a new ID if the user doesn't currently have access to the grain.                                       // 2866
    check(accountId, String);                                                                                          // 2868
    check(grain, Match.ObjectIncluding({                                                                               // 2869
      _id: String,                                                                                                     // 2869
      userId: String,                                                                                                  // 2869
      identityId: String                                                                                               // 2869
    }));                                                                                                               // 2869
                                                                                                                       //
    if (accountId == grain.userId) {                                                                                   // 2871
      return grain.identityId;                                                                                         // 2872
    } else {                                                                                                           // 2873
      // Check if the user has already been introduced to this grain.                                                  // 2874
      var existingToken = this.collections.apiTokens.findOne({                                                         // 2875
        "grainId": grain._id,                                                                                          // 2876
        "owner.user.accountId": accountId,                                                                             // 2877
        "owner.user.identityId": {                                                                                     // 2878
          $exists: true                                                                                                // 2878
        }                                                                                                              // 2878
      }, {                                                                                                             // 2876
        fields: {                                                                                                      // 2879
          "owner.user.identityId": 1                                                                                   // 2879
        }                                                                                                              // 2879
      });                                                                                                              // 2879
                                                                                                                       //
      if (existingToken) {                                                                                             // 2881
        // This user already has a token associated with this grain. Reuse the identity ID.                            // 2882
        // TODO(someday): It's a bit awkward that if the user deletes all their tokens for a grain,                    // 2883
        //   we forget their identity ID. This both means that if the user regains access, they'll                     // 2884
        //   have a new identity in the grain, and it means that we have no way of enumerating all                     // 2885
        //   identity IDs the grain has ever seen (including forgotten ones), which means we can't                     // 2886
        //   give the grain owner a way to remap these identities when needed. Perhaps ApiTokens                       // 2887
        //   should never really be deleted, only hidden?                                                              // 2888
        return existingToken.owner.user.identityId;                                                                    // 2889
      }                                                                                                                // 2890
                                                                                                                       //
      if (!grain.private) {                                                                                            // 2892
        // This grain operates on the old sharing model, where simply knowing the grain ID is                          // 2893
        // sufficient to open it. We need to assign identity IDs in a consistent way without having                    // 2894
        // stored them anywhere. Identity IDs used to be based on credential IDs, which at some                        // 2895
        // point were used to fill in identicon keys in profiles... so use the identicon.                              // 2896
        var user = Meteor.users.findOne(accountId);                                                                    // 2897
                                                                                                                       //
        if (user && user.profile && user.profile.identicon) {                                                          // 2898
          return user.profile.identicon;                                                                               // 2899
        } else {                                                                                                       // 2900
          throw new Meteor.Error(500, "Don't know how to identity user under old sharing model.");                     // 2901
        }                                                                                                              // 2902
      } // This user is new to this grain. Give them a freshly-generated identity ID.                                  // 2903
      // TODO(cleanup): We only ever pass the first 16 bytes of the identity ID to the app. Maybe                      // 2906
      //   we can truncate all identity IDs to 16 chars?                                                               // 2907
                                                                                                                       //
                                                                                                                       //
      return SandstormDb.generateIdentityId();                                                                         // 2908
    }                                                                                                                  // 2909
  };                                                                                                                   // 2910
}                                                                                                                      // 2911
                                                                                                                       //
if (Meteor.isServer) {                                                                                                 // 2913
  SandstormDb.prototype.deleteCredential = function (credentialId) {                                                   // 2914
    check(credentialId, String);                                                                                       // 2915
    Meteor.users.remove({                                                                                              // 2917
      _id: credentialId                                                                                                // 2917
    });                                                                                                                // 2917
  };                                                                                                                   // 2918
                                                                                                                       //
  SandstormDb.prototype.deleteAccount = function (userId, backend) {                                                   // 2920
    check(userId, String);                                                                                             // 2921
                                                                                                                       //
    var _this = this;                                                                                                  // 2923
                                                                                                                       //
    var user = Meteor.users.findOne({                                                                                  // 2924
      _id: userId                                                                                                      // 2924
    });                                                                                                                // 2924
    this.deleteGrains({                                                                                                // 2925
      userId: userId                                                                                                   // 2925
    }, backend, "grain");                                                                                              // 2925
    this.removeApiTokens({                                                                                             // 2926
      "owner.user.accountId": userId                                                                                   // 2926
    });                                                                                                                // 2926
    this.collections.userActions.remove({                                                                              // 2927
      userId: userId                                                                                                   // 2927
    });                                                                                                                // 2927
    this.collections.notifications.remove({                                                                            // 2928
      userId: userId                                                                                                   // 2928
    });                                                                                                                // 2928
    user.loginCredentials.forEach(function (credential) {                                                              // 2929
      if (Meteor.users.find({                                                                                          // 2930
        $or: [{                                                                                                        // 2930
          "loginCredentials.id": credential.id                                                                         // 2931
        }, {                                                                                                           // 2931
          "nonloginCredentials.id": credential.id                                                                      // 2932
        }]                                                                                                             // 2932
      }).count() === 1) {                                                                                              // 2930
        // If this is the only account with the credential, then delete it                                             // 2934
        _this.deleteCredential(credential.id);                                                                         // 2935
      }                                                                                                                // 2936
    });                                                                                                                // 2937
    user.nonloginCredentials.forEach(function (credential) {                                                           // 2938
      if (Meteor.users.find({                                                                                          // 2939
        $or: [{                                                                                                        // 2939
          "loginCredentials.id": credential.id                                                                         // 2940
        }, {                                                                                                           // 2940
          "nonloginCredentials.id": credential.id                                                                      // 2941
        }]                                                                                                             // 2941
      }).count() === 1) {                                                                                              // 2939
        // If this is the only account with the credential, then delete it                                             // 2943
        _this.deleteCredential(credential.id);                                                                         // 2944
      }                                                                                                                // 2945
    });                                                                                                                // 2946
    this.collections.contacts.remove({                                                                                 // 2947
      accountId: userId                                                                                                // 2947
    });                                                                                                                // 2947
    this.collections.contacts.remove({                                                                                 // 2948
      ownerId: userId                                                                                                  // 2948
    });                                                                                                                // 2948
    backend.deleteUser(userId);                                                                                        // 2949
    Meteor.users.remove({                                                                                              // 2950
      _id: userId                                                                                                      // 2950
    });                                                                                                                // 2950
  };                                                                                                                   // 2951
}                                                                                                                      // 2952
                                                                                                                       //
Meteor.methods({                                                                                                       // 2954
  addUserActions: function (packageId) {                                                                               // 2955
    check(packageId, String);                                                                                          // 2956
                                                                                                                       //
    if (!this.userId || !Meteor.user().loginCredentials || !isSignedUpOrDemo()) {                                      // 2957
      throw new Meteor.Exception(403, "Must be logged in as a non-guest to add app actions.");                         // 2958
    }                                                                                                                  // 2959
                                                                                                                       //
    if (this.isSimulation) {                                                                                           // 2961
      // TODO(cleanup): Appdemo code relies on this being simulated client-side but we don't have                      // 2962
      //   a proper DB object to use.                                                                                  // 2963
      new SandstormDb().addUserActions(this.userId, packageId, true);                                                  // 2964
    } else {                                                                                                           // 2965
      this.connection.sandstormDb.addUserActions(this.userId, packageId);                                              // 2966
    }                                                                                                                  // 2967
  },                                                                                                                   // 2968
  removeUserAction: function (actionId) {                                                                              // 2970
    check(actionId, String);                                                                                           // 2971
                                                                                                                       //
    if (this.isSimulation) {                                                                                           // 2972
      UserActions.remove({                                                                                             // 2973
        _id: actionId                                                                                                  // 2973
      });                                                                                                              // 2973
    } else {                                                                                                           // 2974
      if (this.userId) {                                                                                               // 2975
        var result = this.connection.sandstormDb.collections.userActions.findAndModify({                               // 2976
          query: {                                                                                                     // 2977
            _id: actionId,                                                                                             // 2977
            userId: this.userId                                                                                        // 2977
          },                                                                                                           // 2977
          remove: true                                                                                                 // 2978
        });                                                                                                            // 2976
                                                                                                                       //
        if (!result.ok) {                                                                                              // 2981
          throw new Error("Couldn't remove user action " + actionId);                                                  // 2982
        }                                                                                                              // 2983
                                                                                                                       //
        var action = result.value;                                                                                     // 2985
                                                                                                                       //
        if (action) {                                                                                                  // 2986
          this.connection.sandstormDb.deleteUnusedPackages(action.appId);                                              // 2987
        }                                                                                                              // 2988
      }                                                                                                                // 2989
    }                                                                                                                  // 2990
  }                                                                                                                    // 2991
});                                                                                                                    // 2954
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"profile.js":function(require){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// packages/sandstorm-db/profile.js                                                                                    //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var makeIdenticon = void 0;                                                                                            // 17
var httpProtocol = void 0;                                                                                             // 18
                                                                                                                       //
if (Meteor.isServer) {                                                                                                 // 20
  SandstormDb.ensureSubscriberHasCredential = function (publishHandler, credentialId) {                                // 21
    // Helper for publish functions that need to restrict access based on whether the subscriber                       // 22
    // has a given credential linked. Automatically stops the subscription if the user loses the                       // 23
    // credential. Returns a boolean indicating whether the user initially has the credential.                         // 24
    var userId = publishHandler.userId;                                                                                // 26
                                                                                                                       //
    if (userId === credentialId) {                                                                                     // 27
      return true;                                                                                                     // 28
    } else {                                                                                                           // 29
      var hasCredentialCursor = Meteor.users.find({                                                                    // 30
        $or: [{                                                                                                        // 31
          _id: userId,                                                                                                 // 33
          "loginCredentials.id": credentialId                                                                          // 34
        }, {                                                                                                           // 32
          _id: userId,                                                                                                 // 37
          "nonloginCredentials.id": credentialId                                                                       // 38
        }]                                                                                                             // 36
      });                                                                                                              // 30
                                                                                                                       //
      if (hasCredentialCursor.count() == 0) {                                                                          // 42
        publishHandler.stop();                                                                                         // 43
        return false;                                                                                                  // 44
      }                                                                                                                // 45
                                                                                                                       //
      var handle = hasCredentialCursor.observe({                                                                       // 47
        removed: function () {                                                                                         // 48
          publishHandler.stop();                                                                                       // 48
        }                                                                                                              // 48
      });                                                                                                              // 48
      publishHandler.onStop(function () {                                                                              // 50
        handle.stop();                                                                                                 // 50
      });                                                                                                              // 50
      return true;                                                                                                     // 52
    }                                                                                                                  // 53
  };                                                                                                                   // 54
                                                                                                                       //
  Meteor.publish("credentialDetails", function (credentialId) {                                                        // 56
    check(credentialId, String);                                                                                       // 57
    if (!SandstormDb.ensureSubscriberHasCredential(this, credentialId)) return;                                        // 58
    return Meteor.users.find({                                                                                         // 60
      _id: credentialId                                                                                                // 60
    }, {                                                                                                               // 60
      fields: {                                                                                                        // 61
        type: 1,                                                                                                       // 62
        unverifiedEmail: 1,                                                                                            // 63
        expires: 1,                                                                                                    // 64
        createdAt: 1,                                                                                                  // 65
        "services.dev.name": 1,                                                                                        // 67
        "services.demo": 1,                                                                                            // 68
        "services.google.id": 1,                                                                                       // 70
        "services.google.email": 1,                                                                                    // 71
        "services.google.verified_email": 1,                                                                           // 72
        "services.google.name": 1,                                                                                     // 73
        "services.google.picture": 1,                                                                                  // 74
        "services.google.gender": 1,                                                                                   // 75
        "services.google.hd": 1,                                                                                       // 76
        "services.github.id": 1,                                                                                       // 78
        "services.github.email": 1,                                                                                    // 79
        "services.github.emails": 1,                                                                                   // 80
        "services.github.username": 1,                                                                                 // 81
        "services.email.email": 1,                                                                                     // 83
        "services.ldap.id": 1,                                                                                         // 85
        "services.ldap.username": 1,                                                                                   // 86
        "services.ldap.rawAttrs": 1,                                                                                   // 87
        "services.saml.id": 1,                                                                                         // 89
        "services.saml.email": 1,                                                                                      // 90
        "services.saml.displayName": 1                                                                                 // 91
      }                                                                                                                // 61
    });                                                                                                                // 61
  }), Meteor.publish("accountCredentials", function () {                                                               // 94
    // Maybe this should be folded into the "credentials" subscription?                                                // 97
    if (!this.userId) return [];                                                                                       // 99
    return Meteor.users.find({                                                                                         // 101
      _id: this.userId                                                                                                 // 102
    }, {                                                                                                               // 102
      fields: {                                                                                                        // 103
        type: 1,                                                                                                       // 104
        profile: 1,                                                                                                    // 105
        verifiedEmail: 1,                                                                                              // 106
        loginCredentials: 1,                                                                                           // 107
        nonloginCredentials: 1,                                                                                        // 108
        expires: 1,                                                                                                    // 109
        primaryEmail: 1                                                                                                // 110
      }                                                                                                                // 103
    });                                                                                                                // 103
  });                                                                                                                  // 112
                                                                                                                       //
  makeIdenticon = function (id) {                                                                                      // 114
    var hash = id.slice(0, 32);                                                                                        // 115
    return httpProtocol + "//" + makeWildcardHost("static") + "/identicon/" + hash + "?s=256";                         // 116
  };                                                                                                                   // 117
                                                                                                                       //
  var Url = Npm.require("url");                                                                                        // 119
                                                                                                                       //
  httpProtocol = Url.parse(process.env.ROOT_URL).protocol;                                                             // 120
} else {                                                                                                               // 121
  var identiconCache = {};                                                                                             // 122
                                                                                                                       //
  makeIdenticon = function (id) {                                                                                      // 124
    // Given a cryptographic hash as input, generate an identicon.                                                     // 125
    // identicon.js doesn't use more than 32 digits of the hash even if we provide it, but slice it                    // 127
    // anyway to guarantee this, since historically we sliced it for reasons now obsolete.                             // 128
    var hash = id.slice(0, 32);                                                                                        // 129
                                                                                                                       //
    if (hash in identiconCache) {                                                                                      // 131
      return identiconCache[hash];                                                                                     // 132
    }                                                                                                                  // 133
                                                                                                                       //
    var data = new Identicon(hash, 512).toString();                                                                    // 135
    var result = "data:image/svg+xml," + encodeURIComponent(data);                                                     // 136
    identiconCache[hash] = result;                                                                                     // 137
    return result;                                                                                                     // 138
  };                                                                                                                   // 139
                                                                                                                       //
  httpProtocol = window.location.protocol;                                                                             // 141
}                                                                                                                      // 142
                                                                                                                       //
var GENDERS = {                                                                                                        // 144
  male: "male",                                                                                                        // 144
  female: "female",                                                                                                    // 144
  neutral: "neutral",                                                                                                  // 144
  robot: "robot"                                                                                                       // 144
};                                                                                                                     // 144
                                                                                                                       //
function staticAssetUrl(id, staticHost) {                                                                              // 146
  if (id) {                                                                                                            // 147
    return staticHost + "/" + id;                                                                                      // 148
  } else {                                                                                                             // 149
    return undefined;                                                                                                  // 150
  }                                                                                                                    // 151
}                                                                                                                      // 152
                                                                                                                       //
function filterHandle(handle) {                                                                                        // 154
  // Convert disallowed letters into underscores, but no more than one underscore in a row.                            // 155
  return handle && handle.toLowerCase().split(/[^a-z0-9_]+/g).join("_");                                               // 156
}                                                                                                                      // 157
                                                                                                                       //
function emailToHandle(email) {                                                                                        // 159
  // Turn an email address into a handle.                                                                              // 160
  if (!email) return undefined; // Use the stuff before the @. Ignore stuff after '+' because it's commonly used for filters
  // on the same account.                                                                                              // 165
                                                                                                                       //
  var parts = email.split("@");                                                                                        // 166
  var base = filterHandle(parts[0].split("+")[0]);                                                                     // 167
  var domain = (parts[1] || "").split(".");                                                                            // 169
                                                                                                                       //
  if (domain[domain.length - 1] === "name") {                                                                          // 170
    // Oh, a .name domain. Let's use                                                                                   // 171
    base = filterHandle(domain.slice(0, domain.length - 1).join("."));                                                 // 172
  } else if (_.contains(["me", "self", "contact", "admin", "administrator", "root", "info", "sandstorm", "sandstormio", "inbox", "indiegogo", "mail", "email"], base)) {
    // This is probably an address at a vanity domain. Use the domain itself as the handle.                            // 179
    base = filterHandle(domain[0]);                                                                                    // 180
  }                                                                                                                    // 181
                                                                                                                       //
  return filterHandle(base);                                                                                           // 183
}                                                                                                                      // 184
                                                                                                                       //
SandstormDb.fillInProfileDefaults = function (credential, profile) {                                                   // 186
  // Fill in a user profile based on the data obtained from a linked credential.                                       // 187
  //                                                                                                                   // 188
  // `credential` is the source credential object.                                                                     // 189
  // `profile` is the output profile, usually located in the account object.                                           // 190
  //                                                                                                                   // 191
  // This function will only fill in details that aren't already present.                                              // 192
  if (!profile) {                                                                                                      // 194
    throw new Error("missing profile (maybe using old call signature?)");                                              // 195
  }                                                                                                                    // 196
                                                                                                                       //
  var services = credential.services || {};                                                                            // 198
                                                                                                                       //
  if (services.github) {                                                                                               // 200
    profile.name = profile.name || services.github.username || "Name Unknown";                                         // 201
    profile.handle = profile.handle || filterHandle(services.github.username) || filterHandle(profile.name);           // 202
  } else if (services.google) {                                                                                        // 204
    profile.name = profile.name || services.google.name || "Name Unknown";                                             // 205
    profile.handle = profile.handle || emailToHandle(services.google.email) || filterHandle(profile.name);             // 206
    profile.pronoun = profile.pronoun || GENDERS[services.google.gender] || "neutral";                                 // 208
  } else if (services.email) {                                                                                         // 209
    var email = services.email.email;                                                                                  // 210
    profile.name = profile.name || emailToHandle(email);                                                               // 211
    profile.handle = profile.handle || emailToHandle(email);                                                           // 212
  } else if (services.dev) {                                                                                           // 213
    var lowerCaseName = services.dev.name.split(" ")[0].toLowerCase();                                                 // 214
    profile.name = profile.name || services.dev.name;                                                                  // 215
    profile.handle = profile.handle || filterHandle(lowerCaseName);                                                    // 216
    profile.pronoun = profile.pronoun || (_.contains(["alice", "carol", "eve"], lowerCaseName) ? "female" : _.contains(["bob", "dave"], lowerCaseName) ? "male" : "neutral");
  } else if (services.demo) {                                                                                          // 220
    profile.name = profile.name || "Demo User";                                                                        // 221
    profile.handle = profile.handle || "demo";                                                                         // 222
  } else if (services.ldap) {                                                                                          // 223
    var setting = Settings.findOne({                                                                                   // 224
      _id: "ldapNameField"                                                                                             // 224
    });                                                                                                                // 224
    var key = setting ? setting.value : "";                                                                            // 225
    profile.handle = profile.handle || services.ldap.username;                                                         // 226
    profile.name = profile.name || services.ldap.rawAttrs[key] || profile.handle;                                      // 227
  } else if (services.saml) {                                                                                          // 228
    profile.handle = profile.handle || emailToHandle(services.saml.email);                                             // 229
    profile.name = profile.name || services.saml.displayName || profile.handle;                                        // 230
  } else {                                                                                                             // 231
    throw new Error("unrecognized authentication service: " + SandstormDb.getServiceName(credential));                 // 232
  }                                                                                                                    // 234
                                                                                                                       //
  profile.pronoun = profile.pronoun || "neutral"; // Base identicon on primary credential so that it tends to be consistent across servers.
                                                                                                                       //
  profile.identicon = credential._id;                                                                                  // 239
};                                                                                                                     // 240
                                                                                                                       //
SandstormDb.getIntrinsicName = function (credential, usePrivate) {                                                     // 242
  var services = credential.services;                                                                                  // 243
                                                                                                                       //
  if (services.github) {                                                                                               // 244
    return services.github.username;                                                                                   // 245
  } else if (services.google) {                                                                                        // 246
    return usePrivate ? services.google.email : services.google.name;                                                  // 247
  } else if (services.email) {                                                                                         // 248
    return services.email.email;                                                                                       // 249
  } else if (services.dev) {                                                                                           // 250
    return services.dev.name;                                                                                          // 251
  } else if (services.demo) {                                                                                          // 252
    return "demo on " + credential.createdAt.toISOString().substring(0, 10);                                           // 253
  } else if (services.ldap) {                                                                                          // 254
    return services.ldap.username;                                                                                     // 255
  } else if (services.saml) {                                                                                          // 256
    return services.saml.id;                                                                                           // 257
  } else {                                                                                                             // 258
    throw new Error("unrecognized authentication service: " + SandstormDb.getServiceName(credential));                 // 259
  }                                                                                                                    // 261
};                                                                                                                     // 262
                                                                                                                       //
SandstormDb.getServiceName = function (credential) {                                                                   // 264
  var keys = Object.keys(credential.services).filter(function (k) {                                                    // 265
    return k !== "resume";                                                                                             // 265
  });                                                                                                                  // 265
                                                                                                                       //
  if (keys.length !== 1) {                                                                                             // 266
    throw new Error("expected exactly one auth service: " + keys.join(","));                                           // 267
  }                                                                                                                    // 268
                                                                                                                       //
  return keys[0];                                                                                                      // 269
};                                                                                                                     // 270
                                                                                                                       //
SandstormDb.getLoginId = function (credential) {                                                                       // 272
  return Accounts.loginServices[SandstormDb.getServiceName(credential)].getLoginId(credential);                        // 273
};                                                                                                                     // 274
                                                                                                                       //
SandstormDb.prototype.getAccountIntrinsicNames = function (account, usePrivate) {                                      // 276
  // Get the user's identity badges, for display to other users. For example, if the user uses                         // 277
  // GitHub login, this would be their GitHub username. This is displayed to other users to prove                      // 278
  // this user's identity.                                                                                             // 279
  // TODO(someday): Perhaps users should be able to choose which badges are displayed publicly. For                    // 281
  //   now, as a heuristic, we choose the credentials which the user has approved for login.                           // 282
  var credentialIds = account.loginCredentials.map(function (cred) {                                                   // 283
    return cred.id;                                                                                                    // 283
  });                                                                                                                  // 283
  return Meteor.users.find({                                                                                           // 284
    _id: {                                                                                                             // 284
      $in: credentialIds                                                                                               // 284
    }                                                                                                                  // 284
  }).map(function (credential) {                                                                                       // 284
    return {                                                                                                           // 285
      service: SandstormDb.getServiceName(credential),                                                                 // 286
      name: SandstormDb.getIntrinsicName(credential, usePrivate) // TODO(soon): Add profile link?                      // 287
                                                                                                                       //
    };                                                                                                                 // 285
  });                                                                                                                  // 290
};                                                                                                                     // 291
                                                                                                                       //
SandstormDb.getVerifiedEmailsForCredential = function (credential) {                                                   // 293
  var services = credential.services;                                                                                  // 294
                                                                                                                       //
  if (services.google && services.google.email && services.google.verified_email) {                                    // 295
    // jscs:ignore requireCamelCaseOrUpperCaseIdentifiers                                                              // 296
    return [{                                                                                                          // 297
      email: services.google.email,                                                                                    // 297
      primary: true                                                                                                    // 297
    }];                                                                                                                // 297
  } else if (services.email) {                                                                                         // 298
    return [{                                                                                                          // 299
      email: services.email.email,                                                                                     // 299
      primary: true                                                                                                    // 299
    }];                                                                                                                // 299
  } else if (services.github && services.github.emails) {                                                              // 300
    return _.chain(services.github.emails).filter(function (email) {                                                   // 301
      return email.verified;                                                                                           // 302
    }).map(function (email) {                                                                                          // 302
      return _.pick(email, "email", "primary");                                                                        // 303
    }).value();                                                                                                        // 303
  } else if (services.ldap) {                                                                                          // 305
    // TODO(cleanup): don't create a new SandstormDb here, make this non-static                                        // 306
    var email = services.ldap.rawAttrs[new SandstormDb().getLdapEmailField()];                                         // 307
                                                                                                                       //
    if (email) {                                                                                                       // 308
      return [{                                                                                                        // 309
        email: email,                                                                                                  // 309
        primary: true                                                                                                  // 309
      }];                                                                                                              // 309
    }                                                                                                                  // 310
  } else if (services.saml && services.saml.email) {                                                                   // 311
    return [{                                                                                                          // 312
      email: services.saml.email,                                                                                      // 312
      primary: true                                                                                                    // 312
    }];                                                                                                                // 312
  }                                                                                                                    // 313
                                                                                                                       //
  return [];                                                                                                           // 315
};                                                                                                                     // 316
                                                                                                                       //
SandstormDb.prototype.findCredentialsByEmail = function (email) {                                                      // 318
  // Returns an array of credentials which have the given email address as one of their verified                       // 319
  // addresses.                                                                                                        // 320
  check(email, String); // For LDAP, the field containing the e-mail address is configurable...                        // 322
                                                                                                                       //
  var ldapQuery = {};                                                                                                  // 325
  ldapQuery["services.ldap.rawAttrs." + this.getLdapEmailField()] = email;                                             // 326
  return Meteor.users.find({                                                                                           // 328
    $or: [{                                                                                                            // 328
      "services.google.email": email                                                                                   // 329
    }, {                                                                                                               // 329
      "services.email.email": email                                                                                    // 330
    }, {                                                                                                               // 330
      "services.github.emails.email": email                                                                            // 331
    }, ldapQuery, {                                                                                                    // 331
      "services.saml.email": email                                                                                     // 333
    }]                                                                                                                 // 333
  }).fetch().filter(function (credential) {                                                                            // 328
    // Verify that the email is verified, since our query doesn't technically do that.                                 // 335
    return !!_.findWhere(SandstormDb.getVerifiedEmailsForCredential(credential), {                                     // 336
      email: email                                                                                                     // 336
    });                                                                                                                // 336
  });                                                                                                                  // 337
};                                                                                                                     // 338
                                                                                                                       //
SandstormDb.prototype.findAccountsByEmail = function (email) {                                                         // 340
  var credentialIds = _.pluck(this.findCredentialsByEmail(email), "_id");                                              // 341
                                                                                                                       //
  return Meteor.users.find({                                                                                           // 342
    $or: [{                                                                                                            // 342
      "loginCredentials.id": {                                                                                         // 343
        $in: credentialIds                                                                                             // 343
      }                                                                                                                // 343
    }, {                                                                                                               // 343
      "nonloginCredentials.id": {                                                                                      // 344
        $in: credentialIds                                                                                             // 344
      }                                                                                                                // 344
    }]                                                                                                                 // 344
  }).fetch();                                                                                                          // 342
};                                                                                                                     // 346
                                                                                                                       //
SandstormDb.fillInPictureUrl = function (user) {                                                                       // 348
  var staticHost = httpProtocol + "//" + makeWildcardHost("static");                                                   // 349
  var url = staticAssetUrl(user.profile.picture, staticHost);                                                          // 350
                                                                                                                       //
  if (!url && user.profile && user.profile.identicon) {                                                                // 351
    url = makeIdenticon(user.profile.identicon);                                                                       // 352
  }                                                                                                                    // 353
                                                                                                                       //
  if (!url && user.type === "credential") {                                                                            // 354
    url = makeIdenticon(user._id);                                                                                     // 355
  }                                                                                                                    // 356
                                                                                                                       //
  user.profile.pictureUrl = url;                                                                                       // 358
};                                                                                                                     // 359
                                                                                                                       //
SandstormDb.getUserCredentialIds = function (user) {                                                                   // 361
  // Given an account user object, returns an array containing the ID of each credential linked to                     // 362
  // the account. Always returns the most recently added login credential first.                                       // 363
  if (user && user.loginCredentials) {                                                                                 // 364
    return _.pluck(user.nonloginCredentials.concat(user.loginCredentials), "id").reverse();                            // 365
  } else {                                                                                                             // 366
    return [];                                                                                                         // 367
  }                                                                                                                    // 368
};                                                                                                                     // 369
                                                                                                                       //
SandstormDb.getUserEmails = function (user) {                                                                          // 371
  // Given a user object, returns an array containing all email addresses associated with that user.                   // 372
  // Each entry in the array is an object of the form:                                                                 // 373
  //     `{email: String, verified: Bool, primary: Optional(Bool)}`                                                    // 374
  //                                                                                                                   // 375
  // At most one entry in the result has `primary = true`.                                                             // 376
  //                                                                                                                   // 377
  // TODO(cleanup): This actually does need to query the database to fetch profile information                         // 378
  //   for linked credentials, so it probably makes more sense for it to be a non-static method                        // 379
  //   on SandstormDb.                                                                                                 // 380
  var credentialIds = SandstormDb.getUserCredentialIds(user);                                                          // 382
  var verifiedEmails = {};                                                                                             // 383
  var unverifiedEmails = {};                                                                                           // 384
  credentialIds.forEach(function (id) {                                                                                // 386
    var credential = Meteor.users.findOne({                                                                            // 387
      _id: id                                                                                                          // 387
    });                                                                                                                // 387
                                                                                                                       //
    if (credential && credential.services) {                                                                           // 388
      SandstormDb.getVerifiedEmailsForCredential(credential).forEach(function (verifiedEmail) {                        // 389
        if (verifiedEmail) {                                                                                           // 390
          verifiedEmails[verifiedEmail.email] = true;                                                                  // 391
        }                                                                                                              // 392
      });                                                                                                              // 393
    }                                                                                                                  // 394
                                                                                                                       //
    if (credential && credential.unverifiedEmail) {                                                                    // 396
      unverifiedEmails[credential.unverifiedEmail] = true;                                                             // 397
    }                                                                                                                  // 398
  });                                                                                                                  // 399
  var result = [];                                                                                                     // 401
                                                                                                                       //
  _.keys(verifiedEmails).map(function (email) {                                                                        // 402
    result.push({                                                                                                      // 403
      email: email,                                                                                                    // 403
      verified: true,                                                                                                  // 404
      primary: email === user.primaryEmail                                                                             // 405
    });                                                                                                                // 403
  });                                                                                                                  // 406
                                                                                                                       //
  _.keys(unverifiedEmails).map(function (email) {                                                                      // 408
    if (!(email in verifiedEmails)) {                                                                                  // 409
      result.push({                                                                                                    // 409
        email: email,                                                                                                  // 409
        verified: false                                                                                                // 409
      });                                                                                                              // 409
    }                                                                                                                  // 409
  }); // If `user.primaryEmail` is not among the verified emails, mark the first as primary.                           // 410
                                                                                                                       //
                                                                                                                       //
  if (!(user.primaryEmail in verifiedEmails) && result.length > 0) {                                                   // 413
    result[0].primary = true;                                                                                          // 414
  }                                                                                                                    // 415
                                                                                                                       //
  return result;                                                                                                       // 417
};                                                                                                                     // 418
                                                                                                                       //
SandstormDb.prototype.addContact = function () {                                                                       // 420
  function addContact(ownerId, contactId) {                                                                            // 420
    check(ownerId, String);                                                                                            // 421
    check(contactId, String);                                                                                          // 422
    var profile = Meteor.users.findOne(contactId).profile;                                                             // 423
    this.collections.contacts.upsert({                                                                                 // 424
      ownerId: ownerId,                                                                                                // 424
      accountId: contactId                                                                                             // 424
    }, {                                                                                                               // 424
      ownerId: ownerId,                                                                                                // 425
      petname: profile && profile.name,                                                                                // 426
      created: new Date(),                                                                                             // 427
      accountId: contactId                                                                                             // 428
    });                                                                                                                // 424
  }                                                                                                                    // 430
                                                                                                                       //
  return addContact;                                                                                                   // 420
}();                                                                                                                   // 420
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}}}}},{
  "extensions": [
    ".js",
    ".json"
  ]
});
require("./node_modules/meteor/sandstorm-db/db.js");
require("./node_modules/meteor/sandstorm-db/profile.js");

/* Exports */
if (typeof Package === 'undefined') Package = {};
(function (pkg, symbols) {
  for (var s in symbols)
    (s in pkg) || (pkg[s] = symbols[s]);
})(Package['sandstorm-db'] = {}, {
  SandstormDb: SandstormDb
});

})();

//# sourceMappingURL=sandstorm-db.js.map
